#!/bin/sh
# -*- TCL -*-
# the next line restarts using wish \
exec wish8.3 "$0" ${1+"$@"}

#
# Where am I?
#

if {[catch {::freewrap::iswrapped $::argv0}] || \
	[::freewrap::iswrapped $::argv0] == 0} {
    #
    # Not wrapped.
    #
    set homedir [file dirname [info script]]
    set selfdir $homedir
} else {
    #
    # Probably wrapped. Try to find my home directory using $argv0
    #
    set homedir [file join [pwd] [file dirname $::argv0]]
    set selfdir [file dirname [info script]]
}

#
# Load Packages
#

foreach package {
    BWidget-1.3.1
    combat-0.7
    dom_pretty_print-0.2
    tcldom-2.0
    tclxml-2.0
    tkcon-2.3
    tcl++
} {
    if {[file exists [file join $::selfdir lib $package]]} {
	lappend auto_path [file join $::selfdir lib $package]
    } elseif {[file exists [file join $::homedir lib $package]]} {
	lappend auto_path [file join $::homedir lib $package]
    } else {
	# file exists does not seem to work in freewrap
	lappend auto_path [file join $::selfdir lib $package]
    }
}

package require Tk
package require BWidget
package require combat
package require dom
package require dom_pretty_print

#
# create Ass namespace
#

namespace eval Ass {
    variable init

    if {![info exists init]} {
	set init 0
    }
}

#
# ======================================================================
# Set up GUI
# ======================================================================
#

proc Ass::InitGUI {} {
    variable notebook
    variable mainframe
    variable packwidgets
    variable homewidgets
    variable compwidgets

    #
    # Main Frame
    #

    option add *TitleFrame.l.font {helvetica 10 bold}
    option add *Menu*tearOff 0

    set menudesc {
        "&File" "" "" 0 {
            {command "&New" {} "Clear Assembly" {Ctrl n} -command Ass::FileNew}
            {command "&Setup" {} "Set Configuration Values" "" -command "Ass::Setup"}
	    {separator}
            {command "E&xit" {} "Exit Assembly Tool" {Ctrl x} -command "Ass::ExitAss"}
	}
	"&Packages" "" "" 0 {
	    {command "Open &Package" {} "Open Software Package (.zip or .csd)" {Ctrl p} -command "Ass::PackOpenSoftpkg"}
	    {command "Download Package" {} "Download Software Package" "" -command "Ass::PackDownloadSoftpkg"}
	    {separator}
	    {command "Package &Wizard" {} "Run Software Package Wizard" {Ctrl w} -command "Ass::SoftpkgWizard"}
	}
        "&Assembly" "" "" 0 {
            {command "&Open Assembly" {} "Open Assembly (.cad or .aar)" {Ctrl o} -command "Ass::FileOpenAssembly"}
            {command "Download Assembly" {} "Download Assembly" "" -command "Ass::FileDownloadAssembly"}
            {command "&Save Assembly" {} "Save Assembly Descriptor (.cad)" {Ctrl s} -command "Ass::FileSaveAssembly"}
            {command "Build &Archive" {} "Build Assembly Archive (.aar)" {Ctrl a} -command "Ass::FileBuildPackage"}
	    {separator}
            {command "&Deploy Assembly" {} "Deploy this Assembly" {Ctrl d} -command "Ass::FileDeployAssembly"}
        }
	"&Help" "" "" 0 {
	    {command "&Index" {} "Using the MicoCCM Assembly Tool" "" -command "Ass::HelpIndex"}
	    {separator}
	    {command "&Glossary" {} "CCM Packaging and Deployment Terminology" "" -command "Ass::HelpGlossary"}
	    {command "&Deployment Tool" {} "Using the MicoCCM Deployment Tool" "" -command "Ass::HelpDeploy"}
	    {separator}
	    {command "&Credits" {} "Info about Third Party Packages" "" -command "Ass::Credits"}
	    {command "&License" {} "Assembly and Deployment Toolkit License" "" -command "Ass::License"}
	    {command "&About" {} "About the MicoCCM Assembly Tool" "" -command "Ass::About"}
	}
    }

    set mainframe [MainFrame .mainframe \
	    -textvariable ::Ass::status \
	    -menu $menudesc]

    set frame [$mainframe getframe]
    set notebook [NoteBook $frame.nb -font {helvetica 12 bold}]

    #
    # Create Packages page of Notebook
    #

    set pnote [$notebook insert end Packages -text "Packages"]
    set ptitle [TitleFrame $pnote.ptitle -text "Available Packages"]
    set pframe [$ptitle getframe]
    set ppane [PanedWindow $pframe.pane -side top -pad 5]
    set pleft [$ppane add]
    set pright [$ppane add]

    # left pane: tree view of all packages

    set psw [ScrolledWindow $pleft.sw \
	    -relief sunken -borderwidth 2]
    set packtree [Tree $psw.tree -width 35 -height 25 \
	    -relief flat -borderwidth 0 -highlightthickness 0 \
	    -opencmd "Ass::PackOpenNode"]
    $psw setwidget $packtree

    $packtree bindText <ButtonPress-1> "Ass::PackSelect"
    $packtree bindImage <ButtonPress-1> "Ass::PackSelect"
    $packtree bindText <ButtonPress-3> "Ass::PackPopup"
    $packtree bindImage <ButtonPress-3> "Ass::PackPopup"

    pack $psw -side top -expand yes -fill both
    pack $ptitle -fill both -expand yes -pady 5

    # right pane: package properties

    set ppproptitle [TitleFrame $pright.proptitle -text "Package Properties"]
    set ppprops [$ppproptitle getframe]

    set pppropid [LabelEntry $ppprops.id -label "Identifier" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled]
    pack $pppropid -side top -fill x -expand yes -pady 2

    set pppropfile [LabelEntry $ppprops.file -label "Package" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled]
    pack $pppropfile -side top -fill x -expand yes -pady 2

    set ppropempty1 [frame $ppprops.empty1]
    pack $ppropempty1 -side top -fill x -expand yes -pady 10

    set pproploclab [label $ppprops.loclab -anchor w \
	    -text "Implementation Location"]
    pack $pproploclab -side top -fill x -expand yes -pady 2

    set pppropidlurl [LabelEntry $ppprops.idlurl -label "IDL File" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled]
    pack $pppropidlurl -side top -fill x -expand yes -pady 2

    set pppropliburl [LabelEntry $ppprops.liburl -label "Impl File" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled]
    pack $pppropliburl -side top -fill x -expand yes -pady 2

    set ppropempty2 [frame $ppprops.empty2]
    pack $ppropempty2 -side top -fill x -expand yes -pady 10

    set ppropreflab [label $ppprops.reflab -anchor w \
	    -text "In an Assembly Archive,"]
    pack $ppropreflab -side top -fill x -expand yes -pady 2

    set pproprefpkg [radiobutton $ppprops.refpkg \
	    -text "Reference this Package" -value "refpkg" \
	    -state disabled -anchor w]
    $pproprefpkg configure -activebackground [$pproprefpkg cget -bg]
    pack $pproprefpkg -side top -fill x -expand yes -pady 2

    set ppropreffiles [radiobutton $ppprops.reffiles \
	    -text "Reference Impl and IDL files" -value "reffiles" \
	    -state disabled -anchor w]
    $ppropreffiles configure -activebackground [$ppropreffiles cget -bg]
    pack $ppropreffiles -side top -fill x -expand yes -pady 2

    set pproprefnoref [radiobutton $ppprops.refnoref \
	    -text "Include all files in the archive" -value "noref" \
	    -state disabled -anchor w]
    $pproprefnoref configure -activebackground [$pproprefnoref cget -bg]
    pack $pproprefnoref -side top -fill x -expand yes -pady 2

    # pack packages notebook

    pack $ppproptitle -side top -fill x -pady 5
    pack $ppane -fill both -expand yes
    pack $ptitle -fill both -expand yes -pady 2

    # store package widget names

    set packwidgets(tree) $packtree
    set packwidgets(id) $pppropid
    set packwidgets(pkgurl) $pppropfile
    set packwidgets(idlurl) $pppropidlurl
    set packwidgets(liburl) $pppropliburl
    set packwidgets(refpkg) $pproprefpkg
    set packwidgets(reffiles) $ppropreffiles
    set packwidgets(refnoref) $pproprefnoref

    #
    # Create Homes page of Notebook
    #

    set hnote [$notebook insert end Homes -text "Homes"]
    set htitle [TitleFrame $hnote.htitle -text "Home Instances"]
    set hframe [$htitle getframe]
    set hpane [PanedWindow $hframe.pane -side top -pad 5]
    set hleft [$hpane add]
    set hright [$hpane add]

    # left pane: tree view of all homes and attributes

    set hlscroll [ScrolledWindow $hleft.sw \
	    -relief sunken -borderwidth 2]
    set hometree [Tree $hlscroll.tree -width 25 -height 25 \
	    -relief flat -borderwidth 0 -highlightthickness 0 \
	    -opencmd "Ass::HomeOpenNode"]
    $hlscroll setwidget $hometree

    $hometree bindText <ButtonPress-1> "Ass::HomeSelect"
    $hometree bindImage <ButtonPress-1> "Ass::HomeSelect"
    $hometree bindText <ButtonPress-3> "Ass::HomePopup"
    $hometree bindImage <ButtonPress-3> "Ass::HomePopup"

    pack $hlscroll -side top -expand yes -fill both

    # right pane: home properties and attribute values

    # right pane top: home properties

    set hrightproptitle [TitleFrame $hright.proptitle -text "Home Properties"]
    set hrightprops [$hrightproptitle getframe]

    set hrightpropid [LabelEntry $hrightprops.id -label "Identifier" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled \
	    -command Ass::HomeUpdateId \
	    -helptext "Homes and Components must have an identifier that \
	    is unique within the Assembly"]
    pack $hrightpropid -side top -fill x -expand yes -pady 2
    bind $hrightpropid <FocusOut> Ass::HomeUpdateId

    set hrightpropnsreg [checkbutton $hrightprops.nsreg \
	    -text "Register with Naming Service" -variable "" \
	    -anchor w -state disabled -highlightthickness 0]
    $hrightpropnsreg configure -activebackground [$hrightpropnsreg cget -bg]
    pack $hrightpropnsreg -side top -fill x -expand yes -pady 2

    set hrightpropnsname [LabelEntry $hrightprops.nsname -label "NS Name" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled \
	    -helptext "The Name for this home in the Naming Service"]
    pack $hrightpropnsname -side top -fill x -expand yes -pady 2

    # right pane bot: home attributes

    set hrightattrtitle [TitleFrame $hright.attrtitle \
	    -text "Attribute Properties"]
    set hrightattrs [$hrightattrtitle getframe]

    set hrightdummylab [label $hrightattrs.dummy \
	    -text "No Attribute Selected" \
	    -anchor c]
    pack $hrightdummylab -fill both -expand yes

    # pack right pane

    pack $hrightproptitle -side top -fill x -pady 5
    pack $hrightattrtitle -side top -fill both -expand yes -pady 5

    # pack home notebook

    pack $hpane -fill both -expand yes
    pack $htitle -fill both -expand yes -pady 2

    # store homes widget names

    set homewidgets(tree) $hometree
    set homewidgets(id) $hrightpropid
    set homewidgets(nsreg) $hrightpropnsreg
    set homewidgets(nsname) $hrightpropnsname
    set homewidgets(attrs) $hrightattrs

    #
    # Create Components page of Notebook
    #

    set cnote [$notebook insert end Components -text "Components"]
    set ctitle [TitleFrame $cnote.htitle -text "Component Instances"]
    set cframe [$ctitle getframe]
    set cpane [PanedWindow $cframe.pane -side top -pad 5]
    set cleft [$cpane add]
    set cright [$cpane add]

    # left pane: tree view of all components and their ports

    set clscroll [ScrolledWindow $cleft.sw \
	    -relief sunken -borderwidth 2]
    set comptree [Tree $clscroll.tree -width 25 -height 25 \
	    -relief flat -borderwidth 0 -highlightthickness 0 \
	    -opencmd "Ass::CompOpenNode"]
    $clscroll setwidget $comptree

    $comptree bindText <ButtonPress-1> "Ass::CompSelect"
    $comptree bindImage <ButtonPress-1> "Ass::CompSelect"
    $comptree bindText <ButtonPress-3> "Ass::CompPopup"
    $comptree bindImage <ButtonPress-3> "Ass::CompPopup"

    pack $clscroll -side top -expand yes -fill both

    # right pane: component properties and attributes

    # right pane top: component properties

    set crightproptitle [TitleFrame $cright.proptitle \
	    -text "Component Properties"]
    set crightprops [$crightproptitle getframe]

    set crightpropid [LabelEntry $crightprops.id -label "Identifier" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled \
	    -command Ass::CompUpdateId \
	    -helptext "Homes and Components must have an identifier that \
	    is unique within the Assembly"]
    pack $crightpropid -side top -fill x -expand yes -pady 2
    bind $crightpropid <FocusOut> Ass::CompUpdateId

    #
    # Not supported yet
    #
    #
    # set crightpropccc [checkbutton $crightprops.ccc \
    #	    -text "Call configuration_complete" \
    #       -anchor w -state disabled -highlightthickness 0]
    # $crightpropccc configure -activebackground [$crightpropccc cget -bg]
    # pack $crightpropccc -side top -fill x -expand yes -pady 2

    set crightpropnsreg [checkbutton $crightprops.nsreg \
	    -text "Register with Naming Service" -variable "" \
	    -anchor w -state disabled -highlightthickness 0]
    $crightpropnsreg configure -activebackground [$crightpropnsreg cget -bg]
    pack $crightpropnsreg -side top -fill x -expand yes -pady 2

    set crightpropnsname [LabelEntry $crightprops.nsname -label "NS Name" \
	    -labelwidth 12 -width 40 -labelanchor w -state disabled \
	    -helptext "The Name for this component in the Naming Service"]
    pack $crightpropnsname -side top -fill x -expand yes -pady 2

    # right pane bot: component attribute/port properties

    set crightattrtitle [TitleFrame $cright.porttitle \
	    -text "Port Properties"]
    set crightattrs [$crightattrtitle getframe]

    set crightdummylab [label $crightattrs.dummy -text "No Port Selected" \
	    -anchor c]
    pack $crightdummylab -fill both -expand yes

    # pack right pane

    pack $crightproptitle -side top -fill x -pady 5
    pack $crightattrtitle -side top -fill both -expand yes -pady 5

    # pack components notebook

    pack $cpane -fill both -expand yes
    pack $ctitle -fill both -expand yes -pady 2

    # store component widget names

    set compwidgets(tree) $comptree
    set compwidgets(id) $crightpropid
    #set compwidgets(ccc) $crightpropccc
    set compwidgets(nsreg) $crightpropnsreg
    set compwidgets(nsname) $crightpropnsname
    set compwidgets(attrtitle) $crightattrtitle
    set compwidgets(attrs) $crightattrs

    #
    # Finalize main frame
    #

    $notebook compute_size
    pack $notebook -fill both -expand yes -padx 4 -pady 4
    $notebook raise [$notebook pages 0]
    pack $mainframe -fill both -expand yes

    #
    # Some bindings
    #

    bind . <F1> "Ass::HelpIndex"
    bind . <Control-r> "Ass::Refresh"
    bind . <Control-l> "Ass::Reload"
    bind . <Control-c> "Ass::ToggleConsole"
    bind . <Control-t> "Ass::Tetris"
}

proc Ass::ExitAss {} {
    global forever
    set forever 1
}

proc Ass::Refresh {} {
    variable mainframe
    variable status

    set status "Refreshing ..."

    catch {destroy $mainframe}

    InitGUI
    InitImages

    PackDisplayAllPackages
    HomeDisplayAllHomes
    CompDisplayAllComponents

    set status "Refreshing Done."
}

proc Ass::Reload {} {
    variable status
    set status "Reloading [file tail $::argv0] ..."
    uplevel #0 {source $::argv0}
    set status "Reloading Done."
}

proc Ass::Tetris {} {
    if {![info exists ::Tetris::tetris]} {
	if {[file exists [file join $::selfdir feature.tcl]]} {
	    uplevel #0 source [file join $::selfdir feature.tcl]
	}
    }
    if {[info exists ::Tetris::tetris] && ![winfo exists .tetris]} {
	Tetris::Init
    }
}

proc Ass::ToggleConsole {} {
    variable status

    if {[catch {uplevel #0 package require tkcon}]} {
	set status "Console not available."
	return
    }

    if {[winfo exists .tkcon] && [winfo ismapped .tkcon]} {
	set status "Hiding Console."
	::tkcon hide
    } else {
	set status "Showing Console."
	set notfirst [winfo exists .tkcon]

	::tkcon show

	if {!$notfirst} {
	    bind .tkcon <Control-Key-c> "Ass::ToggleConsole"
	}
    }
}

#
# ----------------------------------------------------------------------
# Some menu items
# ----------------------------------------------------------------------
#

proc Ass::SetupSelect {option title} {
    variable initialdir
    variable status
    variable setup

    set status "Please select the $title"

    if {$::tcl_platform(platform) == "windows"} {
	set filetypes {
	    {"Executables" {*.exe}}
	    {"All Files" {*}}
	}
    } else {
	set filetypes {
	    {"All Files" {*}}
	}
    }

    if {[info exists setup($option)] && $setup($option) != ""} {
	catch {set olddir [file dirname $setup($option)]}
    }

    if {![info exists olddir] || ![file exists $olddir]} {
	set olddir $initialdir
    }

    set filename [tk_getOpenFile -title "Select $title" \
	    -parent .setupAss -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	return
    }
    
    set initialdir [file dirname $filename]
    set setup($option) $filename
}

proc Ass::Setup {} {
    variable setup
    variable status
    variable mutex

    set top [toplevel .setupAss]
    wm title $top "Setup Assembly Tool"

    set title [label $top.title -anchor center \
	    -font {Arial 18 bold} \
	    -text "Setup Assembly Tool"]
    pack $title -side top -fill x -pady 10

    set topsep [Separator $top.topsep -orient horizontal]
    pack $topsep -side top -fill x -pady 10

    set i 0
    foreach {option title} {
	ird "Interface Repository"
	idl "IDL Processor"
	zip "Zip Program"
	unzip "Unzip Program"
	kill "Kill Program"
    } {
	incr i
	frame $top.f$i
	LabelEntry $top.f$i.le -label $title \
		-labelwidth 20 -width 40 -labelanchor w \
		-textvariable ::Ass::setup($option)
	button $top.f$i.but -text "Select" -width 10 \
		-anchor c -relief raised -bd 1 \
		-command "Ass::SetupSelect $option {$title}"
	pack $top.f$i.le -side left -expand yes -fill x
	pack $top.f$i.but -side left
	pack $top.f$i -side top -expand yes -fill x -pady 5 -padx 10
    }


    frame $top.of
    LabelEntry $top.of.le -label "Options to IDL" \
	    -labelwidth 20 -width 40 -labelanchor w \
	    -textvariable ::Ass::setup(idlopts)
    pack $top.of.le -side left -expand yes -fill x
    pack $top.of -after $top.f2 -side top -fill x -pady 5 -padx 10

    set botsep [Separator $top.botsep -orient horizontal]
    pack $botsep -side top -fill x -pady 10

    set ptopf3 [frame $top.buts]
    set ptopf3but1 [button $ptopf3.but1 -width 15 \
	    -text "OK" \
	    -command "set ::Ass::mutex 1"]
    set ptopf3but2 [button $ptopf3.but2 -width 15 \
	    -text "Save" \
	    -command "set ::Ass::mutex 2"]
    set ptopf3but3 [button $ptopf3.but3 -width 15 \
	    -text "Cancel" \
	    -command "set ::Ass::mutex 3"]
    pack $ptopf3but1 $ptopf3but2 $ptopf3but3 -side left -pady 10 -padx 20
    pack $ptopf3

    foreach name [array names setup] {
	set oldsetup($name) $setup($name)
    }

    set status "Editing configuration ..."

    bind $top <Return> "set ::Ass::mutex 1"

    wm deiconify $top
    focus $top
    grab set $top
    set mutex 0
    vwait ::Ass::mutex
    grab release $top
    destroy $top

    if {$mutex != 1 && $mutex != 2} {
	foreach name [array names setup] {
	    set setup($name) $oldsetup($name)
	}
	set status "Configuration change cancelled."
	return
    }

    if {$mutex == 2} {
	set status "Saving configuration data ..."
	update

	set configfile [file join $::homedir setup.tcl]

	if {[catch {
	    set file [open $configfile w]
	    foreach name [array names setup] {
		puts $file "set ::Ass::setup($name) \"$setup($name)\""
	    }
	    close $file
	} err]} {
	    set status "Error saving config: $err"
	    return
	}

	set status "Configuration saved as $configfile."
    } else {
	set status "Configuration updated."
    }
}

proc Ass::HelpShowHelp {basename} {
    variable helpwidgets

    set helpwidgets(top) .helpAss

    if {![winfo exists $helpwidgets(top)]} {
	toplevel $helpwidgets(top)
	wm title $helpwidgets(top) "Assembly Tool Help"

	set sw [ScrolledWindow $helpwidgets(top).sw -auto horizontal]

	if {![catch {package require Tkhtml}]} {
	    set text [html $sw.html -relief sunken -bd 2]
	    set helpwidgets(html) $text
	    catch {unset helpwidgets(text)}
	} else {
	    set text [text $sw.text -relief sunken -bd 2]
	    set helpwidgets(text) $text
	    catch {unset helpwidgets(html)}
	}

	$sw setwidget $text
	pack $sw -side top -fill both -expand yes

	set botsep [Separator $helpwidgets(top).botsep -orient horizontal]
	pack $botsep -side top -fill x -pady 10

	set botframe [frame $helpwidgets(top).botframe]
	set botbut [button $botframe.but -width 10 -text "Close" \
		-command "destroy $helpwidgets(top)"]
	pack $botbut
	pack $botframe -side top -fill x -pady 10

	bind $helpwidgets(top) <Prior> "$text yview scroll -1 pages"
	bind $helpwidgets(top) <Next> "$text yview scroll 1 pages"
	bind $helpwidgets(top) <Up> "$text yview scroll -1 units"
	bind $helpwidgets(top) <Down> "$text yview scroll 1 units"
	bind $helpwidgets(top) <Home> "$text yview moveto 0"
	bind $helpwidgets(top) <End> "$text yview moveto 1"
	bind $helpwidgets(top) <Return> "destroy $helpwidgets(top)"
    }

    wm deiconify $helpwidgets(top)
    wm resizable $helpwidgets(top) 0 0
    focus $helpwidgets(top)

    if {[info exists helpwidgets(html)]} {
	append basename ".html"
	set filename [file join $::selfdir help $basename]
	set file [open $filename]
	set data [read $file]
	close $file

	$helpwidgets(html) clear
	$helpwidgets(html) parse $data
    } else {
	append basename ".txt"
	set filename [file join $::selfdir help $basename]
	set file [open $filename]
	set data [read $file]
	close $file

	$helpwidgets(text) configure -state normal
	$helpwidgets(text) delete 1.0 end
	$helpwidgets(text) insert 1.0 $data
	$helpwidgets(text) configure -state disabled
    }

    # for Windows
    after 0 "wm deiconify $helpwidgets(top)"
}

proc Ass::HelpIndex {} {
    HelpShowHelp asstool
}

proc Ass::HelpDeploy {} {
    HelpShowHelp deploy
}

proc Ass::HelpGlossary {} {
    HelpShowHelp glossary
}

proc Ass::License {} {
    HelpShowHelp license
}

proc Ass::About {} {
    variable images

    set top .aboutAss

    if {[winfo exists $top]} {
	wm deiconify $top
	focus $top
	return
    }

    toplevel $top -bg "#ffffff"
    wm title $top "About Assembly Tool"

    set title [label $top.title -anchor center \
	    -font {Arial 18 bold} -bg "#ffffff" \
	    -text "MicoCCM Assembly Tool"]
    pack $title -side top -fill x -pady 10

    set topsep [Separator $top.topsep -orient horizontal]
    pack $topsep -side top -fill x -pady 10

    set imgframe [frame $top.imgs -bg "#ffffff"]

    set imgleft [frame $imgframe.left -bg "#ffffff"]
    set imglefttop [label $imgleft.lab -width 25 -anchor center \
	    -font {Arial 14 bold} -bg "#ffffff" \
	    -text "Implemented by"]
    set imgleftbot [label $imgleft.img -bg "#ffffff" \
	    -width 200 -height 70 -anchor center \
	    -image $images(wfpxsm)]
    pack $imglefttop $imgleftbot -side top

    set imgright [frame $imgframe.right -bg "#ffffff"]
    set imgrighttop [label $imgright.lab -width 25 -anchor center \
	    -font {Arial 14 bold} -bg "#ffffff" \
	    -text "Sponsored by"]
    set imgrightbot [label $imgright.img -bg "#ffffff" \
	    -width 200 -height 70 -anchor center \
	    -image $images(alcatel)]
    pack $imgrighttop $imgrightbot -side top
    
    pack $imgleft $imgright -side left
    pack $imgframe -side top -fill x

    set pbcframe [frame $top.pbc -bg "#ffffff"]
    set pbctop [label $pbcframe.lab -width 25 -anchor center \
	    -font {Arial 14 bold} -bg "#ffffff" \
	    -text "Powered by"]
    set pbcbot [label $pbcframe.img -anchor center \
	    -bg "#ffffff" -image $images(combat)]
    pack $pbctop $pbcbot -side top
    pack $pbcframe -side top -fill x -pady 10

    set copyright [label $top.copyright -anchor center \
	    -font {Arial 14 bold} -bg "#ffffff" \
	    -text "\251 2002 Frank Pilhofer"]
    pack $copyright -side top -fill x -pady 10

    set email [label $top.email -anchor center \
	    -font {Arial 14 bold} -bg "#ffffff" \
	    -text "fp@fpx.de"]
    pack $email -side top -fill x

    set botsep [Separator $top.botsep -orient horizontal]
    pack $botsep -side top -fill x -pady 10

    set botframe [frame $top.botframe -bg "#ffffff"]
    set botbut [button $botframe.but -width 10 -text "OK" \
	    -command "destroy $top"]
    pack $botbut
    pack $botframe -side top -fill x -pady 10

    wm deiconify $top
    wm resizable $top 0 0
    focus $top

    bind $top <Return> "destroy $top"

    # for Windows
    after 0 "wm deiconify $top"
}

proc Ass::Credits {} {
    set top .credits

    if {[winfo exists $top]} {
	wm deiconify $top
	focus $top
	return
    }

    toplevel $top
    wm title $top "Assembly Tool Credits"

    set title [label $top.title -anchor center \
	    -font {Arial 18 bold} \
	    -text "Used Packages"]
    pack $title -side top -fill x -pady 10

    set topsep [Separator $top.topsep -orient horizontal]
    pack $topsep -side top -fill x -pady 10

    set packages [list bwidget tclxml xmlpp tcltk combat]
    array set packinfo {
	combat,Title "Combat"
	combat,URL "http://www.fpx.de/Combat/"
	bwidget,Title "Unifix BWidget"
	bwidget,URL "http://tcllib.sourceforge.net/"
	tclxml,Title "Steve Balls' TclXML and TclDOM"
	tclxml,URL "http://tclxml.sourceforge.net/"
	xmlpp,Title "Cameron Laird's XML Pretty Printer"
	xmlpp,URL "http://starbase.neosoft.com/~claird/"
	tcltk,Title "Tcl/Tk"
	tcltk,URL "http://tcl.sourceforge.net/"
    }

    if {![catch {package require tkcon}]} {
	lappend packages tkcon
	array set packinfo {
	    tkcon,Title "Jeffrey Hobbs' tkcon"
	    tkcon,URL "http://tkcon.sourceforge.net/"
	}
    }

    if {![catch {package require Tkhtml}]} {
	lappend packages tkhtml
	array set packinfo {
	    tkhtml,Title "Richard Hipp's Tkhtml"
	    tkhtml,URL "http://www.hwaci.com/drh/"
	}
    }

    if {$::tcl_platform(platform) == "windows"} {
	lappend packages freewrap
	array set packinfo {
	    freewrap,Title "Dennis LaBelle's Freewrap"
	    freewrap,URL "http://freewrap.sourceforge.net/"
	}
    }

    foreach pack $packages {
	set pf [frame $top.$pack]
	set tl [label $pf.title -anchor center -text $packinfo($pack,Title)]
	set ul [label $pf.url -anchor center -font {Courier 10} \
		-text $packinfo($pack,URL)]
	pack $tl $ul -side top -fill x
	pack $pf -side top -fill x -pady 5 -padx 10
    }

    set botsep [Separator $top.botsep -orient horizontal]
    pack $botsep -side top -fill x -pady 10

    set botframe [frame $top.botframe]
    set botbut [button $botframe.but -width 10 -text "OK" \
	    -command "destroy $top"]
    pack $botbut
    pack $botframe -side top -fill x -pady 10

    wm deiconify $top
    wm resizable $top 0 0
    focus $top

    bind $top <Return> "destroy $top"

    # for Windows
    after 0 "wm deiconify $top"
}

#
# ======================================================================
# Introspection via Interface Repository
# ======================================================================
#

proc Ass::StartInterfaceRepository {} {
    variable InterfaceRepository
    variable tempfiles
    variable status
    variable setup
    global ifrpid

    if {$InterfaceRepository != 0} {
	return
    }

    set status "Looking for Interface Repository ..."
    update

    corba::try {
	set InterfaceRepository \
		[corba::resolve_initial_references InterfaceRepository]
    } catch {...} {
	set InterfaceRepository 0
    }

    if {$InterfaceRepository != 0} {
        if {![$InterfaceRepository _is_a IDL:omg.org/CORBA/ComponentIR/Repository:1.0]} {
	    error "the initial Interface Repository is not a Component Interface Repository"
	}
	return
    }

    set status "Starting Interface Repository ..."
    update

    set irdfile [file join $::homedir ird.ior]
    catch {file delete $irdfile}

    corba::try {
	set ifrpid [eval exec $setup(ird) --ior $irdfile &]
    } catch {... err} {
	catch {file delete ird.ior}
	error "could not start Interface Repository $setup(ird): $err"
    }

    set status "Starting Interface Repository ... (pid $ifrpid) ..."
    update

    for {set i 0} {$i < 50} {incr i} {
	if {($i % 2) == 0} {
	    append status "."
	    update
	}

	if {[file readable $irdfile]} {
	    corba::try {
		set f [open $irdfile]
		set ior [gets $f]
		close $f
	    } catch {...} {
		catch {close $f}
		set ior ""
	    }
	    if {$ior != ""} {
		break
	    }
	}
	after 500
    }

    after 1000

    if {![info exists ior] || $ior == ""} {
	catch {exec $setup(kill) $ifrpid}
	catch {file delete $irdfile}
	set status "Error: failure starting Interface Repository"
	error "failure starting Interface Repository: IOR file did not appear"
    }

    corba::try {
	set InterfaceRepository [corba::string_to_object $ior]
	$InterfaceRepository _is_a IDL:omg.org/CORBA/Repository:1.0
    } catch {... err} {
	catch {exec $setup(kill) $ifrpid}
	catch {file delete $irdfile}
	set status "Error: could not contact Interface Repository after startup"
	error "could not contact Interface Repository after startup: $err"
    }
    
    if {![$InterfaceRepository _is_a IDL:omg.org/CORBA/ComponentIR/Repository:1.0]} {
	catch {exec setup(kill) $ifrpid}
	catch {file delete $irdfile}
	set status "Error: Interface Repository is not component-enabled"
	error "Interface Repository is not component-enabled"
    }

    lappend tempfiles $irdfile
}

proc Ass::LoadIDLFile {idlfile {options ""}} {
    variable InterfaceRepository
    variable irdata
    variable status
    variable setup

    set localname [file tail $idlfile]

    if {[info exists irdata(__files__,$idlfile)]} {
	return
    }

    StartInterfaceRepository
    set status "Loading $localname into Interface Repository ..."
    update

    #
    # Windows hack: Use "start /w", otherwise we get a strange deadlock
    #               Add dirname of idl.exe to PATH, otherwise mico-cpp
    #               is not found.
    #

    if {$::tcl_platform(platform) == "windows"} {
	set idlprog "exec start /w $setup(idl)"
	set idldir [file nativename [file dirname $setup(idl)]]
	if {[string first $idldir $::env(PATH)] == -1} {
	    append ::env(PATH) ";" $idldir
	}
    } else {
	set idlprog "tk_exec $setup(idl)"
    }

    corba::try {
	set irior [corba::object_to_string $InterfaceRepository]
	eval $idlprog --feed-ir --feed-included-defs \
		-ORBInitRef InterfaceRepository=$irior \
		$setup(idlopts) $options $idlfile
    } catch {... err} {
	set status "Error loading IDL file: $err"
	error "error loading IDL file: $err"
    }

    set irdata(__files__,$idlfile) 1
    set status "$localname loaded."
    return 1
}

proc Ass::IntrospectHome {homename} {
    variable InterfaceRepository
    variable iridtoname
    variable irdata
    variable status

    if {[info exists irdata($homename,type)]} {
	return
    }

    StartInterfaceRepository
    set status "Introspecting Home $homename ..."
    update

    corba::try {
	set homedef [$InterfaceRepository lookup $homename]
    } catch {...} {
	set homedef 0
    }

    if {$homedef == 0} {
	error "Error: cannot find $homename in Interface Repository"
    }

    if {![$homedef _is_a IDL:omg.org/CORBA/ComponentIR/HomeDef:1.0]} {
	error "Error: $homename is not a HomeDef"
    }

    set id [$homedef id]
    set irdata($homename,type) Home
    set irdata($homename,id) $id
    set iridtoname($id) $homename

    set comp [$homedef managed_component]
    set irdata($homename,managed) [$comp absolute_name]

    set basedef [$homedef base_home]
    if {$basedef == 0} {
	set irdata($homename,base) 0
    } else {
	set irdata($homename,base) [$basedef absolute_name]
    }
    
    set keydef [$homedef primary_key]

    if {$keydef == 0} {
	set irdata($homename,key) 0
    } else {
	set irdata($homename,key) [$keydef describe]
    }

    set irdata($homename,supp) [list]
    foreach supported [$homedef supported_interfaces] {
	lappend irdata($homename,supp) [$supported absolute_name]
    }

    set irdata($homename,ports) [list]
    set irdata($homename,contents) \
	    [$homedef describe_contents dk_all 0 -1]

    foreach description $irdata($homename,contents) {
	array set desc $description
	array set port [lindex $desc(value) 1]

	switch -- $desc(kind) {
	    dk_Attribute {
		lappend irdata($homename,ports) $port(name)
		set irdata($homename,port,$port(name),kind) Attribute
		set irdata($homename,port,$port(name),type) $port(type)
		set irdata($homename,port,$port(name),mode) $port(mode)
	    }
	    dk_Factory {
		lappend irdata($homename,ports) $port(name)
		set irdata($homename,port,$port(name),kind) Factory
		set irdata($homename,port,$port(name),params) \
			$port(parameters)
	    }
	}
    }

    #
    # Initialize is_a data
    #

    IntrospectInitIsA $homedef

    #
    # Introspect managed component, base home and supported interfaces
    #

    IntrospectComponent $irdata($homename,managed)

    if {$irdata($homename,base) != 0} {
	IntrospectHome $irdata($homename,base)
    }

    foreach supp $irdata($homename,supp) {
	IntrospectInterface $supp
    }

    set status "Home $homename introspected."
}

proc Ass::IntrospectComponent {compname} {
    variable InterfaceRepository
    variable irdata
    variable status

    if {[info exists irdata($compname,type)]} {
	return
    }

    StartInterfaceRepository
    set status "Introspecting Component $compname ..."
    update

    corba::try {
	set compdef [$InterfaceRepository lookup $compname]
    } catch {...} {
	set compdef 0
    }

    if {$compdef == 0} {
	error "Error: cannot find $compname in Interface Repository"
    }

    if {![$compdef _is_a IDL:omg.org/CORBA/ComponentIR/ComponentDef:1.0]} {
	error "Error: $compname is not a ComponentDef"
    }

    set id [$compdef id]
    set irdata($compname,type) Component
    set irdata($compname,id) $id
    set iridtoname($id) $compname

    set basedef [$compdef base_component]
    if {$basedef == 0} {
	set irdata($compname,base) 0
    } else {
	set irdata($homename,base) [$basedef absolute_name]
    }

    set irdata($compname,supp) [list]
    foreach supported [$compdef supported_interfaces] {
	lappend irdata($compname,supp) [$supported absolute_name]
    }

    set irdata($compname,contents) [$compdef describe_contents dk_all 0 -1]

    #
    # Initialize is_a data
    #

    IntrospectInitIsA $compdef

    #
    # Introspect base component supported interfaces
    #

    if {$irdata($compname,base) != 0} {
	IntrospectComponent $irdata($compname,base)
    }

    foreach supp $irdata($compname,supp) {
	IntrospectInterface $supp
    }

    #
    # Look at each port and introspect its interfaces and events
    #

    set irdata($compname,ports) [list]

    foreach description $irdata($compname,contents) {
	array set desc $description
	array set port [lindex $desc(value) 1]

	switch -- $desc(kind) {
	    dk_Provides -
	    dk_Uses {
		set def [$InterfaceRepository lookup_id $port(interface_type)]
		set absname [$def absolute_name]
		IntrospectInterface $absname
	    }
	    dk_Emits -
	    dk_Publishes -
	    dk_Consumes {
		set def [$InterfaceRepository lookup_id $port(event)]
		set absname [$def absolute_name]
		IntrospectValue $absname
	    }
	}

	switch -- $desc(kind) {
	    dk_Attribute {
		lappend irdata($compname,ports) $port(name)
		set irdata($compname,port,$port(name),kind) Attribute
		set irdata($compname,port,$port(name),type) $port(type)
		set irdata($compname,port,$port(name),mode) $port(mode)
	    }
	    dk_Provides {
		lappend irdata($compname,ports) $port(name)
		set irdata($compname,port,$port(name),kind) Provides
		set irdata($compname,port,$port(name),type) \
			$port(interface_type)
	    }
	    dk_Uses {
		lappend irdata($compname,ports) $port(name)
		set irdata($compname,port,$port(name),kind) Uses
		set irdata($compname,port,$port(name),type) \
			$port(interface_type)
		set irdata($compname,port,$port(name),multiplex) \
			$port(is_multiple)
	    }
	    dk_Emits {
		lappend irdata($compname,ports) $port(name)
		set irdata($compname,port,$port(name),kind) Emits
		set irdata($compname,port,$port(name),type) $port(event)
	    }
	    dk_Publishes {
		lappend irdata($compname,ports) $port(name)
		set irdata($compname,port,$port(name),kind) Publishes
		set irdata($compname,port,$port(name),type) $port(event)
	    }
	    dk_Consumes {
		lappend irdata($compname,ports) $port(name)
		set irdata($compname,port,$port(name),kind) Consumes
		set irdata($compname,port,$port(name),type) $port(event)
	    }
	}
    }

    set status "Component $compname introspected."
}

proc Ass::IntrospectInterface {ifacename} {
    variable InterfaceRepository
    variable irdata
    variable status

    if {[info exists irdata($ifacename,type)]} {
	return
    }

    StartInterfaceRepository
    set status "Introspecting Interface $ifacename ..."
    update

    corba::try {
	set ifacedef [$InterfaceRepository lookup $ifacename]
    } catch {...} {
	set ifacedef 0
    }

    if {$ifacedef == 0} {
	error "Error: cannot find $ifacename in Interface Repository"
    }

    if {![$ifacedef _is_a IDL:omg.org/CORBA/InterfaceDef:1.0]} {
	error "Error: $ifacename is not an InterfaceDef"
    }

    set id [$ifacedef id]
    set irdata($ifacename,type) Interface
    set irdata($ifacename,id) $id
    set iridtoname($id) $ifacename

    IntrospectInitIsA $ifacedef

    set status "Interface $ifacename introspected."
}

proc Ass::IntrospectValue {valuename} {
    variable InterfaceRepository
    variable irdata
    variable status

    if {[info exists irdata($valuename,type)]} {
	return
    }

    StartInterfaceRepository
    set status "Introspecting Value $valuename ..."
    update

    corba::try {
	set valuedef [$InterfaceRepository lookup $valuename]
    } catch {...} {
	set valuedef 0
    }

    if {$valuedef == 0} {
	error "Error: cannot find $valuename in Interface Repository"
    }

    if {![$valuedef _is_a IDL:omg.org/CORBA/ValueDef:1.0]} {
	error "Error: $valuename is not a ValueDef"
    }

    set id [$valuedef id]
    set irdata($valuename,type) Valuetype
    set irdata($valuename,id) $id
    set iridtoname($id) $valuename

    IntrospectInitIsA $valuedef

    set status "Value $valuename introspected."
}

proc Ass::IntrospectId {id} {
    variable InterfaceRepository

    StartInterfaceRepository

    corba::try {
	set def [$InterfaceRepository lookup_id $id]
	set absname [$def absolute_name]
    } catch {...} {
	set def 0
    }

    if {$def == 0} {
	error "Error: cannot find $id in Interface Repository"
    }

    set kind [$def def_kind]

    switch -- $kind {
	dk_Component {
	    IntrospectComponent $absname
	}
	dk_Home {
	    IntrospectHome $absname
	}
	dk_Interface {
	    IntrospectInterface $absname
	}
	dk_Value {
	    IntrospectValue $absname
	}
    }

    return $absname
}

proc Ass::IntrospectInitIsA {def} {
    variable InterfaceRepository
    variable irisadata

    StartInterfaceRepository

    if {$def == 0} {
	error "oops"
    }

    set typeid [$def id]

    if {[info exists irisadata($typeid,$typeid)]} {
	return
    }

    set irisadata($typeid,$typeid) 1

    set kind [$def def_kind]
    set bases [list]

    switch -- $kind {
	dk_Component {
	    set bases [$def supported_interfaces]
	    set base_component [$def base_component]
	    if {$base_component != 0} {
		lappend bases $base_component
	    }
	}
	dk_Home {
	    set bases [$def supported_interfaces]
	    set base_home [$def base_home]
	    if {$base_home != 0} {
		lappend bases $base_home
	    }
	}
	dk_Interface {
	    set bases [$def base_interfaces]
	}
	dk_Value {
	    set bases [$def abstract_base_values]
	    set base_value [$def base_value]
	    if {$base_value != 0} {
		lappend bases $base_value
	    }
	}
    }

    set baseids [list]
    foreach base $bases {
	lappend baseids [$base id]
	IntrospectInitIsA $base
    }

    foreach baseid $baseids {
	foreach isaentry [array names irisadata $baseid,*] {
	    set compid [lindex [split $isaentry ,] 1]
	    set irisadata($typeid,$compid) 1
	}
    }
}

proc Ass::IntrospectIsA {typeid1 typeid2} {
    variable irisadata
    if {[info exists irisadata($typeid1,$typeid2)]} {
	return $irisadata($typeid1,$typeid2)
    }
    return 0
}

#
# ======================================================================
# Management of Packages, Homes and Components
# ======================================================================
#

proc Ass::PackIntrospectIDLFile {packnode} {
    variable InterfaceRepository
    variable packdata
    variable irdata
    variable status
    variable setup

    if {$packdata($packnode,idlintrospected)} {
	return 0
    }

    set idlurl $packdata($packnode,idlurl)

    if {$packdata($packnode,idllocalfile) == ""} {
	set idllocalfile [DownloadFile $idlurl]
	set packdata($packnode,idllocalfile) $idllocalfile
    }
    
    LoadIDLFile $packdata($packnode,idllocalfile)
    set packdata($packnode,idlintrospected) 1
    return 1
}

proc Ass::HomeIntrospectHome {homenode} {
    variable nodecounter
    variable configdata
    variable homedata
    variable packdata
    variable status
    variable irdata

    if {$homedata($homenode,introspected)} {
	return
    }

    StartInterfaceRepository

    if {![info exists homedata($homenode,name)]} {
	set packnode $homedata($homenode,package)

	PackIntrospectIDLFile $packnode

	if {[info exists packdata($packnode,homename)]} {
	    set homedata($homenode,name) $packdata($packnode,homename)
	} elseif {[info exists packdata($packnode,repoid)]} {
	    set homedata($homenode,name) \
		    [IntrospectId $packdata($packnode,repoid)]
	}
    }

    set homename $homedata($homenode,name)
    IntrospectHome $homename
    set homedata($homenode,introspected) 1
    
    foreach attribute $irdata($homename,ports) {
	if {$irdata($homename,port,$attribute,kind) != "Attribute"} {
	    continue
	}

	if {$irdata($homename,port,$attribute,mode) == "ATTR_READONLY"} {
	    continue
	}

	set configdata($homenode,$attribute,kind) Attribute

	incr nodecounter
	lappend homedata($homenode,children) $nodecounter
	set homedata($nodecounter,parent) $homenode
	set homedata($nodecounter,nodetype) Attribute
	set homedata($nodecounter,name) $attribute
    }
}

proc Ass::CompIntrospectComponent {compnode} {
    variable nodecounter
    variable configdata
    variable compdata
    variable homedata
    variable irdata

    if {$compdata($compnode,introspected)} {
	return
    }

    StartInterfaceRepository

    if {![info exists compdata($compnode,name)]} {
	set homenode $compdata($compnode,home)
	HomeIntrospectHome $homenode
	set homename $homedata($homenode,name)
	set compdata($compnode,name) $irdata($homename,managed)
    }

    set compname $compdata($compnode,name)
    IntrospectComponent $compname
    set compdata($compnode,introspected) 1

    foreach portname $irdata($compname,ports) {
	set portkind $irdata($compname,port,$portname,kind)

	if {$portkind != "Attribute" && \
		$portkind != "Emits" && \
		$portkind != "Publishes" && \
		$portkind != "Consumes" && \
		$portkind != "Provides" && \
		$portkind != "Uses"} {
	    continue
	}

	switch -- $portkind {
	    Attribute {
		if {$irdata($compname,port,$portname,mode) == "ATTR_READONLY"} {
		    continue
		}
		set portimg attr
	    }
	    Provides {
		set portimg facet
	    }
	    Uses {
		set portimg recep
	    }
	    Emits {
		set portimg source
	    }
	    Publishes {
		set portimg source
	    }
	    Consumes {
		set portimg sink
	    }
	}

	set configdata($compnode,$portname,kind) $portkind

	incr nodecounter
	lappend compdata($compnode,children) $nodecounter
	set compdata($nodecounter,parent) $compnode
	set compdata($nodecounter,nodetype) $portkind
	set compdata($nodecounter,name) $portname
	set compdata($nodecounter,bitmap) $portimg
    }
}

#
# ======================================================================
# Packages Notebook Page
# ======================================================================
#

proc Ass::PackDisplayPackage {node} {
    variable packwidgets
    variable packdata
    variable images

    if {![$packwidgets(tree) exists $node]} {
	$packwidgets(tree) insert end root $node \
		-text $packdata($node,id) \
		-open 1 -drawcross auto \
		-image [Bitmap::get folder]
    }

    foreach child $packdata($node,children) {
	if {![$packwidgets(tree) exists $child]} {
	    $packwidgets(tree) insert end $node $child \
		    -text $packdata($child,text) \
		    -open 0 -drawcross never \
		    -image $images(home)
	}
    }
}

proc Ass::PackDisplayAllPackages {} {
    variable packdata

    foreach node $packdata(allpackages) {
	PackDisplayPackage $node
    }
}

proc Ass::PackOpenNode {packnode} {
    variable packwidgets
    variable packdata

    if {[$packwidgets(tree) itemcget $packnode -drawcross] == "allways"} {
	$packwidgets(tree) itemconfigure $packnode -drawcross auto
	PackDisplayPackage $packnode
    }
}

#
# Handle Softpkg files
#

proc Ass::PackOpenSoftpkg {} {
    variable PackLoadSoftpkgFile
    variable initialdir
    variable packdata
    variable status

    set status "Please a Softpkg File (.zip or .csd)"

    set filetypes {
	{"Software Package" {.zip}}
	{"Software Package Descriptor" {.csd}}
	{"All Files" {*}}
    }

    if {[info exists PackLoadSoftpkgFile] && \
	    $PackLoadSoftpkgFile != ""} {
	catch {set olddir [file dirname $PackLoadSoftpkgFile]}
    }

    if {![info exists olddir] || ![file exists $olddir]} {
	set olddir $initialdir
    }

    set filename [tk_getOpenFile -title "Select Softpkg File" \
	    -parent . -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	set status "Cancelled."
	return
    }

    set initialdir [file dirname $filename]

    if {![file exists $filename]} {
	set status "Error: $filename does not exist."
	return
    }

    set PackLoadSoftpkgFile $filename
    set packname [file rootname [file tail $filename]]
    set dirname [file dirname $filename]

    set status "Adding package $packname ..."
    update

    if {[catch {set homenode [FileReadSoftpkg $PackLoadSoftpkgFile $dirname]} err]} {
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error reading Software Package" \
		-message $err
	return
    }

    #
    # FileReadSoftpkg returns the home's node
    #

    set packnode $packdata($homenode,parent)
    PackDisplayPackage $packnode
    set status "Package $packname added."
}

proc Ass::PackDownloadSoftpkg {} {
    variable PackLoadSoftpkgURL
    variable packdata
    variable status
    variable mutex

    set top [toplevel .packDownloadSoftpkg]
    wm title $top "Open URL"
    set toplab [label $top.lab -width 50 -anchor center \
	    -text "Please enter the URL of a Software Package"]
    pack $toplab -side top -fill x -pady 10

    set sep1 [Separator $top.sep1 -orient horizontal]
    pack $sep1 -side top -fill x -pady 10

    set ptopf1 [frame $top.f1]
    set ptopf1lab [label $ptopf1.lab -text "Package" \
	    -width 15 -anchor w]
    set ptopf1ent [entry $ptopf1.ent -width 40 \
	    -textvariable Ass::PackLoadSoftpkgURL]
    pack $ptopf1lab -side left
    pack $ptopf1ent -side left -expand yes -fill x
    pack $ptopf1 -side top -fill x -padx 10

    set sep2 [Separator $top.sep2 -orient horizontal]
    pack $sep2 -side top -fill x -pady 10

    set ptopf3 [frame $top.f3]
    set ptopf3but1 [button $ptopf3.but1 -width 15 \
	    -text "Download" \
	    -command "set ::Ass::mutex 1"]
    set ptopf3but2 [button $ptopf3.but2 -width 15 \
	    -text "Cancel" \
	    -command "set ::Ass::mutex 2"]
    pack $ptopf3but1 $ptopf3but2 -side left -pady 10 -padx 20
    pack $ptopf3

    bind $ptopf1ent <Return> "focus $ptopf3but1"
    bind $ptopf3but1 <Return> "set ::Ass::mutex 1"
    bind $ptopf3but2 <Return> "set ::Ass::mutex 2"

    set status "Please enter URL ..."
    wm deiconify $top
    focus $top
    grab set $top
    set mutex 0
    vwait ::Ass::mutex
    grab release $top
    destroy $top

    if {$mutex != 1} {
	set status "Download cancelled."
	return
    }

    set status "Downloading package $PackLoadSoftpkgURL ..."
    update

    if {[catch {set homenode [FileReadSoftpkg $PackLoadSoftpkgURL]} err]} {
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error downloading Software Package" \
		-message $err
	return
    }

    #
    # FileReadSoftpkg returns the home's node
    #

    set packnode $packdata($homenode,parent)
    PackDisplayPackage $packnode
    set status "Package downloaded."
}

proc Ass::PackSelect {node} {
    variable packwidgets
    variable packdata

    set currentselnode [lindex [$packwidgets(tree) selection get] 0]

    if {$node == $currentselnode} {
	return
    }

    focus $packwidgets(tree)
    $packwidgets(tree) selection set $node
    $packwidgets(tree) see $node

    if {$packdata($node,nodetype) == "Softpkg"} {
	if {[llength $packdata($node,children)] == 1} {
	    set node [lindex $packdata($node,children) 0]
	}
    }

    switch -- $packdata($node,nodetype) {
	Home {
	    PackSelectHome $node
	}
	Softpkg {
	    PackSelectPackage $node
	}
    }
}

proc Ass::PackSelectPackage {node} {
    variable packwidgets
    variable packdata

    if {$node == "none"} {
	$packwidgets(id) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(pkgurl) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(idlurl) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(liburl) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(refpkg) configure \
		-variable "" \
		-state disabled

	$packwidgets(reffiles) configure \
		-variable "" \
		-state disabled

	$packwidgets(refnoref) configure \
		-variable "" \
		-state disabled

	return
    }

    $packwidgets(id) configure \
	    -textvariable ::Ass::packdata($node,id) \
	    -state disabled

    $packwidgets(pkgurl) configure \
	    -textvariable ::Ass::packdata($node,pkgurl) \
	    -state disabled

    $packwidgets(idlurl) configure \
	    -textvariable "" -text "" \
	    -state disabled

    $packwidgets(liburl) configure \
	    -textvariable "" -text "" \
	    -state disabled

    $packwidgets(refpkg) configure \
	    -variable "" \
	    -state disabled

    $packwidgets(reffiles) configure \
	    -variable "" \
	    -state disabled

    $packwidgets(refnoref) configure \
	    -variable "" \
	    -state disabled
}

proc Ass::PackSelectHome {node} {
    variable packwidgets
    variable packdata

    if {$node == "none"} {
	$packwidgets(id) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(pkgurl) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(idlurl) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(liburl) configure \
		-textvariable "" -text "" \
		-state disabled

	$packwidgets(refpkg) configure \
		-variable "" \
		-state disabled

	$packwidgets(reffiles) configure \
		-variable "" \
		-state disabled

	$packwidgets(refnoref) configure \
		-variable "" \
		-state disabled

	return
    }

    set packnode $packdata($node,parent)

    $packwidgets(id) configure \
	    -textvariable ::Ass::packdata($node,id) \
	    -state normal

    if {$packdata($packnode,nodetype) != "Softpkg"} {
	$packwidgets(pkgurl) configure \
		-textvariable "" -text "" \
		-state disabled
    } else {
	$packwidgets(pkgurl) configure \
		-textvariable ::Ass::packdata($node,pkgurl) \
		-state normal
    }

    $packwidgets(idlurl) configure \
	    -textvariable ::Ass::packdata($node,idlurl) \
	    -state normal

    $packwidgets(liburl) configure \
	    -textvariable ::Ass::packdata($node,liburl) \
	    -state normal

    if {$packdata($packnode,nodetype) != "Softpkg"} {
	$packwidgets(refpkg) configure \
		-variable "" \
		-state disabled
    } else {
	$packwidgets(refpkg) configure \
		-variable ::Ass::packdata($node,whatref) \
		-state normal
    }

    $packwidgets(reffiles) configure \
	    -variable ::Ass::packdata($node,whatref) \
	    -state normal

    $packwidgets(refnoref) configure \
	    -variable ::Ass::packdata($node,whatref) \
	    -state normal
}

proc Ass::PackPopup {node} {
    variable packwidgets
    variable packdata

    PackSelect $node

    if {![info exists packwidgets(popupforpkg)]} {
	set packwidgets(popupforpkg) [menu .packPopupMenuForPkg]
	$packwidgets(popupforpkg) add command \
		-label "Make Softpkg Descriptor (.csd)" \
		-command "Ass::PackSaveSoftpkg 0"
	$packwidgets(popupforpkg) add command \
		-label "Make Software Package (.zip)" \
		-command "Ass::PackSaveSoftpkg 1"
	$packwidgets(popupforpkg) add separator
	$packwidgets(popupforpkg) add command \
		-label "Remove" \
		-command "Ass::PackRemovePackage"
    }

    if {![info exists packwidgets(popupforhome)]} {
	set packwidgets(popupforhome) [menu .packPopupMenuForHome]
	$packwidgets(popupforhome) add command \
		-label "Deploy Home in Assembly" \
		-command "Ass::PackDeployHome"
	$packwidgets(popupforhome) add separator
	$packwidgets(popupforhome) add command \
		-label "Make Softpkg Descriptor (.csd)" \
		-command "Ass::PackSaveSoftpkg 0"
	$packwidgets(popupforhome) add command \
		-label "Make Software Package (.zip)" \
		-command "Ass::PackSaveSoftpkg 1"
    }

    set xpos [expr [winfo pointerx .] + 5]
    set ypos [expr [winfo pointery .] + 5]
    

    switch -- $packdata($node,nodetype) {
	Softpkg {
	    tk_popup $packwidgets(popupforpkg) $xpos $ypos
	}
	Home {
	    tk_popup $packwidgets(popupforhome) $xpos $ypos
	}
    }
}

proc Ass::PackSaveSoftpkg {zipit} {
    variable packwidgets
    variable packdata
    variable status
    variable setup

    set packnode [lindex [$packwidgets(tree) selection get] 0]

    if {$packdata($packnode,nodetype) == "Softpkg"} {
	set packnode [lindex $packdata($packnode,children) 0]
    }

    set packid $packdata($packnode,id)
    set initialfilename $packid

    if {$zipit} {
	set status "Please select a filename for the Software Package"
	append initialfilename ".zip"
	set filetypes {
	    {"Software Package" {.zip}}
	    {"All Files" {*}}
	}
    } else {
	set status "Please select a filename for the Software Package Descriptor"
	append initialfilename ".csd"
	set filetypes {
	    {"Software Package Descriptor" {.csd}}
	    {"All Files" {*}}
	}
    }

    #set olddir $initialdir
    set olddir [pwd]

    set filename [tk_getSaveFile -title "Save As ..." -parent . \
	    -filetypes $filetypes -initialdir $olddir \
	    -initialfile $initialfilename]

    if {$filename == ""} {
	set status "Cancelled."
	return
    }

    if {[file extension [file tail $filename]] == ""} {
	if {$zipit} {
	    append filename ".zip"
	} else {
	    append filename ".csd"
	}
    }

    set initialdir [file dirname $filename]
    set localname [file tail $filename]

    if {$zipit} {
	#
	# make a new directory for the archive's contents
	#

	set aardir [MkTempDir]
	set metadir [file join $aardir meta-inf]

	if {[catch {file mkdir $metadir}]} {
	    set status "Cannot create $metadir directory."
	    return
	}

	#
	# Browse package, and copy IDL files and DLL files to this directory
	#

	if {$packdata($packnode,whatref) == "noref"} {
	    set packliburl $packdata($packnode,liburl)
	    set packidlurl $packdata($packnode,idlurl)

	    set lastidx [string last / $packidlurl]
	    incr lastidx
	    set localidlname [string range $packidlurl $lastidx end]
	    
	    set status "Copying $localidlname into Package ..."
	    update

	    set idllocalfile [file join $aardir $localidlname]

	    if {[catch {DownloadFile $packidlurl $idllocalfile} err]} {
		set status "Error: cannot copy $localidlname to $aardir: $err"
		catch {file delete -force $aardir}
		return
	    }
	    
	    set lastidx [string last / $packliburl]
	    incr lastidx
	    set locallibname [string range $packliburl $lastidx end]
	    
	    set status "Copying $locallibname into Package ..."
	    update
	    
	    set liblocalfile [file join $aardir $locallibname]

	    if {[catch {DownloadFile $packliburl $liblocalfile} err]} {
		set status "Error: cannot copy $locallibname to $aardir: $err"
		catch {file delete -force $aardir}
		return
	    }
	}

	set csdfilename $packid
	append csdfilename ".csd"
	set csdfile [file join $metadir $csdfilename]
    } else {
	set csdfile $filename
    }

    FileDumpSoftpkg $packnode $csdfile $zipit

    if {$zipit} {
	set status "Compressing Package $localname ..."
	update

	catch {file delete -force $filename}

	set oldpwd [pwd]
	cd $aardir
	if {[catch {tk_exec $setup(zip) -q -r $filename .} err]} {
	    set status "Error: zipping error: $err"
	    catch {file delete -force $aardir}
	    cd $oldpwd
	    return
	}
	cd $oldpwd

	#
	# Remove temp directory
	#

	catch {file delete -force $aardir}
	set status "Done writing Software Package $localname"
    } else {
	set status "Done writing Software Package Descriptor $localname"
    }
}

proc Ass::PackDeployHome {} {
    variable unnamedhomeid
    variable packwidgets
    variable packdata
    variable irdata
    variable status

    set node [lindex [$packwidgets(tree) selection get] 0]

    if {$packdata($node,nodetype) == "Softpkg"} {
	set node [lindex $packdata($node,children) 0]
    }

    if {[catch {PackIntrospectIDLFile $node} err]} {
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error introspecting IDL file" \
		-message $err
	return
    }

    if {![info exists packdata($node,homename)]} {
	if {[info exists packdata($node,repoid)]} {
	    set packdata($node,homename) [IntrospectId $packdata($node,repoid)]
	} else {
	    set status "Cannot Deploy this home, no info"
	}
    }

    set name $packdata($node,homename)
    set relidx [string last : $name]
    set relname [string range $name [expr {$relidx+1}] end]

    if {![info exists unnamedhomeid($relname)]} {
	set unnamedhomeid($relname) 1
    } else {
	incr unnamedhomeid($relname)
    }

    if {$unnamedhomeid($relname) == 1} {
	set newhomeid $relname
    } else {
	set newhomeid "$relname ($unnamedhomeid($relname))"
    }

    set status "Adding home $newhomeid to Assembly ..."
    update

    set homenode [HomeAddHome $node $newhomeid]

    if {[catch {HomeIntrospectHome $homenode} err]} {
	set status "Error introspecting home"
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error introspecting Home" \
		-message $err
	return
    }

    HomeDisplayHome $homenode

    set status "Home $newhomeid added to Assembly."
}

proc Ass::PackRemovePackage {} {
    variable packwidgets
    variable packdata
    variable status

    set packnode [lindex [$packwidgets(tree) selection get] 0]
    set packid $packdata($packnode,id)

    if {$packdata($packnode,refs) != 0} {
	set status "Error: cannot remove $packid."
	tk_messageBox -parent . -type ok -icon info -default ok \
		-title "Cannot remove package" \
		-message "$packid cannot be removed from the assembly,\
		because there are still deployed homes from this package\
		in the assembly."
	return
    }

    PackSelectPackage none

    $packwidgets(tree) delete $packnode

    set idx [lsearch -exact $packdata(allpackages) $packnode]
    set packdata(packageshomes) [lreplace $packdata(allpackages) $idx $idx]

    catch {array unset packdata $packnode,*}

    set status "Package $packid removed."
}

#
# ======================================================================
# Homes Notebook Page
# ======================================================================
#

proc Ass::HomeDisplayHome {node} {
    variable homewidgets
    variable homedata
    variable images

    set opened $homedata($node,introspected)

    if {![$homewidgets(tree) exists $node]} {
	if {$opened} {
	    set drawcross auto
	} else {
	    set drawcross allways
	}

	$homewidgets(tree) insert end root $node \
		-text $homedata($node,id) \
		-open $opened -drawcross $drawcross \
		-image $images(home)
    }

    if {$opened} {
	foreach child $homedata($node,children) {
	    if {![$homewidgets(tree) exists $child]} {
		$homewidgets(tree) insert end $node $child \
			-text $homedata($child,name) \
			-open 0 -drawcross never \
			-image $images(attr)
	    }
	}
    }
}

proc Ass::HomeDisplayAllHomes {} {
    variable homedata

    foreach node $homedata(allhomes) {
	HomeDisplayHome $node
    }
}

proc Ass::HomeOpenNode {homenode} {
    variable homewidgets
    variable homedata

    if {!$homedata($homenode,introspected)} {
	if {[catch {HomeIntrospectHome $homenode} err]} {
	    set status "Error introspecting home"
	    tk_messageBox -parent . \
		    -type ok -icon error -default ok \
		    -title "Error introspecting Home" \
		    -message $err
	    return
	}
    }

    if {[$homewidgets(tree) itemcget $homenode -drawcross] == "allways"} {
	$homewidgets(tree) itemconfigure $homenode -drawcross auto
	HomeDisplayHome $homenode
    }
}

proc Ass::HomeSelect {node} {
    variable homewidgets
    variable homedata

    set currentselnode [lindex [$homewidgets(tree) selection get] 0]

    if {$node == $currentselnode} {
	return
    }

    focus $homewidgets(tree)
    $homewidgets(tree) selection set $node
    $homewidgets(tree) see $node

    switch -- $homedata($node,nodetype) {
	Attribute {
	    HomeSelectHome $homedata($node,parent)
	    HomeSelectAttribute $node
	}
	Home {
	    HomeSelectHome $node
	    HomeSelectAttribute none
	}
    }
}

proc Ass::HomeSelectHome {node} {
    variable homewidgets

    if {$node == "none"} {
	$homewidgets(id) configure \
		-textvariable "" -text "" \
		-state disabled
	
	$homewidgets(nsreg) configure \
		-variable "" \
		-state disabled
	
	$homewidgets(nsname) configure \
		-textvariable "" -text "" \
		-state disabled

	return
    }

    $homewidgets(id) configure \
	    -textvariable ::Ass::homedata($node,id) \
	    -state normal

    $homewidgets(nsreg) configure \
	    -variable ::Ass::homedata($node,nsreg) \
	    -state normal

    $homewidgets(nsname) configure \
	    -textvariable ::Ass::homedata($node,nsname) \
	    -state normal
}

proc Ass::HomeSelectAttribute {node} {
    variable homewidgets
    variable homedata
    variable irdata

    foreach subitem [winfo children $homewidgets(attrs)] {
	catch {destroy $subitem}
    }

    if {$node == "none"} {
	set dummy [label $homewidgets(attrs).dummy \
		-text "No Attribute Selected" -anchor c]
	pack $dummy -fill both -expand yes

	return
    }

    set attrname $homedata($node,name)
    set homenode $homedata($node,parent)
    set homename $homedata($homenode,name)
    set attrtype $irdata($homename,port,$attrname,type)
    
    AttributeMakeFrame home $homewidgets(attrs) \
	    $homenode $attrname $attrtype
}

proc Ass::HomeUpdateId {} {
    variable homewidgets
    variable homedata

    set currentsel [$homewidgets(tree) selection get]

    if {[llength $currentsel] == 0} {
	return
    }

    set node [lindex $currentsel 0]

    if {$homedata($node,nodetype) == "Attribute"} {
	set node $homedata($node,parent)
    }

    $homewidgets(tree) itemconfigure $node -text $homedata($node,id)
}

proc Ass::HomePopup {node} {
    variable HomePopupMenus
    variable homewidgets
    variable homedata
    
    focus $homewidgets(tree)
    $homewidgets(tree) selection set $node
    $homewidgets(tree) see $node

    if {$homedata($node,nodetype) != "Home"} {
	return
    }

    if {[catch {HomeIntrospectHome $node} err]} {
	set status "Error introspecting home"
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error introspecting Home" \
		-message $err
	return
    }

    set name $homedata($node,name)

    if {![info exists HomePopupMenus($name)]} {
	HomePopupBuild $name
    }

    set xpos [expr [winfo pointerx .] + 5]
    set ypos [expr [winfo pointery .] + 5]
    tk_popup $HomePopupMenus($name) $xpos $ypos
}

proc Ass::HomePopupBuild {name} {
    variable HomePopupMenus
    variable irdata

    #
    # Prepare Popup menu
    #

    if {![info exists HomePopupMenus($name)]} {
	set flatname [string map {:: _} $name]
	set HomePopupMenus($name) \
		[menu .homePopupMenus_$flatname]

	if {$irdata($name,key) == 0} {
	    $HomePopupMenus($name) add command -label "create" \
		    -command "Ass::HomeCreateComponent"
	}

	foreach factory $irdata($name,ports) {
	    if {$irdata($name,port,$factory,kind) != "Factory"} {
		continue
	    }

	    set parameters $irdata($name,port,$factory,params)

	    set parlist "("
	    set paridx 0
	    foreach par $parameters {
		array set pd $par
		if {$paridx > 0} {
		    append parlist ", "
		}
		append parlist $pd(name)
		incr paridx
	    }
	    append parlist ")"

	    set fullname "$factory $parlist"

	    $HomePopupMenus($name) add command -label $fullname \
		    -command "Ass::HomeFactoryComponent $factory"
	}

	$HomePopupMenus($name) add separator
	$HomePopupMenus($name) add command -label "Remove" \
		-command "Ass::HomeRemoveHome"
    }
}

proc Ass::HomeRemoveHome {} {
    variable homewidgets
    variable homedata
    variable packdata
    variable status

    set hnode [lindex [$homewidgets(tree) selection get] 0]
    set homeid $homedata($hnode,id)
    set homename $homedata($hnode,name)

    if {$homedata($hnode,refs) != 0} {
	set status "Error: cannot remove $homeid."
	tk_messageBox -parent . -type ok -icon info -default ok \
		-title "Cannot remove home" \
		-message "$homeid cannot be removed from the assembly,\
		because there are still component instances created\
		from this home."
	return
    }

    set status "Removing home $homeid ..."
    update

    HomeSelectHome none
    HomeSelectAttribute none

    $homewidgets(tree) delete $hnode

    set idx [lsearch -exact $homedata(allhomes) $hnode]
    set homedata(allhomes) [lreplace $homedata(allhomes) $idx $idx]

    set packnode $homedata($hnode,package)
    set package $packdata($packnode,parent)
    incr packdata($package,refs) -1

    catch {array unset homedata $hnode,*}
    catch {array unset configdata $hnode,*}

    set status "Home $homeid removed."
}

proc Ass::HomeCreateComponent {} {
    variable unnamedcompid
    variable homewidgets
    variable homedata
    variable compdata
    variable irdata
    variable status

    set hnode [lindex [$homewidgets(tree) selection get] 0]
    set homename $homedata($hnode,name)
    set compname $irdata($homename,managed)
    set relidx [string last : $compname]
    set relname [string range $compname [expr {$relidx+1}] end]

    if {![info exists unnamedcompid($relname)]} {
	set unnamedcompid($relname) 1
    } else {
	incr unnamedcompid($relname)
    }

    if {$unnamedcompid($relname) == 1} {
	set newcompid $relname
    } else {
	set newcompid "$relname ($unnamedcompid($relname))"
    }

    set status "Adding component $newcompid to Assembly ..."
    update

    set compnode [CompAddComponent $hnode $newcompid]

    if {[catch {CompIntrospectComponent $compnode} err]} {
	set status "Error introspecting component"
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error introspecting Component" \
		-message $err
	return
    }

    CompDisplayComponent $compnode

    set status "Component $newcompid added to Assembly."
}

proc Ass::HomeFactoryComponent {factoryname} {
    variable status
    set status "Sorry, component construction via factory is not implemented yet."
}

#
# ======================================================================
# Components Notebook Page
# ======================================================================
#

proc Ass::CompDisplayComponent {node} {
    variable compwidgets
    variable compdata
    variable images

    set opened $compdata($node,introspected)

    if {![$compwidgets(tree) exists $node]} {
	if {$opened} {
	    set drawcross auto
	} else {
	    set drawcross allways
	}

	$compwidgets(tree) insert end root $node \
		-text $compdata($node,id) \
		-open $opened -drawcross $drawcross \
		-image $images(comp)
    }

    if {$opened} {
	foreach child $compdata($node,children) {
	    if {![$compwidgets(tree) exists $child]} {
		$compwidgets(tree) insert end $node $child \
			-text $compdata($child,name) \
			-open 0 -drawcross never \
			-image $images($compdata($child,bitmap))
	    }
	}
    }
}

proc Ass::CompDisplayAllComponents {} {
    variable compdata

    foreach node $compdata(allcomponents) {
	CompDisplayComponent $node
    }
}

proc Ass::CompOpenNode {compnode} {
    variable compwidgets
    variable compdata

    if {!$compdata($compnode,introspected)} {
	if {[catch {CompIntrospectComponent $compnode} err]} {
	    set status "Error introspecting component"
	    tk_messageBox -parent . \
		    -type ok -icon error -default ok \
		    -title "Error introspecting Component" \
		    -message $err
	    return
	}
    }

    if {[$compwidgets(tree) itemcget $compnode -drawcross] == "allways"} {
	$compwidgets(tree) itemconfigure $compnode -drawcross auto
	CompDisplayComponent $compnode
    }
}

proc Ass::CompSelect {node} {
    variable compwidgets
    variable compdata

    set currentselnode [lindex [$compwidgets(tree) selection get] 0]

    if {$node == $currentselnode} {
	return
    }

    focus $compwidgets(tree)
    $compwidgets(tree) selection set $node
    $compwidgets(tree) see $node

    switch -- $compdata($node,nodetype) {
	Component {
	    CompSelectComp $node
	    CompSelectPort none
	}
	Attribute -
	Provides -
	Uses -
	Emits -
	Publishes -
	Consumes {
	    CompSelectComp $compdata($node,parent)
	    CompSelectPort $node
	}
    }
}

proc Ass::CompSelectComp {node} {
    variable compwidgets

    if {$node == "none"} {
	$compwidgets(id) configure \
		-textvariable "" -text "" \
		-state disabled

	#$compwidgets(ccc) configure \
	#	-variable "" \
	#	-state disabled
	
	$compwidgets(nsreg) configure \
		-variable "" \
		-state disabled
	
	$compwidgets(nsname) configure \
		-textvariable "" -text "" \
		-state disabled

	return
    }

    $compwidgets(id) configure \
	    -textvariable ::Ass::compdata($node,id) \
	    -state normal

    #$compwidgets(ccc) configure \
    #	    -variable ::Ass::compdata($node,ccc) \
    #	    -state normal

    $compwidgets(nsreg) configure \
	    -variable ::Ass::compdata($node,nsreg) \
	    -state normal

    $compwidgets(nsname) configure \
	    -textvariable ::Ass::compdata($node,nsname) \
	    -state normal
}

proc Ass::CompSelectPort {node} {
    variable compwidgets
    variable attrwidgets
    variable compdata
    variable irdata

    foreach subitem [winfo children $compwidgets(attrs)] {
	catch {destroy $subitem}
    }

    if {$node == "none"} {
	$compwidgets(attrtitle) configure -text "Port Properties"
	set dummy [label $compwidgets(attrs).dummy \
		-text "No Port Selected" -anchor c]
	pack $dummy -fill both -expand yes

	return
    }
    
    set portkind $compdata($node,nodetype)
    set portname $compdata($node,name)
    set compnode $compdata($node,parent)
    set compname $compdata($compnode,name)
    set porttype $irdata($compname,port,$portname,type)

    switch -- $portkind {
	Attribute {
	    $compwidgets(attrtitle) configure -text "Attribute Properties"
	    AttributeMakeFrame comp $compwidgets(attrs) \
		    $compnode $portname $porttype
	}
	Provides {
	    $compwidgets(attrtitle) configure -text "Facet Properties"
	    SinkPortMakeFrame comp $compwidgets(attrs) \
		    $compnode $portname $portkind $porttype
	}
	Uses {
	    if {$irdata($compname,port,$portname,multiplex)} {
		$compwidgets(attrtitle) configure \
			-text "Multiplex Receptacle Properties"
		SourcePortMakeFrame comp $compwidgets(attrs) \
			$compnode $portname $portkind $porttype
	    } else {
		$compwidgets(attrtitle) configure \
			-text "Simplex Receptacle Properties"
		SimplexSourceMakeFrame comp $compwidgets(attrs) \
			$compnode $portname $portkind $porttype
	    }
	}
	Emits {
	    $compwidgets(attrtitle) configure -text "Emitter Properties"
	    SimplexSourceMakeFrame comp $compwidgets(attrs) \
		    $compnode $portname $portkind $porttype
	}
	Publishes {
	    $compwidgets(attrtitle) configure -text "Publisher Properties"
	    SourcePortMakeFrame comp $compwidgets(attrs) \
		    $compnode $portname $portkind $porttype
	}
	Consumes {
	    $compwidgets(attrtitle) configure -text "Consumer Properties"
	    SinkPortMakeFrame comp $compwidgets(attrs) \
		    $compnode $portname $portkind $porttype
	}
    }
}

proc Ass::CompUpdateId {} {
    variable compwidgets
    variable compdata

    set currentsel [$compwidgets(tree) selection get]

    if {[llength $currentsel] == 0} {
	return
    }

    set node [lindex $currentsel 0]

    if {$compdata($node,nodetype) != "Component"} {
	set node $compdata($node,parent)
    }

    $compwidgets(tree) itemconfigure $node -text $compdata($node,id)
}

#
# Popup for Components and Ports
#

proc Ass::CompPopup {node} {
    variable compwidgets
    variable compdata
    variable irdata

    set type $compdata($node,nodetype)

    if {$type != "Component" && \
	    $type != "Provides" && \
	    $type != "Emits" && \
	    $type != "Publishes" && \
	    $type != "Uses" && \
	    $type != "Consumes"} {
	#
	# no popup for attributes
	#
	return
    }

    if {![info exists compwidgets(popup)]} {
	set compwidgets(popup) [menu .compPopupMenu]
    }

    $compwidgets(popup) delete 0 end

    if {$type == "Component"} {
	if {[catch {CompIntrospectComponent $node} err]} {
	    set status "Error introspecting component"
	    tk_messageBox -parent . \
		    -type ok -icon error -default ok \
		    -title "Error introspecting Component" \
		    -message $err
	    return
	}
	set compname $compdata($node,name)
    } else {
	set portname $compdata($node,name)
	set compnode $compdata($node,parent)
	set compname $compdata($compnode,name)
    }

    switch -- $type {
	Component {
	    CompBuildPopupForSinkPort $node $type $irdata($compname,id)
	}
	Uses -
	Emits -
	Publishes {
	    CompBuildPopupForSourcePort $node $type \
		    $irdata($compname,port,$portname,type)
	}
	Provides -
	Consumes {
	    CompBuildPopupForSinkPort $node $type \
		    $irdata($compname,port,$portname,type)
	}
	default {
	    error "oops"
	}
    }

    set xpos [expr [winfo pointerx .] + 5]
    set ypos [expr [winfo pointery .] + 5]
    tk_popup $compwidgets(popup) $xpos $ypos
}

proc Ass::CompBuildPopupForSourcePort {sourcenode sourcetype sourceid} {
    variable compwidgets
    variable compdata
    variable irdata

    #
    # Look for matching source ports
    #

    foreach compnode $compdata(allcomponents) {
	CompIntrospectComponent $compnode

	set ctype $compdata($compnode,nodetype)
	set cname $compdata($compnode,name)
	set compid $compdata($compnode,id)

	if {$ctype != "Component"} {
	    continue
	}

	if {$sourcetype == "Uses" && \
		[IntrospectIsA $irdata($cname,id) $sourceid]} {
	    $compwidgets(popup) add command -label $compid \
		    -command "Ass::CompPopupConnectPort $sourcenode $compnode"
	}

	foreach portnode $compdata($compnode,children) {
	    set sinkporttype $compdata($portnode,nodetype)
	    
	    if {$sourcetype == "Uses" && $sinkporttype != "Provides"} {
		continue
	    } elseif {($sourcetype == "Emits" || \
		    $sourcetype == "Publishes") && \
		    $sinkporttype != "Consumes"} {
		continue
	    }

	    set sinkname $compdata($portnode,name)
	    set sinkid $irdata($cname,port,$sinkname,type)

	    if {$sourcetype == "Uses" && \
		    ![IntrospectIsA $sinkid $sourceid]} {
		continue
	    } elseif {($sourcetype == "Emits" || \
		    $sourcetype == "Publishes") && \
		    ![IntrospectIsA $sourceid $sinkid]} {
		continue
	    }

	    set label "$sinkname of $compid"
		
	    $compwidgets(popup) add command -label $label \
		    -command "Ass::CompPopupConnectPort $sourcenode $portnode"
	}
    }
}

proc Ass::CompBuildPopupForSinkPort {sinknode sinktype sinkid} {
    variable compwidgets
    variable compdata
    variable irdata

    #
    # Look for matching source ports
    #

    set count 0

    foreach compnode $compdata(allcomponents) {
	CompIntrospectComponent $compnode

	set ctype $compdata($compnode,nodetype)
	set cname $compdata($compnode,name)
	set compid $compdata($compnode,id)

	if {$ctype != "Component"} {
	    continue
	}

	foreach portnode $compdata($compnode,children) {
	    set sourceporttype $compdata($portnode,nodetype)
	    
	    if {$sinktype == "Component" && $sourceporttype != "Uses"} {
		continue
	    } elseif {$sinktype == "Provides" && $sourceporttype != "Uses"} {
		continue
	    } elseif {$sinktype == "Consumes" && \
		    $sourceporttype != "Emits" && \
		    $sourceporttype != "Publishes"} {
		continue
	    }

	    set sourcename $compdata($portnode,name)
	    set sourceid $irdata($cname,port,$sourcename,type)

	    if {$sourceporttype == "Uses" && \
		    ![IntrospectIsA $sinkid $sourceid]} {
		continue
	    } elseif {($sourceporttype == "Emits" || \
		    $sourceporttype == "Publishes") && \
		    ![IntrospectIsA $sourceid $sinkid]} {
		continue
	    }

	    incr count
	    set label "$sourcename of $compid"

	    $compwidgets(popup) add command -label $label \
		    -command "Ass::CompPopupConnectPort $portnode $sinknode"
	}
    }

    if {$sinktype == "Component"} {
	if {$count} {
	    $compwidgets(popup) add separator
	}
	$compwidgets(popup) add command -label "Remove" \
		-command "Ass::CompRemove $sinknode"
    }
}

proc Ass::CompPopupConnectPort {sourcenode sinknode} {
    variable compwidgets
    variable configdata
    variable compdata
    variable irdata

    set sourcetype $compdata($sourcenode,nodetype)
    set sourceportname $compdata($sourcenode,name)
    set sourcecompnode $compdata($sourcenode,parent)
    set sourcecompname $compdata($sourcecompnode,name)

    set sinktype $compdata($sinknode,nodetype)

    if {$sinktype == "Component"} {
	set sinkcompnode $sinknode
	set sinkportname -1
    } else {
	set sinkcompnode $compdata($sinknode,parent)
	set sinkportname $compdata($sinknode,name)
    }

    if {![info exists configdata($sourcecompnode,$sourceportname,conn)]} {
	set conno 0
	set configdata($sourcecompnode,$sourceportname,conn) 1
    } elseif {$sourcetype == "Uses" && \
	    $irdata($sourcecompname,port,$sourceportname,multiplex) == 0} {
	set conno 0
	set configdata($sourcecompnode,$sourceportname,conn) 1
    } elseif {$sourcetype == "Emits"} {
	set conno 0
	set configdata($sourcecompnode,$sourceportname,conn) 1
    } else {
	set conno $configdata($sourcecompnode,$sourceportname,conn)
	incr configdata($sourcecompnode,$sourceportname,conn)
    }

    if {[info exists configdata($sourcecompnode,$sourceportname,$conno,type)] && \
	    $configdata($sourcecompnode,$sourceportname,$conno,type) == "Port"} {
	set oldcompnode [lindex $configdata($sourcecompnode,$sourceportname,$conno,sink) 0]
	incr compdata($oldcompnode,refs) -1
    }

    incr compdata($sinkcompnode,refs)
    set configdata($sourcecompnode,$sourceportname,kind) $sourcetype
    set configdata($sourcecompnode,$sourceportname,$conno,type) "Port"
    set configdata($sourcecompnode,$sourceportname,$conno,sink) \
	    [list $sinkcompnode $sinkportname]

    #
    # If the source is selected, update its widget
    #

    set currentselnode [lindex [$compwidgets(tree) selection get] 0]

    if {$sourcenode == $currentselnode} {
	CompSelectPort $sourcenode
    }
}

proc Ass::CompRemove {compnode} {
    variable compwidgets
    variable configdata
    variable homedata
    variable compdata
    variable status

    set compid $compdata($compnode,id)

    if {$compdata($compnode,refs) != 0} {
	set status "Error: cannot remove $compid."
	tk_messageBox -parent . -type ok -icon info -default ok \
		-title "Cannot remove component" \
		-message "$compid cannot be removed, because it is\
		still referenced by one or more connections."
	return
    }

    set status "Removing component $compid ..."
    update

    #
    # If removed node is selected, unselect
    #

    set currentsel [$compwidgets(tree) selection get]

    if {[llength $currentsel] != 0} {
	set currentselnode [lindex $currentsel 0]

	while {[$compwidgets(tree) parent $currentselnode] != "root"} {
	    set currentselnode [$compwidgets(tree) parent $currentselnode]
	}

	if {$compnode == $currentselnode} {
	    CompSelectComp none
	    CompSelectPort none
	}
    }

    #
    # Delete node and associated data
    #

    set homenode $compdata($compnode,home)
    incr homedata($homenode,refs) -1

    $compwidgets(tree) delete $compnode

    set idx [lsearch -exact $compdata(allcomponents) $compnode]
    set compdata(allcomponents) [lreplace $compdata(allcomponents) $idx $idx]

    catch {array unset compdata $compnode,*}
    catch {array unset configdata $compnode,*}

    set status "Component $compid removed."
}

#
# ======================================================================
# Show Attribute Information
# ======================================================================
#

proc Ass::AttributeMakeFrame {top widget id name type} {
    variable nodecounter
    variable configdata
    variable attrwidgets

    if {![info exists configdata($id,$name,set)]} {
	set configdata($id,$name,set) 0
    }

    set attrconfig [checkbutton $widget.config \
	    -text "Configure this attribute" \
	    -variable ::Ass::configdata($id,$name,set) \
	    -anchor w]
    $attrconfig configure -activebackground [$attrconfig cget -bg]
    pack $attrconfig -side top -fill x

    switch -- $type {
	boolean -
	short -
	long -
	float -
	double {
	    set defaultvalue 0
	    set typename $type
	}
	{unsigned short} {
	    set defaultvalue 0
	    set typename ushort
	}
	{unsigned long} {
	    set defaultvalue 0
	    set typename ulong
	}
	{long long} {
	    set defaultvalue 0
	    set typename longlong
	}
	{unsigned long long} {
	    set defaultvalue 0
	    set typename ulonglong
	}
	{long double} {
	    set defaultvalue 0
	    set typename longdouble
	}
	char -
	octet -
	wchar {
	    set defaultvalue "x"
	    set typename $type
	}
	string -
	wstring {
	    set defaultvalue ""
	    set typename $type
	}
    }

    switch -- $type {
	boolean -
	short -
	long -
	{unsigned short} -
	{unsigned long} -
	float -
	double -
	char -
	octet -
	string -
	{long long} -
	{unsigned long long} -
	{long double} -
	wchar -
	wstring {
	    set configdata($id,$name,type,$name) $typename

	    if {![info exists configdata($id,$name,value,$name)]} {
		set configdata($id,$name,value,$name) $defaultvalue
	    }

	    set attrtype [LabelEntry $widget.type -label "Type" \
		    -labelwidth 12 -width 20 -labelanchor w \
		    -editable 0 -relief flat -text $type]
	    $attrtype configure -entrybg [$attrtype cget -bg]
	    pack $attrtype -side top -fill x

	    set attrvalue [LabelEntry $widget.value -label "Value" \
		    -labelwidth 12 -width 20 -labelanchor w \
		    -textvariable ::Ass::configdata($id,$name,value,$name)]
	    pack $attrvalue -side top -fill x

	    return
	}
    }

    set tckind [lindex $type 0]

    if {$tckind != "sequence" && $tckind != "struct" && \
	    $tckind != "valuetype"} {
	set attrtype [LabelEntry $widget.type -label "Type" \
		-labelwidth 12 -width 20 -labelanchor w \
		-editable 0 -relief flat -text $tckind]
	$attrtype configure -entrybg [$attrtype cget -bg]
	pack $attrtype -side top -fill x
	$attrconfig configure -state disabled
	return
    }

    #
    # create tree widget to display complex data types
    #

    set attrtype [label $widget.type -text $tckind \
	    -width 20 -anchor w -relief flat]
    pack $attrtype -side top -fill x

    set attrsw [ScrolledWindow $widget.sw \
	    -relief sunken -borderwidth 2]
    set attrtree [Tree $attrsw.tree -height 4 \
	    -relief flat -borderwidth 0 -highlightthickness 0]
    $attrsw setwidget $attrtree

    $attrtree bindText <ButtonPress-1> "Ass::AttributeSelect $top $attrtree"
    $attrtree bindImage <ButtonPress-1> "Ass::AttributeSelect $top $attrtree"

    pack $attrsw -side top -expand yes -fill both

    set itemname [LabelEntry $widget.itemname -label "Name" \
	    -labelwidth 12 -width 20 -labelanchor w \
	    -state disabled -editable 0 -relief flat]
    $itemname configure -entrybg [$itemname cget -bg]
    pack $itemname -side top -fill x

    set itemtype [LabelEntry $widget.itemtype -label "Type" \
	    -labelwidth 12 -width 20 -labelanchor w \
	    -state disabled -editable 0 -relief flat]
    $itemtype configure -entrybg [$itemtype cget -bg]
    pack $itemtype -side top -fill x

    set itemvalue [LabelEntry $widget.itemvalue -label "Value" \
	    -labelwidth 12 -width 20 -labelanchor w \
	    -state disabled]
    pack $itemvalue -side top -fill x

    #
    # populate tree widget
    #

    AttributePopulateTree $attrtree root $id $name $name $type

    #
    # save widgets for later use
    #

    catch {array unset attrwidgets $top,*}
    set attrwidgets($top,tree) $attrtree
    set attrwidgets($top,name) $itemname
    set attrwidgets($top,type) $itemtype
    set attrwidgets($top,value) $itemvalue
}

proc Ass::AttributePopulateTree {tree node id attrname name type} {
    variable nodecounter
    variable configdata
    variable images

    switch -- $type {
	boolean -
	short -
	long -
	float -
	double {
	    set defaultvalue 0
	    set typename $type
	}
	{unsigned short} {
	    set defaultvalue 0
	    set typename ushort
	}
	{unsigned long} {
	    set defaultvalue 0
	    set typename ulong
	}
	{long long} {
	    set defaultvalue 0
	    set typename longlong
	}
	{unsigned long long} {
	    set defaultvalue 0
	    set typename ulonglong
	}
	{long double} {
	    set defaultvalue 0
	    set typename longdouble
	}
	char -
	octet -
	wchar {
	    set defaultvalue "x"
	    set typename $type
	}
	string -
	wstring {
	    set defaultvalue ""
	    set typename $type
	}
    }

    switch -- $type {
	boolean -
	short -
	long -
	{unsigned short} -
	{unsigned long} -
	float -
	double -
	char -
	octet -
	string -
	{long long} -
	{unsigned long long} -
	{long double} -
	wchar -
	wstring -
	Object {
	    set configdata($id,$attrname,type,$name) $typename

	    if {![info exists configdata($id,$attrname,value,$name)]} {
		set configdata($id,$attrname,value,$name) $defaultvalue
	    }

	    incr nodecounter
	    $tree insert end $node $nodecounter \
		    -text $name \
		    -data [list $id $attrname $name $type] \
		    -open 0 -drawcross never \
		    -image $images(attr)

	    return
	}
    }

    set tckind [lindex $type 0]

    if {$tckind == "sequence"} {
	set configdata($id,$attrname,type,$name) sequence

	if {![info exists configdata($id,$attrname,value,$name)]} {
	    set configdata($id,$attrname,value,$name) 0
	}

	set mynode [incr nodecounter]
	$tree insert end $node $mynode \
		-text $name \
		-data [list $id $attrname $name $type] \
		-open 0 -drawcross auto \
		-image [Bitmap::get folder]

	set length $configdata($id,$attrname,value,$name)
	set subtype [lindex $type 1]

	for {set i 0} {$i < $length} {incr i} {
	    set subname "$name\[$i\]"
	    AttributePopulateTree $tree $mynode $id $attrname \
		    $subname $subtype
	}
    } elseif {$tckind == "struct" || $tckind == "valuetype"} {
	set configdata($id,$attrname,type,$name) $tckind
	set configdata($id,$attrname,members,$name) [list]
	set members [lindex $type 2]

	if {$tckind == "valuetype"} {
	    set basetc [lindex $type 3]
	    while {$basetc != "" && $basetc != "0"} {
		foreach {member_name member_type} [lindex $basetc 2] {
		    set members [linsert $members 0 $member_name $member_type]
		}
		set basetc [lindex $basetc 3]
	    }
	}

	set mynode [incr nodecounter]
	$tree insert end $node $mynode \
		-text $name \
		-data [list $id $attrname $name $type] \
		-open 0 -drawcross auto \
		-image [Bitmap::get folder]

	foreach {member_name member_type} $members {
	    set subname "$name.$member_name"
	    lappend configdata($id,$attrname,members,$name) $member_name
	    AttributePopulateTree $tree $mynode $id $attrname \
		    $subname $member_type
	}
    }
}

proc Ass::AttributeSelect {top tree node} {
    variable configdata
    variable attrwidgets

    set currentselnode [lindex [$tree selection get] 0]

    if {$node == $currentselnode} {
	return
    }

    focus $tree
    $tree selection set $node
    $tree see $node

    set data [$tree itemcget $node -data]

    set id   [lindex $data 0]
    set attr [lindex $data 1]
    set name [lindex $data 2]
    set type [lindex $data 3]

    set tckind [lindex $type 0]

    $attrwidgets($top,name) configure \
	    -text $name \
	    -state normal

    $attrwidgets($top,type) configure \
	    -text $tckind \
	    -state normal

    switch -- $tckind {
	boolean -
	short -
	long -
	{unsigned short} -
	{unsigned long} -
	float -
	double -
	char -
	octet -
	string -
	{long long} -
	{unsigned long long} -
	{long double} -
	wchar -
	wstring -
	Object {
	    $attrwidgets($top,value) configure \
		    -textvariable ::Ass::configdata($id,$attr,value,$name) \
		    -state normal -command ""
	}
	sequence {
	    $attrwidgets($top,value) configure \
		    -textvariable ::Ass::configdata($id,$attr,value,$name) \
		    -state normal \
		    -command "Ass::AttributeAdjustSequence $top $tree $node"
	}
	array -
	struct -
	valuetype {
	    $attrwidgets($top,value) configure \
		    -textvariable "" -text "" \
		    -state disabled -command ""
	    return
	}
    }
}

proc Ass::AttributeAdjustSequence {top tree node} {
    variable nodecounter
    variable configdata
    variable attrwidgets

    set data [$tree itemcget $node -data]

    set id   [lindex $data 0]
    set attr [lindex $data 1]
    set name [lindex $data 2]
    set type [lindex $data 3]

    set newlength $configdata($id,$attr,value,$name)
    set curlength [llength [$tree nodes $node]]

    if {$curlength == $newlength} {
	return
    }

    if {$newlength < $curlength} {
	$tree delete [$tree nodes $node $newlength end]
    } else {
	set subtype [lindex $type 1]

	for {set i $curlength} {$i < $newlength} {incr i} {
	    set subname "$name\[$i\]"

	    AttributePopulateTree $tree $node $id $attr \
		    $subname $subtype
	}
    }
}

#
# ======================================================================
# Handle Facet and Consumes Ports
# ======================================================================
#

proc Ass::SinkPortMakeFrame {top widget id name porttype conntype} {
    variable configdata

    if {![info exists configdata($id,$name,nsreg)]} {
	set configdata($id,$name,nsreg) 0
	set configdata($id,$name,nsname) ""
    }

    catch {array unset attrwidgets $top,*}

    set sourcetype [LabelEntry $widget.type -label "Type" \
	    -labelwidth 12 -width 20 -labelanchor w \
	    -editable 0 -relief flat -text $conntype]
    $sourcetype configure -entrybg [$sourcetype cget -bg]
    pack $sourcetype -side top -fill x -pady 2

    set sourcensreg [checkbutton $widget.nsreg \
	    -text "Register with Naming Service" \
	    -variable ::Ass::configdata($id,$name,nsreg) \
	    -anchor w -highlightthickness 0]
    $sourcensreg configure -activebackground [$sourcensreg cget -bg]
    pack $sourcensreg -side top -fill x -pady 2

    set sourcensname [LabelEntry $widget.nsname -label "NS Name" \
	    -labelwidth 12 -width 40 -labelanchor w \
	    -textvariable ::Ass::configdata($id,$name,nsname) \
	    -helptext "The Name for this connection in the Naming Service"]
    pack $sourcensname -side top -fill x -pady 2
}

#
# ======================================================================
# Handle Simplex Receptacles and Emitters
# ======================================================================
#

proc Ass::SimplexSourceMakeFrame {top widget id name porttype conntype} {
    variable attrwidgets
    variable configdata
    variable irdata

    if {![info exists configdata($id,$name,conn)]} {
	set configdata($id,$name,conn) 0
	set configdata($id,$name,0,type) "none"
	set configdata($id,$name,0,sink) ""
    }

    set receptype [LabelEntry $widget.type -label "Type" \
	    -labelwidth 12 -width 20 -labelanchor w \
	    -editable 0 -relief flat -text $conntype]
    $receptype configure -entrybg [$receptype cget -bg]
    pack $receptype -side top -fill x -pady 2

    set recepconn [checkbutton $widget.nsreg \
	    -text "Configure Connection" \
	    -variable ::Ass::configdata($id,$name,conn) \
	    -anchor w -highlightthickness 0]
    $recepconn configure -activebackground [$recepconn cget -bg]
    pack $recepconn -side top -fill x -pady 2

    set connframe [frame $widget.conn0]

    set connlabel [label $connframe.lab -text "Connect To" \
	    -width 12 -anchor w]
    pack $connlabel -side left

    set connbut [menubutton $connframe.but -text "Select" -width 12 \
	    -menu $connframe.but.m -direction below -relief raised]
    pack $connbut -side left

    set connent [Entry $connframe.ent -width 10]
    pack $connent -side left -expand yes -fill x

    pack $connframe -side top -fill x -pady 2

    catch {array unset attrwidgets $top,*}
    set attrwidgets($top,0,button) $connbut
    set attrwidgets($top,0,menu) $connbut.m
    set attrwidgets($top,0,entry) $connent

    SourcePortBuildSelectMenu $top $id $name 0 $connbut.m \
	    $porttype $conntype
    SourcePortSelectConfig $top $id $name 0 \
	    $configdata($id,$name,0,type) \
	    $configdata($id,$name,0,sink)
}

#
# ======================================================================
# Handle Multiplex Receptacles and Event Sources
# ======================================================================
#

proc Ass::SourcePortMakeFrame {top widget id name porttype conntype} {
    variable attrwidgets
    variable configdata

    if {![info exists configdata($id,$name,conn)]} {
	set configdata($id,$name,conn) 0
    }

    set configdata($id,$name,kind) $porttype

    set sinktype [LabelEntry $widget.type -label "Type" \
	    -labelwidth 12 -width 20 -labelanchor w \
	    -editable 0 -relief flat -text $conntype]
    $sinktype configure -entrybg [$sinktype cget -bg]
    pack $sinktype -side top -fill x -pady 2

    set numconnframe [frame $widget.numconn]

    set numconnlabel [label $numconnframe.lab -text "Connections" \
	    -width 12 -anchor w]
    pack $numconnlabel -side left

    set numconnbut1 [button $numconnframe.but1 -text "+" -width 5 \
	    -command "::Ass::SourcePortAddConn $top $id $name $porttype $conntype 1" \
	    -relief raised]
    pack $numconnbut1 -side left -padx 3

    set numconnbut2 [button $numconnframe.but2 -text "-" -width 5 \
	    -command "::Ass::SourcePortAddConn $top $id $name $porttype $conntype -1" \
	    -relief raised]
    pack $numconnbut2 -side left -padx 3

    set numconnent [Entry $numconnframe.ent -width 10 \
	    -command "::Ass::SourcePortUpdateConns $top $id $name $porttype $conntype" \
	    -textvariable ::Ass::configdata($id,$name,conn)]
    pack $numconnent -side left -expand yes -fill x
    bind $numconnent <FocusOut> "Ass::SourcePortUpdateConns $top $id $name $porttype $conntype"

    pack $numconnframe -side top -fill x -pady 2

    set connssw [ScrolledWindow $widget.conns \
	    -relief sunken -borderwidth 2]
    set connssf [ScrollableFrame $connssw.sf -constrainedwidth 1]
    set connsframe [$connssf getframe]
    $connssw setwidget $connssf

    pack $connssw -side top -fill both -expand yes

    catch {array unset attrwidgets $top,*}
    set attrwidgets($top,frames) 0
    set attrwidgets($top,numconns) $numconnent
    set attrwidgets($top,conns) $connsframe

    SourcePortUpdateConns $top $id $name $porttype $conntype
}

proc Ass::SourcePortUpdateConns {top id name porttype conntype} {
    variable attrwidgets
    variable configdata
    variable compdata

    set numconns $configdata($id,$name,conn)

    if {$numconns == $attrwidgets($top,frames)} {
	return
    }

    if {$numconns < $attrwidgets($top,frames)} {
	for {set i $numconns} {$i < $attrwidgets($top,frames)} {incr i} {
	    if {$configdata($id,$name,$i,type) == "Port"} {
		set compnode [lindex $configdata($id,$name,$i,sink) 0]
		incr compdata($compnode,refs) -1
	    }
	    catch {array unset configdata $id,$name,$i,*}
	    catch {destroy $attrwidgets($top,$i,frame)}
	}
	set attrwidgets($top,frames) $numconns
	return
    }

    set connsframe $attrwidgets($top,conns)

    for {set i $attrwidgets($top,frames)} {$i < $numconns} {incr i} {
	set connframe [frame $connsframe.c$i]

	set connlabel [label $connframe.lab -text "Conn # $i" \
		-width 12 -anchor w]
	pack $connlabel -side left

	set connbut [menubutton $connframe.but -text "Select" -width 12 \
		-menu $connframe.but.m -direction below -relief raised]
	pack $connbut -side left

	set connent [Entry $connframe.ent -width 20]
	pack $connent -side left -expand yes -fill x -expand yes

	pack $connframe -side top -fill x -pady 1

	set attrwidgets($top,$i,frame) $connframe
	set attrwidgets($top,$i,button) $connbut
	set attrwidgets($top,$i,menu) $connbut.m
	set attrwidgets($top,$i,entry) $connent

	if {![info exists ::Ass::configdata($id,$name,$i,type)]} {
	    set configdata($id,$name,$i,type) "none"
	    set configdata($id,$name,$i,sink) ""
	}

	SourcePortBuildSelectMenu $top $id $name $i $connbut.m \
		$porttype $conntype

	SourcePortSelectConfig $top $id $name $i \
		$configdata($id,$name,$i,type) \
		$configdata($id,$name,$i,sink)
    }

    set attrwidgets($top,frames) $numconns
}

proc Ass::SourcePortAddConn {top id name porttype conntype incr} {
    variable configdata

    if {[incr configdata($id,$name,conn) $incr] < 0} {
	set configdata($id,$name,conn) 0
    }

    SourcePortUpdateConns $top $id $name $porttype $conntype
}

proc Ass::SourcePortBuildSelectMenu {top id name item menu porttype conntype} {
    variable compwidgets
    variable compdata
    variable irdata

    menu $menu
    $menu add command -label "Object URL" \
	    -command "::Ass::SourcePortSelectConfig $top $id $name $item URL {}"
    $menu add command -label "Naming Service Entry" \
	    -command "::Ass::SourcePortSelectConfig $top $id $name $item NS {}"
    $menu add separator

    #
    # Look for matching sink ports
    #

    set count 0

    foreach compnode $compdata(allcomponents) {
	CompIntrospectComponent $compnode

	set ctype $compdata($compnode,nodetype)
	set cname $compdata($compnode,name)
	set compid $compdata($compnode,id)

	if {$ctype != "Component"} {
	    continue
	}

	if {$porttype == "Uses" && \
		[IntrospectIsA $irdata($cname,id) $conntype]} {
	    incr count
	    $menu add command -label $compid \
		    -command "Ass::SourcePortSelectConfig $top $id $name $item Port {$compnode -1}"
	}

	foreach portnode $compdata($compnode,children) {
	    set sinkporttype $compdata($portnode,nodetype)
	    
	    if {$porttype == "Uses" && $sinkporttype != "Provides"} {
		continue
	    } elseif {($porttype == "Emits" || $porttype == "Publishes") && \
		    $sinkporttype != "Consumes"} {
		continue
	    }

	    set sinkname $compdata($portnode,name)
	    set sinkconntype $irdata($cname,port,$sinkname,type)

	    if {$porttype == "Uses" && \
		    ![IntrospectIsA $sinkconntype $conntype]} {
		continue
	    } elseif {($porttype == "Emits" || \
		    $porttype == "Publishes") && \
		    ![IntrospectIsA $conntype $sinkconntype]} {
		continue
	    }

	    incr count
	    set label "$sinkname of $compid"
		
	    $menu add command -label $label \
		    -command "Ass::SourcePortSelectConfig $top $id $name $item Port {$compnode $sinkname}"
	}
    }

    if {$count} {
	$menu add separator
    }

    $menu add command -label "Remove" \
	    -command "Ass::SourcePortRemove $top $id $name $item"
}

proc Ass::SourcePortSelectConfig {top id name item conntype conndata} {
    variable compwidgets
    variable attrwidgets
    variable configdata
    variable compdata

    set oldtype $configdata($id,$name,$item,type)

    if {$oldtype == "Port"} {
	set compnode [lindex $configdata($id,$name,$item,sink) 0]
	incr compdata($compnode,refs) -1
    }

    set configdata($id,$name,$item,type) $conntype

    switch -- $conntype {
	none {
	    set configdata($id,$name,$item,sink) ""
	    $attrwidgets($top,$item,button) configure -text "Select"
	    $attrwidgets($top,$item,entry) configure \
		    -state disabled -editable 0 \
		    -textvariable ""
	}
	URL {
	    if {$oldtype == "Port"} {
		set configdata($id,$name,$item,sink) ""
	    }

	    $attrwidgets($top,$item,button) configure -text "Object URL"
	    $attrwidgets($top,$item,entry) configure \
		    -state normal -editable 1 \
		    -textvariable ::Ass::configdata($id,$name,$item,sink)
	}
	NS {
	    if {$oldtype == "Port"} {
		set configdata($id,$name,$item,sink) ""
	    }

	    $attrwidgets($top,$item,button) configure -text "Naming Entry"
	    $attrwidgets($top,$item,entry) configure \
		    -state normal -editable 1 \
		    -textvariable ::Ass::configdata($id,$name,$item,sink)
	}
	Port {
	    set compnode [lindex $conndata 0]
	    set portname [lindex $conndata 1]
	    set compid $compdata($compnode,id)

	    if {$portname == -1} {
		set label $compid
	    } else {
		set label "$compid.$portname"
	    }

	    $attrwidgets($top,$item,button) configure -text "Port"
	    $attrwidgets($top,$item,entry) configure \
		    -state normal -editable 0 \
		    -textvariable ""
	    $attrwidgets($top,$item,entry) configure -text $label

	    incr compdata($compnode,refs)
	    set configdata($id,$name,$item,sink) [list $compnode $portname]
	}
	default {
	    error "oops"
	}
    }
}

proc Ass::SourcePortRemove {top id name item} {
    variable compwidgets
    variable attrwidgets
    variable configdata
    variable compdata

    if {$configdata($id,$name,conn) == 0} {
	return
    }

    if {![info exists attrwidgets($top,frames)]} {
	#
	# simplex receptacle or emits port
	#
	set configdata($id,$name,conn) 0
	SourcePortSelectConfig $top $id $name 0 "none" ""
	return
    }

    set numconns [incr attrwidgets($top,frames) -1]
    
    for {set i $item} {$i < $numconns} {incr i} {
	set j [expr {$i + 1}]
	SourcePortSelectConfig $top $id $name $i \
		$configdata($id,$name,$j,type) \
		$configdata($id,$name,$j,sink)
    }

    SourcePortSelectConfig $top $id $name $numconns "none" ""
    catch {array unset configdata $id,$name,$numconns,*}
    catch {destroy $attrwidgets($top,$numconns,frame)}
    set attrwidgets($top,frames) $numconns
    set configdata($id,$name,conn) $numconns
}

#
# ======================================================================
# File -> New
# ======================================================================
#

proc Ass::FileNew {} {
    variable packwidgets
    variable compwidgets
    variable homewidgets
    variable configdata
    variable packdata
    variable homedata
    variable compdata
    variable status

    set status "Clearing Assembly ..."

    PackSelectPackage "none"
    
    HomeSelectHome "none"
    HomeSelectAttribute "none"

    CompSelectComp "none"
    CompSelectPort "none"

    $packwidgets(tree) delete [$packwidgets(tree) nodes root]
    $compwidgets(tree) delete [$compwidgets(tree) nodes root]
    $homewidgets(tree) delete [$homewidgets(tree) nodes root]

    catch {unset packdata}
    catch {unset homedata}
    catch {unset compdata}
    catch {unset configdata}

    set packdata(allpackages) [list]
    set homedata(allhomes) [list]
    set compdata(allcomponents) [list]

    #
    # Remove temp files
    #

    CleanupTempFiles

    #
    # Keep IR data around
    #

    set status "Assembly cleared."
}

#
# ======================================================================
# File -> Open Assembly
# ======================================================================
#

proc Ass::FileOpenAssembly {} {
    variable initialdir
    variable status

    set status "Please select a Component Assembly file"

    set olddir $initialdir

    set filetypes {
	{"Assembly Archive" {.aar}}
	{"Assembly Descriptor" {.cad}}
	{"All Files" {*}}
    }

    set filename [tk_getOpenFile -title "Open Assembly ..." \
	    -parent . -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	set status "Cancelled."
	return
    }

    set initialdir [file dirname $filename]

    if {![file exists $filename]} {
	set status "Error: $filename does not exist."
	return
    }

    #
    # Clear Assembly
    #

    FileNew

    #
    # Is it an Component Assembly File (.aar)? Then unzip to a temp
    # directory.
    #

    if {[catch {FileReadAssembly $filename} err]} {
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error downloading Assembly" \
		-message $err
	return
    }

    #
    # Update Display
    #

    PackDisplayAllPackages
    HomeDisplayAllHomes
    CompDisplayAllComponents

    set status "Done."
}

#
# ======================================================================
# File -> Download Assembly
# ======================================================================
#

proc Ass::FileDownloadAssembly {} {
    variable PackLoadAssemblyURL
    variable initialdir
    variable status
    variable mutex

    set top [toplevel .packDownloadAssembly]
    wm title $top "Open URL"
    set toplab [label $top.lab -width 50 -anchor center \
	    -text "Please enter the URL of an Assembly Archive"]
    pack $toplab -side top -fill x -pady 10

    set sep1 [Separator $top.sep1 -orient horizontal]
    pack $sep1 -side top -fill x -pady 10

    set ptopf1 [frame $top.f1]
    set ptopf1lab [label $ptopf1.lab -text "Archive URL" \
	    -width 15 -anchor w]
    set ptopf1ent [entry $ptopf1.ent -width 40 \
	    -textvariable Ass::PackLoadAssemblyURL]
    pack $ptopf1lab -side left
    pack $ptopf1ent -side left -expand yes -fill x
    pack $ptopf1 -side top -fill x -padx 10

    set sep2 [Separator $top.sep2 -orient horizontal]
    pack $sep2 -side top -fill x -pady 10

    set ptopf3 [frame $top.f3]
    set ptopf3but1 [button $ptopf3.but1 -width 15 \
	    -text "Download" \
	    -command "set ::Ass::mutex 1"]
    set ptopf3but2 [button $ptopf3.but2 -width 15 \
	    -text "Cancel" \
	    -command "set ::Ass::mutex 2"]
    pack $ptopf3but1 $ptopf3but2 -side left -pady 10 -padx 20
    pack $ptopf3

    bind $ptopf1ent <Return> "focus $ptopf3but1"
    bind $ptopf3but1 <Return> "set ::Ass::mutex 1"
    bind $ptopf3but2 <Return> "set ::Ass::mutex 2"

    set status "Please enter URL ..."
    wm deiconify $top
    focus $top
    grab set $top
    set mutex 0
    vwait ::Ass::mutex
    grab release $top
    destroy $top

    if {$mutex != 1} {
	set status "Download cancelled."
	return
    }

    #
    # Clear Assembly
    #

    FileNew

    #
    # Download
    #

    set status "Downloading assembly $PackLoadAssemblyURL ..."
    update

    if {[catch {FileReadAssembly $PackLoadAssemblyURL} err]} {
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error downloading Assembly" \
		-message $err
	return
    }

    #
    # Update Display
    #

    PackDisplayAllPackages
    HomeDisplayAllHomes
    CompDisplayAllComponents

    set status "Done."
}

#
# ======================================================================
# File -> Save Assembly
# ======================================================================
#

proc Ass::FileSaveAssembly {} {
    variable initialdir
    variable status

    set status "Please select a filename for the Component Assembly Descriptor"

    #set olddir $initialdir
    set olddir [pwd]

    set filetypes {
	{"Assembly Descriptor" {.cad}}
	{"All Files" {*}}
    }

    set filename [tk_getSaveFile -title "Save As ..." \
	    -parent . -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	set status "Cancelled."
	return
    }

    if {[file extension [file tail $filename]] == ""} {
	append filename ".cad"
    }

    set initialdir [file dirname $filename]
    set localname [file tail $filename]

    if {[catch {FileDumpAssembly $localname $initialdir 0} err]} {
	set status "Error writing assembly."
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error writing Assembly" \
		-message $err
	return
    }

    set status "Done writing Assembly Descriptor $localname"
}

#
# ======================================================================
# Build Package
# ======================================================================
#

proc Ass::FileBuildPackage {} {
    variable initialdir
    variable homedata
    variable packdata
    variable status
    variable setup

    set status "Please select a filename for the Component Assembly Archive"

    #set olddir $initialdir
    set olddir [pwd]

    set filetypes {
	{"Assembly Archive" {.aar}}
	{"All Files" {*}}
    }

    set filename [tk_getSaveFile -title "Save As ..." \
	    -parent . -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	set status "Cancelled."
	return
    }

    if {[file extension [file tail $filename]] == ""} {
	append filename ".aar"
    }

    set initialdir [file dirname $filename]
    set localname [file tail $filename]
    set rootname [file rootname $localname]

    #
    # make a new directory for the archive's contents
    #

    set aardir [MkTempDir]
    set metadir [file join $aardir meta-inf]

    if {[catch {file mkdir $metadir}]} {
	set status "Cannot create $metadir directory."
	return
    }

    #
    # Browse packages, and copy IDL files and DLL files to this directory
    #

    foreach homenode $homedata(allhomes) {
	set packnode $homedata($homenode,package)
	set packid $packdata($packnode,id)

	if {[info exists packaged($packid)]} {
	    continue
	}

	set packaged($packid) 1

	if {$packdata($packnode,whatref) != "noref"} {
	    continue
	}

	set packliburl $packdata($packnode,liburl)
	set packidlurl $packdata($packnode,idlurl)

	set lastidx [string last / $packidlurl]
	incr lastidx
	set localidlname [string range $packidlurl $lastidx end]

	if {![info exists copied($localidlname)]} {
	    set status "Copying $localidlname into Package ..."
	    update

	    set idllocalfile [file join $aardir $localidlname]
	    if {[catch {DownloadFile $packidlurl $idllocalfile} err]} {
		set status "Error: cannot copy $localidlname to $aardir: $err"
		catch {file delete -force $aardir}
		return
	    }

	    set copied($localidlname) 1
	}

	set lastidx [string last / $packliburl]
	incr lastidx
	set locallibname [string range $packliburl $lastidx end]

	if {![info exists copied($locallibname)]} {
	    set status "Copying $locallibname into Package ..."
	    update

	    set liblocalfile [file join $aardir $locallibname]
	    if {[catch {DownloadFile $packliburl $liblocalfile} err]} {
		set status "Error: cannot copy $locallibname to $aardir: $err"
		catch {file delete -force $aardir}
		return
	    }

	    set copied($locallibname) 1
	}
    }
	    
    #
    # Write meta-inf
    #

    set cadname $rootname
    append cadname ".cad"

    if {[catch {FileDumpAssembly $cadname $metadir 1} err]} {
	set status "Error writing assembly."
	tk_messageBox -parent . \
		-type ok -icon error -default ok \
		-title "Error writing Assembly" \
		-message $err
	catch {file delete -force $aardir}
	return
    }

    #
    # Zip it
    #

    set status "Compressing Package $localname ..."
    update

    catch {file delete -force $filename}

    set oldpwd [pwd]
    cd $aardir
    if {[catch {tk_exec $setup(zip) -q -r $filename .} err]} {
	set status "Error: zipping error: $err"
	catch {file delete -force $aardir}
	cd $oldpwd
	return
    }
    cd $oldpwd

    #
    # Remove temp directory
    #

    catch {file delete -force $aardir}

    #
    # Okay
    #

    set status "Done writing Assembly Archive $localname"
}

#
# ======================================================================
# Deploy Assembly
# ======================================================================
#

proc Ass::FileDeployStatusDisplay {name1 name2 op} {
    if {[winfo exists .traceDeployment.sw.text]} {
	.traceDeployment.sw.text insert end $::Ass::status
	.traceDeployment.sw.text insert end "\n"
	.traceDeployment.sw.text see end
	update
    }
}

proc Ass::FileDeployAssembly {} {
    variable status
    variable setup
    variable mutex

    set top [toplevel .deployAssembly]
    wm title $top "Deploy Assembly"

    set title [label $top.title -anchor center \
	    -font {Arial 18 bold} \
	    -text "Deploy Assembly"]
    pack $title -side top -fill x -pady 10

    set topsep [Separator $top.topsep -orient horizontal]
    pack $topsep -side top -fill x -pady 10

    if {![info exists setup(ccmd)]} {
	set setup(ccmd) [info hostname]
	append setup(ccmd) ":1234"
    }

    if {![info exists setup(localaddr)]} {
	if {[info exists setup(host)]} {
	    if {$setup(host) == "<auto>" || $setup(host) == ""} {
		set setup(localaddr) [info hostname]
	    } else {
		set setup(localaddr) $setup(host)
	    }
	} else {
	    set setup(localaddr) [info hostname]
	}
	if {[info exists setup(port)]} {
	    append setup(localaddr) ":" $setup(port)
	} else {
	    append setup(localaddr) ":1235"
	}
    }

    if {![info exists setup(nsior)]} {
	set setup(nsior) ""
    }

    set ccmd [LabelEntry $top.ccmd -label "MicoCCMD Address" \
	    -labelwidth 20 -width 40 -labelanchor w \
	    -textvariable ::Ass::setup(ccmd)]
    set localaddr [LabelEntry $top.localaddr -label "Local Address" \
	    -labelwidth 20 -width 40 -labelanchor w \
	    -textvariable ::Ass::setup(localaddr)]
    set nsior [LabelEntry $top.nsior -label "Naming Service" \
	    -labelwidth 20 -width 40 -labelanchor w \
	    -textvariable ::Ass::setup(nsior)]
    pack $ccmd $localaddr $nsior \
	    -side top -fill x -pady 5 -padx 10

    set botsep [Separator $top.botsep -orient horizontal]
    pack $botsep -side top -fill x -pady 10

    set ptopf3 [frame $top.f3]
    set ptopf3but1 [button $ptopf3.but1 -width 15 \
	    -text "Deploy" \
	    -command "set ::Ass::mutex 1"]
    set ptopf3but2 [button $ptopf3.but2 -width 15 \
	    -text "Cancel" \
	    -command "set ::Ass::mutex 2"]
    pack $ptopf3but1 $ptopf3but2 -side left -pady 10 -padx 20
    pack $ptopf3

    set status "Getting deploy information ..."

    wm deiconify $top
    focus $top
    grab set $top
    set mutex 0
    vwait ::Ass::mutex
    grab release $top
    destroy $top

    if {$mutex != 1} {
	set status "Deployment cancelled."
	return
    }

    #
    # Dissect the local address
    #

    set colon [string first : $setup(localaddr)]
    if {$colon == -1} {
	set setup(host) $setup(localaddr)
	set setup(port) 1235
    } else {
	set setup(host) [string range $setup(localaddr) 0 [expr {$colon - 1}]]
	set setup(port) [string range $setup(localaddr) [expr {$colon + 1}] end]
    }

    if {$setup(host) == [info hostname] || $setup(host) == ""} {
	set setup(host) "<auto>"
    }

    #
    # verify MicoCCMD Address
    #

    set status "Contacting MicoCCMD ..."
    set colon [string first : $setup(ccmd)]
    set atype [string range $setup(ccmd) 0 $colon]

    if {$atype != "corbaloc:" && $atype != "file:" && $atype != "http:" && \
	    $atype != "IOR:"} {
	set ccmdhost $setup(ccmd)
	if {$colon == -1} {
	    append ccmdhost ":1234"
	}
	set ccmdior "corbaloc::$ccmdhost/MicoCCMD"
    } else {
	set ccmdior $setup(ccmd)
    }

    corba::try {
	set ccmd [corba::string_to_object $ccmdior]
    } catch {... err} {
	set status "Illegal MicoCCM Daemon IOR."
	tk_messageBox -parent . -type ok -icon info -default ok \
		-title "Illegal MicoCCMD IOR" \
		-message "The address of the MicoCCM Daemon seems to\
		be illegal; string_to_object reports: $err"
	return
    }

    corba::try {
	set ia1 [$ccmd _is_a IDL:omg.org/Components/ComponentInstallation:1.0]
	set ia2 [$ccmd _is_a IDL:omg.org/Components/ServerActivator:1.0]
	set ia3 [$ccmd _is_a IDL:mico.org/MICOCCM/MicoCCMD:1.0]
    } catch {... err} {
	set status "Cannot talk to MicoCCM Daemon."
	tk_messageBox -parent . -type ok -icon error -default ok \
		-title "Cannot talk to MicoCCMD" \
		-message "The address of the MicoCCM Daemon seems to\
		be wrong: communication with that address failed."
	return
    }

    if {!$ia1 || !$ia2} {
	set status "Not a MicoCCM Daemon."
	tk_messageBox -parent . -type ok -icon error -default ok \
		-title "Cannot talk to MicoCCMD" \
		-message "The address of the MicoCCM Daemon seems to\
		be wrong: there's something at that address that speaks\
		CORBA, but it does not seem to be a MicoCCM Daemon."
	return
    }

    #
    # If there's something in the Naming Service field, test it
    #

    if {$setup(nsior) != ""} {
	set status "Contacting Naming Service ..."
	set colon [string first : $setup(nsior)]
	set atype [string range $setup(nsior) 0 $colon]

	if {$atype != "corbaloc:" && $atype != "file:" && \
		$atype != "http:" && $atype != "IOR:"} {
	    set nshost $setup(nsior)
	    set nsior "corbaloc::$nshost/NameService"
	} else {
	    set nsior $setup(nsior)
	}

	corba::try {
	    set nsref [corba::string_to_object $nsior]
	    set isns1 [$nsref _is_a IDL:omg.org/CosNaming/NamingContext:1.0]
	    set isns2 [$nsref _is_a IDL:omg.org/CosNaming/NamingContextExt:1.0]
	    if {!$isns1 || !$isns2} {
		error "not a Naming Service"
	    }
	} catch {... err} {
	    set status "Illegal Naming Service IOR."
	    tk_messageBox -parent . -type ok -icon error -default ok \
		    -title "Illegal Naming Service IOR" \
		    -message "Either the address of the Naming Service\
		    is illegal, or it points to something that is not a\
		    Naming Service: $err"
	    return
	}

	corba::register_initial_reference NameService $nsref
    }

    set status "Starting Deployment ..."

    #
    # Make a text window and log status in it
    #

    set tracetop .traceDeployment

    if {![winfo exists $tracetop]} {
    	toplevel $tracetop
	wm title $tracetop "Deployment Log"
	set sw [ScrolledWindow $tracetop.sw -auto horizontal]
	set text [text $sw.text -relief sunken -bd 2]
	$sw setwidget $text
	pack $sw -side top -fill both -expand yes

	set botsep [Separator $tracetop.botsep -orient horizontal]
	pack $botsep -side top -fill x -pady 10

	set botframe [frame $tracetop.botframe]
	set botbut [button $botframe.but -width 10 -text "Close" \
		-command "destroy $tracetop"]
	pack $botbut
	pack $botframe -side top -fill x -pady 10

	bind $tracetop <Prior> "$text yview scroll -1 pages"
	bind $tracetop <Next> "$text yview scroll 1 pages"
	bind $tracetop <Up> "$text yview scroll -1 units"
	bind $tracetop <Down> "$text yview scroll 1 units"
	bind $tracetop <Home> "$text yview moveto 0"
	bind $tracetop <End> "$text yview moveto 1"
	bind $tracetop <Return> "destroy $tracetop"
    } else {
	.traceDeployment.sw.text see end
	.traceDeployment.sw.text insert end "\n--- new log starts ---\n\n"
    }

    wm deiconify .traceDeployment
    focus .traceDeployment

    .traceDeployment.sw.text insert end "Starting deployment ...\n"
    trace variable ::Ass::status w ::Ass::FileDeployStatusDisplay

    corba::try {
	DeployAssembly $ccmd 0 1
	set status "Deployment finished."
    } catch {... err} {
	global errorInfo
	set status "Error: $err"
	.traceDeployment.sw.text insert end $errorInfo
	tk_messageBox -parent . -type ok -icon error -default ok \
		-title "Deployment Error" \
		-message "The following error occured during\
		deployment. Please see the log for a complete\
		problem report: $err"
    }

    trace vdelete ::Ass::status w ::Ass::FileDeployStatusDisplay
    .traceDeployment.sw.text insert end "Deployment log ends.\n"
}

#
# ======================================================================
# Save Assembly
# ======================================================================
#

proc Ass::FileDumpAssembly {filename metadir mkpackage} {
    variable configdata
    variable homedata
    variable compdata
    variable packdata
    variable status

    #
    # (1) Make softpkg and property files for each home
    #

    set allsoftpkgs [list]

    foreach homenode $homedata(allhomes) {
	update ;# strategic update

	set compinstances($homenode) [list]
	set packnode $homedata($homenode,package)
	set packid $packdata($packnode,id)

	if {![info exists packaged($packid)]} {
	    set packaged($packid) 1
	    lappend allsoftpkgs $packnode

	    if {$packdata($packnode,whatref) == "noref" || \
		    $packdata($packnode,whatref) == "reffiles"} {
		set softpkgfilename $packid
		append softpkgfilename ".csd"
		set fullname [file join $metadir $softpkgfilename]
		FileDumpSoftpkg $packnode $fullname $mkpackage
	    }
	}

	HomeIntrospectHome $homenode

	#
	# does this home have any configurable properties?
	#

	set homeid $homedata($homenode,id)
	set homedata($homeid,cpf) [FileHasProperties homedata $homenode]

	if {$homedata($homeid,cpf)} {
	    set propertyfilename $homeid
	    append propertyfilename ".cpf"
	    set fullname [file join $metadir $propertyfilename]
	    FileDumpProperties homedata $homenode $fullname
	}
    }

    #
    # (2) Go through components; find their homes and make property files
    #

    foreach compnode $compdata(allcomponents) {
	update ;# strategic update
	set compid $compdata($compnode,id)
	set homenode $compdata($compnode,home)
	lappend compinstances($homenode) $compnode

	CompIntrospectComponent $compnode

	set compdata($compid,cpf) [FileHasProperties \
		compdata $compnode]

	if {$compdata($compid,cpf)} {
	    set propertyfilename $compid
	    append propertyfilename ".cpf"
	    set fullname [file join $metadir $propertyfilename]
	    FileDumpProperties compdata $compnode $fullname
	}
    }

    #
    # (3) Build Assembly data
    #

    set fileroot [file rootname [file tail $filename]]
    set status "Writing Assembly Descriptor $filename ..."
    update

    set doc [::dom::DOMImplementation create]
    set assembly [::dom::document createElement $doc "componentassembly"]
    ::dom::element setAttribute $assembly id $fileroot

    set compfiles [::dom::document createElement $assembly "componentfiles"]
    set partitioning [::dom::document createElement $assembly "partitioning"]
    set connections [::dom::document createElement $assembly "connections"]

    foreach packnode $allsoftpkgs {
	update ;# strategic update
	set packid $packdata($packnode,id)

	set compfile [::dom::document createElement $compfiles "componentfile"]
	::dom::element setAttribute $compfile id $packid

	switch -- $packdata($packnode,whatref) {
	    noref -
	    reffiles {
		set softpkgfilename $packid
		append softpkgfilename ".csd"
		set file [::dom::document createElement \
			$compfile "fileinarchive"]
		::dom::element setAttribute $file name $softpkgfilename
	    }
	    refpkg {
		set softpkgurl $packdata($packnode,pkgurl)
		if {[file exists $softpkgurl]} {
		    if {[file pathtype $softpkgurl] == "relative"} {
			set pkgurl "file://[pwd]/$softpkgurl"
		    } else {
			set pkgurl "file://$softpkgurl"
		    }
		} else {
		    set pkgurl $softpkgurl
		}
		set link [::dom::document createElement $compfile "link"]
		::dom::element setAttribute $link href $pkgurl
	    }
	}
    }

    foreach homenode $homedata(allhomes) {
	update ;# strategic update
	set homeplacement [FileDumpHome $homenode \
		$partitioning $connections]

	foreach compnode $compinstances($homenode) {
	    FileDumpComponent $compnode $homeplacement $connections
	    update ;# strategic update
	}
    }

    set assembly [::dom::pretty_print $doc]
    ::dom::DOMImplementation destroy $doc

    set fullname [file join $metadir $filename]
    set file [open $fullname w]
    puts $file $assembly
    close $file

    set status "Done."
}

proc Ass::FileDumpSoftpkg {packnode filename mkpackage} {
    variable packdata
    variable status

    set localname [file tail $filename]
    set status "Writing Software Package Descriptor $localname ..."
    update

    set packid $packdata($packnode,id)
    set homerepoid $packdata($packnode,repoid)
    set homeentrypt $packdata($packnode,entrypt)
    set packliburl $packdata($packnode,liburl)
    set packidlurl $packdata($packnode,idlurl)
    set whatref $packdata($packnode,whatref)
    set codetype $packdata($packnode,codetype)

    set doc [::dom::DOMImplementation create]
    set softpkg [::dom::document createElement $doc "softpkg"]
    ::dom::element setAttribute $softpkg name $packid

    set idl [::dom::document createElement $softpkg "idl"]
    ::dom::element setAttribute $idl id $homerepoid

    if {$whatref == "reffiles" || !$mkpackage} {
	if {[file exists $packidlurl]} {
	    if {[file pathtype $packidlurl] == "relative"} {
		set idlurl "file://[pwd]/$packidlurl"
	    } else {
		set idlurl "file://$packidlurl"
	    }
	} else {
	    set idlurl $packidlurl
	}
	set link [::dom::document createElement $idl "link"]
	::dom::element setAttribute $link href $idlurl
    } elseif {$whatref == "noref"} {
	set lastidx [string last / $packidlurl]
	incr lastidx
	set idlfilename [string range $packidlurl $lastidx end]
	set idlfile [::dom::document createElement $idl "fileinarchive"]
	::dom::element setAttribute $idlfile name [file tail $idlfilename]
    } else {
	error "oops"
    }

    set impl [::dom::document createElement $softpkg "implementation"]
    ::dom::element setAttribute $impl id "*"

    set code [::dom::document createElement $impl code]
    ::dom::element setAttribute $code type $codetype

    if {$whatref == "reffiles" || !$mkpackage} {
	if {[file exists $packliburl]} {
	    if {[file pathtype $packliburl] == "relative"} {
		set liburl "file://[pwd]/$packliburl"
	    } else {
		set liburl "file://$packliburl"
	    }
	} else {
	    set liburl $packliburl
	}
	set link [::dom::document createElement $code "link"]
	::dom::element setAttribute $link href $liburl
    } elseif {$whatref == "noref"} {
	set lastidx [string last / $packliburl]
	incr lastidx
	set libfilename [string range $packliburl $lastidx end]
	set libfile [::dom::document createElement $code "fileinarchive"]
	::dom::element setAttribute $libfile name [file tail $libfilename]
    } else {
	error "oops"
    }

    if {$codetype == "DLL"} {
	set entrypt [::dom::document createElement $impl "entrypoint"]
	::dom::document createTextNode $entrypt $homeentrypt
    }

    set softpkgdoc [::dom::pretty_print $doc]
    ::dom::DOMImplementation destroy $doc

    set file [open $filename w]
    puts $file $softpkgdoc
    close $file
}

proc Ass::FileHasProperties {homeorcompdata parentnode} {
    variable configdata
    variable irdata
    upvar $homeorcompdata data

    set irname $data($parentnode,name)

    foreach portname $irdata($irname,ports) {
	update ;# strategic update

	if {$irdata($irname,port,$portname,kind) != "Attribute"} {
	    continue
	}

	if {![info exists configdata($parentnode,$portname,set)]} {
	    continue
	} elseif {!$configdata($parentnode,$portname,set)} {
	    continue
	}

	return 1
    }

    return 0
}

proc Ass::FileDumpProperties {homeorcompdata parentnode filename} {
    variable configdata
    variable irdata
    upvar $homeorcompdata data
    variable status

    set localname [file tail $filename]
    set status "Writing Property File $localname ..."
    update
    
    set doc [::dom::DOMImplementation create]
    set props [::dom::document createElement $doc "properties"]

    set irname $data($parentnode,name)

    foreach portname $irdata($irname,ports) {
	update ;# strategic update

	if {$irdata($irname,port,$portname,kind) != "Attribute"} {
	    continue
	}

	if {![info exists configdata($parentnode,$portname,set)]} {
	    continue
	} elseif {!$configdata($parentnode,$portname,set)} {
	    continue
	}

	set porttype $irdata($irname,port,$portname,type)

	FileDumpPropertyItem $props $parentnode \
		$portname $portname $portname \
		$porttype
    }

    set propdoc [::dom::pretty_print $doc]
    ::dom::DOMImplementation destroy $doc

    set file [open $filename w]
    puts $file $propdoc
    close $file
}

proc Ass::FileDumpPropertyItem {doc id attrname name simplename type} {
    variable configdata

    switch -- $type {
	boolean -
	short -
	long -
	float -
	double {
	    set defaultvalue 0
	    set typename $type
	}
	{unsigned short} {
	    set defaultvalue 0
	    set typename ushort
	}
	{unsigned long} {
	    set defaultvalue 0
	    set typename ulong
	}
	{long long} {
	    set defaultvalue 0
	    set typename longlong
	}
	{unsigned long long} {
	    set defaultvalue 0
	    set typename ulonglong
	}
	{long double} {
	    set defaultvalue 0
	    set typename longdouble
	}
	char -
	octet -
	wchar {
	    set defaultvalue "x"
	    set typename $type
	}
	string -
	wstring {
	    set defaultvalue ""
	    set typename $type
	}
    }

    switch -- $type {
	boolean -
	short -
	long -
	{unsigned short} -
	{unsigned long} -
	float -
	double -
	char -
	octet -
	string -
	{long long} -
	{unsigned long long} -
	{long double} -
	wchar -
	wstring {
	    if {![info exists configdata($id,$attrname,value,$name)]} {
		set configdata($id,$attrname,value,$name) $defaultvalue
	    }

	    set propnode [::dom::document createElement $doc simple]

	    if {$simplename != ""} {
		::dom::element setAttribute $propnode name $simplename
	    }

	    ::dom::element setAttribute $propnode type $typename

	    set valuenode [::dom::document createElement $propnode value]
	    ::dom::document createTextNode $valuenode \
		    $configdata($id,$attrname,value,$name)

	    return
	}
    }

    set tckind [lindex $type 0]

    if {$tckind != "sequence" && $tckind != "struct" && \
	    $tckind != "valuetype" && $tckind != "array"} {
	return
    }

    if {$tckind == "sequence"} {
	if {![info exists configdata($id,$attrname,value,$name)]} {
	    set configdata($id,$attrname,value,$name) 0
	}

	set subtype [lindex $type 1]
	set seqnode [::dom::document createElement $doc sequence]

	if {$simplename != ""} {
	    ::dom::element setAttribute $seqnode name $simplename
	}

	::dom::element setAttribute $seqnode type "sequence<$subtype>"

	set length $configdata($id,$attrname,value,$name)

	for {set i 0} {$i < $length} {incr i} {
	    set subname "$name\[$i\]"
	    FileDumpPropertyItem $seqnode $id \
		    $attrname $subname "" $subtype
	}
    } elseif {$tckind == "struct" || $tckind == "valuetype"} {
	set members [lindex $type 2]

	if {$tckind == "valuetype"} {
	    set basetc [lindex $type 3]
	    while {$basetc != "" && $basetc != "0"} {
		foreach {member_name member_type} [lindex $basetc 2] {
		    set members [linsert $members 0 $member_name $member_type]
		}
		set basetc [lindex $basetc 3]
	    }

	    set node [::dom::document createElement $doc valuetype]
	} else {
	    set node [::dom::document createElement $doc struct]
	}

	set typeid [lindex $type 1]
	::dom::element setAttribute $node name $name
	::dom::element setAttribute $node type $typeid

	foreach {member_name member_type} $members {
	    set subname "$name.$member_name"
	    FileDumpPropertyItem $node $id \
		    $attrname $subname $member_name \
		    $member_type
	}
    }
}

proc Ass::FileDumpHome {homenode partitioning connections} {
    variable configdata
    variable packdata
    variable homedata

    #
    # add node to componentfiles
    #

    set homeid $homedata($homenode,id)
    set packnode $homedata($homenode,package)
    set packid $packdata($packnode,id)

    #
    # add home placement node to partitioning
    #

    set homeplacement [::dom::document createElement \
	    $partitioning "homeplacement"]
    ::dom::element setAttribute $homeplacement id $homeid

    set componentfileref [::dom::document createElement \
	    $homeplacement "componentfileref"]
    ::dom::element setAttribute $componentfileref idref $packid

    if {$homedata($homeid,cpf)} {
	set propertyfilename $homeid
	append propertyfilename ".cpf"
	set homeproperties [::dom::document createElement \
		$homeplacement "homeproperties"]
	set propfile [::dom::document createElement \
		$homeproperties "fileinarchive"]
	::dom::element setAttribute $propfile name $propertyfilename
    }
    
    if {[info exists homedata($homenode,nsreg)] && \
	    [info exists homedata($homenode,nsname)] && \
	    $homedata($homenode,nsreg) && \
	    $homedata($homenode,nsname) != ""} {
	set regns [::dom::document createElement \
		$homeplacement "registerwithnaming"]
	::dom::element setAttribute $regns \
		name $homedata($homenode,nsname)
    }

    return $homeplacement
}

proc Ass::FileDumpComponent {compnode homeplacement connections} {
    variable configdata
    variable compdata
    variable irdata

    set compid $compdata($compnode,id)
    set compname $compdata($compnode,name)
	
    set ci [::dom::document createElement \
	    $homeplacement "componentinstantiation"]
    ::dom::element setAttribute $ci id $compid

    if {$compdata($compid,cpf)} {
	set comppropertyfilename $compid
	append comppropertyfilename ".cpf"
	set compproperties [::dom::document createElement \
		$ci "componentproperties"]
	set comppropfile [::dom::document createElement \
		$compproperties "fileinarchive"]
	::dom::element setAttribute $comppropfile \
		name $comppropertyfilename
    }
	
    # register component itself
	
    if {[info exists compdata($compnode,nsreg)] && \
	    [info exists compdata($compnode,nsname)] && \
	    $compdata($compnode,nsreg) && \
	    $compdata($compnode,nsname) != ""} {
	set registercomponent [::dom::document createElement \
		$ci "registercomponent"]
	set regns [::dom::document createElement \
		$registercomponent "registerwithnaming"]
	::dom::element setAttribute $regns \
		name $compdata($compnode,nsname)
    }
    
    # go through ports
	
    foreach portname $irdata($compname,ports) {
	update ;# strategic update
	set portkind $irdata($compname,port,$portname,kind)

	#
	# Register facets and consumers with Naming Service
	#

	if {$portkind == "Provides" || $portkind == "Consumes"} {
	    if {[info exists configdata($compnode,$portname,nsreg)] && \
		    [info exists configdata($compnode,$portname,nsname)] && \
		    $configdata($compnode,$portname,nsreg) && \
		    $configdata($compnode,$portname,nsname) != ""} {
		set registercomponent [::dom::document createElement \
			$ci "registercomponent"]

		switch -- $portkind {
		    Provides {
			set xmltype "providesidentifier"
		    }
		    Consumes {
			set xmltype "consumesidentifier"
		    }
		}

		set identifier [::dom::document createElement \
			$registercomponent $xmltype]
		::dom::document createTextNode $identifier $portname

		set regns [::dom::document createElement \
			$registercomponent "registerwithnaming"]
		::dom::element setAttribute $regns \
			name $configdata($compnode,$portname,nsname)
	    }
	}

	#
	# For receptacles and event sources, make their connections
	#

	if {($portkind == "Uses" || \
		$portkind == "Publishes" || \
		$portkind == "Emits") && \
		[info exists configdata($compnode,$portname,conn)]} {

	    FileDumpConnectionsForPort $compnode $portname $connections
	}
    }
}

proc Ass::FileDumpConnectionsForPort {compnode portname connections} {
    variable configdata
    variable compdata

    set compid $compdata($compnode,id)
    set portkind $configdata($compnode,$portname,kind)
    set numconns $configdata($compnode,$portname,conn)

    for {set i 0} {$i < $numconns} {incr i} {
	update ;# strategic update

	if {![info exists configdata($compnode,$portname,$i,type)]} {
	    continue
	}

	set sinktype $configdata($compnode,$portname,$i,type)
	set sinkdata $configdata($compnode,$portname,$i,sink)

	if {$sinktype == "none"} {
	    continue
	}

	if {$portkind == "Uses"} {
	    set xmltype1 "connectinterface"
	    set xmltype2 "usesport"
	    set xmltype3 "usesidentifier"
	} elseif {$portkind == "Emits"} {
	    set xmltype1 "connectevent"
	    set xmltype2 "emitsport"
	    set xmltype3 "emitsidentifier"
	} else {
	    set xmltype1 "connectevent"
	    set xmltype2 "publishesport"
	    set xmltype3 "publishesidentifier"
	}

	set connection [::dom::document createElement \
		$connections $xmltype1]
	set sourceport [::dom::document createElement \
		$connection $xmltype2]

	set sourceidentifier [::dom::document createElement \
		$sourceport $xmltype3]
	::dom::document createTextNode $sourceidentifier $portname

	set cidref [::dom::document createElement \
		$sourceport "componentinstantiationref"]
	::dom::element setAttribute $cidref idref $compid

	switch -- $sinktype {
	    URL {
		set exif [::dom::document createElement \
			$connection "existinginterface"]
		set findby [::dom::document createElement \
			$exif "findby"]
		set objref [::dom::document createElement \
			$findby "stringifiedobjectref"]
		::dom::document createTextNode $objref $sinkdata
	    }
	    NS {
		set exif [::dom::document createElement \
			$connection "existinginterface"]
		set findby [::dom::document createElement \
			$exif "findby"]
		set naming [::dom::document createElement \
			$findby "namingservice"]
		::dom::element setAttribute $naming name $sinkdata
	    }
	    Port {
		set sinkcompnode [lindex $sinkdata 0]
		set sinkportname [lindex $sinkdata 1]

		set sinkcompname $compdata($sinkcompnode,name)
		set sinkcompid $compdata($sinkcompnode,id)

		if {$sinkportname != -1} {
		    set sinkportkind $configdata($sinkcompnode,$sinkportname,kind)
		} else {
		    set sinkportkind "Component"
		}

		switch -- $sinkportkind {
		    Component {
			set xmltype1 "componentsupportedinterface"
		    }
		    Provides {
			set xmltype1 "providesport"
			set xmltype2 "providesidentifier"
		    }
		    Consumes {
			set xmltype1 "consumesport"
			set xmltype2 "consumesidentifier"
		    }
		    default {
			error "oops: $compnode $portname $i $sinkportkind"
		    }
		}

		set sinkport [::dom::document createElement \
			$connection $xmltype1]

		if {$sinkportname != -1} {
		    set sinkidentifier [::dom::document \
			    createElement $sinkport $xmltype2]
		    ::dom::document createTextNode \
			    $sinkidentifier \
			    $sinkportname
		}

		set scidref [::dom::document createElement \
			$sinkport "componentinstantiationref"]
		::dom::element setAttribute $scidref idref \
			$sinkcompid
	    }
	    default {
		error "oops: $compnode $portname $i $sinktype"
	    }
	}
    }
}

#
# ======================================================================
# Softpkg Wizard
# ======================================================================
#

proc Ass::SoftpkgWizSetup1 {} {
    variable softwizwidgets
    variable softwizdata

    $softwizwidgets(title) configure -text "Software Package Wizard"

    $softwizwidgets(info) configure -text \
	    "Welcome to the Software Package Wizard. It will help\
	    you to create a CORBA Software Package for your Component.\
	    To create a CORBA Software Package, this Wizard requires a\
	    readily built Component - usually a shared library (DLL)\
	    or executable - and the corresponding IDL file. Please\
	    click on the \"Next\" button to proceed."

    $softwizwidgets(prev) configure -state disabled
    $softwizwidgets(ok) configure -state disabled
    $softwizwidgets(cancel) configure -state normal
    $softwizwidgets(next) configure -state normal
}

proc Ass::SoftpkgWizNext1 {} {
    # nothing to do
}

proc Ass::SoftpkgWizSelectImplFile {top} {
    variable softwizdata
    variable initialdir

    set filetypes [list \
	    [list "Shared Libraries" [list [info sharedlibextension]]] \
	    [list "All Files" [list *]] \
	    ]

    if {[info exists softwizdata(implfile)] && \
	    $softwizdata(implfile) != ""} {
	catch {set olddir [file dirname $softwizdata(implfile)]}
    }

    if {![info exists olddir] || ![file exists $olddir]} {
	set olddir $initialdir
    }

    set filename [tk_getOpenFile -title "Select Implementation File" \
	    -parent $top -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	return
    }

    set initialdir [file dirname $filename]
    set softwizdata(implfile) $filename

    switch -- [file extension $filename] {
	.dll -
	.so -
	.sl {
	    set softwizdata(impltype) "DLL"
	}
	.exe {
	    set softwizdata(impltype) "Executable"
	}
    }
}

proc Ass::SoftpkgWizSetup2 {} {
    variable softwizwidgets
    variable softwizdata

    $softwizwidgets(title) configure -text "Step 1: Implementation File"
    $softwizwidgets(info) configure -text \
	    "Please select the file that contains your component\
	    implementation. Please also select whether this file\
	    is a shared library (DLL) or executable. If you have\
	    made both selections, click on the \"Next\" button."

    if {![info exists softwizdata(implfile)]} {
	set softwizdata(implfile) ""
    }

    if {![info exists softwizdata(impltype)]} {
	set softwizdata(impltype) ""
    }

    set f1 [frame $softwizwidgets(wiz).f1]
    set f1lab [label $f1.lab -text "Implementation" -width 15 -anchor w]
    set f1ent [entry $f1.ent -width 40 \
	    -textvariable ::Ass::softwizdata(implfile)]
    set f1but [button $f1.but -width 10 \
	    -text "Select" -anchor c -relief raised -bd 1 \
	    -command "Ass::SoftpkgWizSelectImplFile $softwizwidgets(top)"]
    pack $f1lab -side left
    pack $f1ent -side left -fill x -expand yes
    pack $f1but -side left -padx 10
    pack $f1 -side top -fill x -padx 10

    set empty1 [frame $softwizwidgets(wiz).empty1]
    pack $empty1 -side top -fill x -pady 10

    set f2 [frame $softwizwidgets(wiz).f2]
    set impltypelab [label $f2.impltypelab -anchor w \
	    -text "This implementation file is a ..."]
    set impltype1 [radiobutton $f2.impltype1 \
	    -text "Shared Library (DLL)" -anchor w \
	    -variable ::Ass::softwizdata(impltype) -value "DLL"]
    $impltype1 configure -activebackground [$impltype1 cget -bg]
    set impltype2 [radiobutton $f2.impltype2 \
	    -text "Standalone Executable" -anchor w \
	    -variable ::Ass::softwizdata(impltype) -value "Executable"]
    $impltype2 configure -activebackground [$impltype2 cget -bg]
    pack $impltypelab $impltype1 $impltype2 -side top -fill x
    pack $f2 -side top -fill x -padx 10

    $softwizwidgets(prev) configure -state normal
    $softwizwidgets(ok) configure -state disabled
    $softwizwidgets(cancel) configure -state normal
    $softwizwidgets(next) configure -state normal
}

proc Ass::SoftpkgWizNext2 {} {
    variable softwizwidgets
    variable softwizdata

    if {$softwizdata(implfile) == "" || \
	    ($softwizdata(impltype) != "DLL" && \
	    $softwizdata(impltype) != "Executable")} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Not so fast" \
		-message "Please complete both entries on this page first."
	error "not complete"
    }
}

proc Ass::SoftpkgWizSelectIDLFile {top} {
    variable softwizdata
    variable initialdir

    set filetypes {
	{"IDL Files" {.idl}}
	{"All Files" {*}}
    }

    if {[info exists softwizdata(idlfile)] && \
	    $softwizdata(idlfile) != ""} {
	catch {set olddir [file dirname $softwizdata(idlfile)]}
    }

    if {![info exists olddir] || ![file exists $olddir]} {
	set olddir $initialdir
    }

    set filename [tk_getOpenFile -title "Select IDL File" \
	    -parent $top -filetypes $filetypes -initialdir $olddir]

    if {$filename == ""} {
	return
    }

    set initialdir [file dirname $filename]
    set softwizdata(idlfile) $filename
    set status "Ok"
}

proc Ass::SoftpkgWizSetup3 {} {
    variable softwizwidgets
    variable softwizdata

    $softwizwidgets(title) configure -text "Step 2: IDL File"
    $softwizwidgets(info) configure -text \
	    "Please select the IDL file that contains the IDL for\
	    your component. If necessary, you can add options to\
	    the IDL compiler here, e.g. to define a search path\
	    for included files. After pressing \"Next\", please\
	    be patient, as loading the file into the Interface\
	    Repository may take some time."

    if {![info exists softwizdata(idlfile)]} {
	set softwizdata(idlfile) ""
    }

    if {![info exists softwizdata(idlopts)]} {
	set softwizdata(idlopts) ""
    }

    set f1 [frame $softwizwidgets(wiz).f1]
    set f1lab [label $f1.lab -text "IDL File" -width 15 -anchor w]
    set f1ent [entry $f1.ent -width 40 \
	    -textvariable ::Ass::softwizdata(idlfile)]
    set f1but [button $f1.but -width 10 \
	    -text "Select" -anchor c -relief raised -bd 1 \
	    -command "Ass::SoftpkgWizSelectIDLFile $softwizwidgets(top)"]
    pack $f1lab -side left
    pack $f1ent -side left -expand yes -fill x
    pack $f1but -side left -padx 10
    pack $f1 -side top -fill x -padx 10

    set f2 [frame $softwizwidgets(wiz).f2]
    set f2lab [label $f2.lab -text "IDL Options" -width 15 -anchor w]
    set f2ent [entry $f2.ent -width 40 \
	    -textvariable ::Ass::softwizdata(idlopts)]
    pack $f2lab -side left
    pack $f2ent -side left -expand yes -fill x
    pack $f2 -side top -fill x -padx 10

    $softwizwidgets(prev) configure -state normal
    $softwizwidgets(ok) configure -state disabled
    $softwizwidgets(cancel) configure -state normal
    $softwizwidgets(next) configure -state normal
}

proc Ass::SoftpkgWizNext3 {} {
    variable softwizwidgets
    variable softwizdata

    if {$softwizdata(idlfile) == ""} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Not so fast" \
		-message "Please select an IDL file."
	error "not complete"
    }

    set idlfile $softwizdata(idlfile)

    if {[catch {set idllocalfile [DownloadFile $idlfile]} err]} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Error downloading IDL file" \
		-message $err
	error "not complete"
    }

    if {[catch {LoadIDLFile $idllocalfile $softwizdata(idlopts)} err]} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Error loading IDL file" \
		-message "The IDL file could not be loaded into the\
		Interface Repository: $err"
	error "not complete"
    }
}

proc Ass::SoftpkgWizHomeSelected {} {
    variable softwizwidgets
    variable softwizdata

    if {$softwizdata(impltype) == "DLL"} {
	set homename [string range $softwizdata(home) 2 end]
	set flatname [string map {:: _} $homename]
	set softwizdata(entrypt) "create_$flatname"
    }
}

proc Ass::SoftpkgWizSetup4 {} {
    variable InterfaceRepository
    variable softwizwidgets
    variable softwizdata
    variable irdata

    $softwizwidgets(title) configure -text "Step 3: Home Type"
    $softwizwidgets(info) configure -text \
	    "Please select the home type implemented by this package.\
	    A file may contain multiple component implementations, but\
	    each package may reference only one of them. If you want to\
	    use more than one home from a package, you will have to\
	    repeat this process. The name of the entry point is derived\
	    from the home name; there is usually no need to change it."

    #
    # locate all homes in the Interface Repository
    #

    set allhomes [list]
    set modules [list $InterfaceRepository]

    corba::try {
	while {[llength $modules] > 0} {
	    set newmodules [list]

	    foreach module $modules {
		foreach submodule [$module contents dk_Module 1] {
		    lappend newmodules $submodule
		}
		foreach home [$module contents dk_Home 1] {
		    lappend allhomes [$home absolute_name]
		}
	    }

	    set modules $newmodules
	}
    }
    catch {... err} {
	puts "oops: $err"
    }

    foreach homename $allhomes {
	IntrospectHome $homename
    }
    
    if {![info exists softwizdata(home)]} {
	set softwizdata(home) ""
    }

    if {![info exists softwizdata(entrypt)]} {
	set softwizdata(entrypt) ""
    }

    set f1 [frame $softwizwidgets(wiz).f1]
    set f1lab [label $f1.lab -text "Select a Home" -width 15 -anchor w]
    set f1com [ComboBox $f1.combo -width 50 -values $allhomes \
	    -textvariable ::Ass::softwizdata(home) \
	    -modifycmd ::Ass::SoftpkgWizHomeSelected]
    pack $f1lab $f1com -side left
    pack $f1 -side top -fill x -padx 10

    set f2 [frame $softwizwidgets(wiz).f2]
    set f2lab [label $f2.lab -text "Entry Point" -width 15 -anchor w]
    set f2ent [entry $f2.ent -width 40 \
	    -textvariable ::Ass::softwizdata(entrypt)]
    pack $f2lab -side left
    pack $f2ent -side left -expand yes -fill x
    pack $f2 -side top -fill x -padx 10

    if {$softwizdata(impltype) == "Executable"} {
	set softwizdata(entrypt) "main"
	$f2ent configure -state disabled
    }

    $softwizwidgets(prev) configure -state normal
    $softwizwidgets(ok) configure -state disabled
    $softwizwidgets(cancel) configure -state normal
    $softwizwidgets(next) configure -state normal
}

proc Ass::SoftpkgWizNext4 {} {
    variable softwizwidgets
    variable softwizdata
    variable irdata

    if {$softwizdata(home) == "" || \
	    $softwizdata(entrypt) == "" || \
	    ![info exists irdata($softwizdata(home),type)] || \
	    $irdata($softwizdata(home),type) != "Home"} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Not so fast" \
		-message "Please select a Home."
	error "not complete"
    }
}

proc Ass::SoftpkgWizSetup5 {} {
    variable softwizwidgets
    variable softwizdata

    $softwizwidgets(title) configure -text "Step 4: Just About Finished"
    $softwizwidgets(info) configure -text \
	    "The wizard now has all the information to create the new\
	    Software Package. Please select a unique identifier for the\
	    package. Once you press the OK button, this package will be\
	    added to the Assembly. You can then right-click on it on the\
	    \"Packages\"\ page to create a CORBA Software Package Descriptor\
	    (.csd) or a CORBA Software Package (.zip) file."

    set abshomename $softwizdata(home)
    set lastidx [string last : $abshomename]
    incr lastidx
    set shorthomename [string range $abshomename $lastidx end]

    if {![info exists softwizdata(identifier)]} {
	set softwizdata(identifier) $shorthomename
    }

    set f1 [frame $softwizwidgets(wiz).f1]
    set f1lab [label $f1.lab -text "Identifier" -width 15 -anchor w]
    set f1ent [entry $f1.ent -width 40 \
	    -textvariable ::Ass::softwizdata(identifier)]
    pack $f1lab -side left
    pack $f1ent -side left -expand yes -fill x
    pack $f1 -side top -fill x -padx 10

    $softwizwidgets(prev) configure -state normal
    $softwizwidgets(ok) configure -state normal
    $softwizwidgets(cancel) configure -state normal
    $softwizwidgets(next) configure -state disabled
}

proc Ass::SoftpkgWizOkay5 {} {
    variable softwizwidgets
    variable softwizdata
    variable nodecounter
    variable packdata
    variable irdata

    if {$softwizdata(identifier) == ""} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Not so fast" \
		-message "Please set an Identifier."
	error "not complete"
    }

    #
    # Add Package Node
    #

    corba::try {
	set packnode [PackAddPackage $softwizdata(identifier) Softpkg ""]

	#
	# Add Node for Home
	#

	set abshomename $softwizdata(home)
	set homerepoid $irdata($abshomename,id)
	set homename [string range $softwizdata(home) 2 end]

	set lastidx [string last : $homename]
	incr lastidx
	set shorthomename [string range $homename $lastidx end]
	
	incr nodecounter
	lappend packdata($packnode,children) $nodecounter
	set packdata($nodecounter,nodetype) Home
	set packdata($nodecounter,id) $softwizdata(identifier)
	set packdata($nodecounter,text) $shorthomename
	set packdata($nodecounter,parent) $packnode
	set packdata($nodecounter,codetype) $softwizdata(impltype)
	set packdata($nodecounter,whatref) "noref"
	set packdata($nodecounter,pkgurl) ""
	set packdata($nodecounter,idlurl) $softwizdata(idlfile)
	set packdata($nodecounter,liburl) $softwizdata(implfile)
	set packdata($nodecounter,repoid) $homerepoid
	set packdata($nodecounter,entrypt) $softwizdata(entrypt)
	set packdata($nodecounter,idlintrospected) 0
	set packdata($nodecounter,idllocalfile) ""

	PackIntrospectIDLFile $nodecounter
	PackDisplayPackage $packnode
    } catch {... err} {
	tk_messageBox -parent $softwizwidgets(top) \
		-type ok -icon error -default ok \
		-title "Oops" \
		-message "Oops: $err"
    }
}

proc Ass::SoftpkgWizard {} {
    variable softwizwidgets
    variable softwizdata
    variable status

    if {[winfo exists .softpkgWizard]} {
	set status "Wizardry in Progress ..."
	return
    }

    set top [toplevel .softpkgWizard]
    wm title $top "Software Package Wizard"

    set title [label $top.title -anchor center \
	    -font {Arial 18 bold} \
	    -text "Software Package Wizard"]
    pack $title -side top -fill x -pady 10

    set topsep [Separator $top.topsep -orient horizontal]
    pack $topsep -side top -fill x -pady 10

    set infolabel [label $top.info -wraplength 500 -height 5 \
	    -anchor n -justify left]
    pack $infolabel -side top -fill x -pady 10

    set wizardry [frame $top.wiz -width 600 -height 200]
    pack $wizardry -side top -fill both -expand yes
    pack propagate $wizardry 0

    set botsep [Separator $top.botsep -orient horizontal]
    pack $botsep -side top -fill x -pady 10

    set botbuts [frame $top.botbuts]
    set botbut1 [button $botbuts.but1 -width 12 \
	    -text "<< Previous" \
	    -command "set ::Ass::softwizdata(mutex) 1"]
    set botbut2 [button $botbuts.but2 -width 12 \
	    -text "Cancel" \
	    -command "set ::Ass::softwizdata(mutex) 2"]
    set botbut3 [button $botbuts.but3 -width 12 \
	    -text "OK" \
	    -command "set ::Ass::softwizdata(mutex) 3"]
    set botbut4 [button $botbuts.but4 -width 12 \
	    -text "Next >>" \
	    -command "set ::Ass::softwizdata(mutex) 4"]
    pack $botbut1 $botbut2 $botbut3 $botbut4 -side left -pady 10 -padx 10
    pack $botbuts

    set softwizwidgets(top) $top
    set softwizwidgets(title) $title
    set softwizwidgets(info) $infolabel
    set softwizwidgets(wiz) $wizardry
    set softwizwidgets(prev) $botbut1
    set softwizwidgets(cancel) $botbut2
    set softwizwidgets(ok) $botbut3
    set softwizwidgets(next) $botbut4

    set softwizdata(mutex) 0
    set softwizdata(curpage) 1
    set softwizdata(pagedirty) 1

    wm deiconify $top
    wm minsize $top 600 400
    focus $top
    # grab set $top

    set status "Wizardry in Progress ..."

    while {42} {
	if {$softwizdata(pagedirty)} {
	    foreach subitem [winfo children $softwizwidgets(wiz)] {
		catch {destroy $subitem}
	    }
	    SoftpkgWizSetup$softwizdata(curpage)
	    set softwizdata(pagedirty) 0
	}

	vwait ::Ass::softwizdata(mutex)

	switch -- $::Ass::softwizdata(mutex) {
	    1 {
		incr softwizdata(curpage) -1
		set softwizdata(pagedirty) 1
	    }
	    2 {
		break
	    }
	    3 {
		if {![catch {SoftpkgWizOkay$softwizdata(curpage)}]} {
		    break
		}
	    }
	    4 {
		if {![catch {SoftpkgWizNext$softwizdata(curpage)}]} {
		    incr softwizdata(curpage) 1
		    set softwizdata(pagedirty) 1
		}
	    }
	}
    }

    # grab release $top
    destroy $top

    if {$::Ass::softwizdata(mutex) == 3} {
	catch {unset softwizwidgets}
	catch {unset softwizdata}
	set status "Wizardry complete."
    } elseif {$::Ass::softwizdata(mutex) == 2} {
	set status "Wizardry interrupted."
    }
}

#
# ======================================================================
# Main
# ======================================================================
#

proc Ass::Init {argv} {
    variable InterfaceRepository
    variable PackLoadSoftpkgFile ""
    variable PackLoadSoftpkgURL ""
    variable PackLoadAssemblyURL ""
    variable packdata
    variable homedata
    variable compdata
    variable nodecounter
    variable initialdir
    variable tempfiles
    variable tempidx
    variable tempdir
    variable status
    variable setup
    variable init
    global ifrpid

    if {![info exists nodecounter]} {
	set nodecounter 0
    }

    if {![info exists packdata(allpackages)]} {
	set packdata(allpackages) [list]
    }

    if {![info exists homedata(allhomes)]} {
	set homedata(allhomes) [list]
    }

    if {![info exists compdata(allcomponents)]} {
	set compdata(allcomponents) [list]
    }

    if {![info exists initialdir]} {
	set initialdir [pwd]
    }

    if {![info exists tempdir]} {
	if {[file exists /tmp] && [file isdirectory /tmp]} {
	    set tempdir /tmp
	} else {
	    set tempdir [pwd]
	}
    }

    if {![info exists tempfiles]} {
	set tempfiles [list]
    }

    if {![info exists tempidx]} {
	set tempidx 0
    }

    if {![info exists status]} {
	set status ""
    }

    if {![info exists InterfaceRepository]} {
	set InterfaceRepository 0
    }

    if {![info exists ifrpid]} {
	set ifrpid 0
    }

    if {![info exists setup(idl)]} {
	if {$::tcl_platform(platform) == "windows"} {
	    set setup(idl) "idl.exe"
	} else {
	    set setup(idl) "idl"
	}
    }

    if {![info exists setup(idlopts)]} {
	set setup(idlopts) ""
    }

    if {![info exists setup(ird)]} {
	if {$::tcl_platform(platform) == "windows"} {
	    set setup(ird) "ird.exe"
	} else {
	    set setup(ird) "ird"
	}
    }

    if {![info exists setup(zip)]} {
	if {$::tcl_platform(platform) == "windows"} {
	    set setup(zip) "zip.exe"
	} else {
	    set setup(zip) "zip"
	}
    }

    if {![info exists setup(unzip)]} {
	if {$::tcl_platform(platform) == "windows"} {
	    set setup(unzip) "unzip.exe"
	} else {
	    set setup(unzip) "unzip"
	}
    }

    if {![info exists setup(kill)]} {
	if {$::tcl_platform(platform) == "windows"} {
	    set setup(kill) "kill.exe"
	} else {
	    set setup(kill) "kill"
	}
    }

    if {![info exists setup(host)]} {
	set setup(host) "<auto>"
    }

    if {![info exists setup(port)]} {
	set setup(port) 1235
    }

    wm withdraw .
    wm title . "MicoCCM Assembly Tool"
    InitImages
    InitGUI
    update
    wm deiconify .

    set status "Initializing Combat ..."
    update

    #if {[llength $argv] == 0 && [file exists ./ir.ior]} {
    #    set argv [list -ORBInitRef InterfaceRepository=file://[pwd]/ir.ior]
    #}

    set argv [eval corba::init $argv]
    source [file join $::selfdir lib ir3.tcl]
    source [file join $::selfdir lib CCM.tcl]
    combat::ir add $_ir_ir3
    combat::ir add $_ir_CCM

    set status "Welcome to the MicoCCM Assembly Tool"
    set init 1
}

#
# Source support files
#

if {[file exists [file join $::selfdir setup.tcl]]} {
    source [file join $::selfdir setup.tcl]
}

if {[file exists [file join $::homedir setup.tcl]]} {
    source [file join $::homedir setup.tcl]
}

source [file join $::selfdir assio.tcl]
source [file join $::selfdir assimg.tcl]
source [file join $::selfdir fpsfhttpd.tcl]

#
# If not initialized, run. (Otherwise, we have been sourced again.)
#

if {$::Ass::init == 0} {
    Ass::Init $argv

    vwait forever

    if {[info exists ifrpid] && $ifrpid != 0} {
	catch {exec $::Ass::setup(kill) $ifrpid}
    }

    Ass::CleanupTempFiles

    exit
}
