#
# Check for availability of packages
#

foreach package {
    BWidget-1.3.1
    combat-0.7
    dom_pretty_print-0.2
    tcldom-2.0
    tclxml-2.0
    tkcon-2.3
    tcl++
} {
    set packpath [file join [file dirname [info script]] lib $package]
    if {[file exists $packpath]} {
	lappend auto_path $packpath
    }
}

#
# initializes packages info
#

catch {package require foo}

proc check_package {package {reqversion ""}} {
    puts -nonewline "checking for package $package ... "
    flush stdout
    if {[lsearch -exact [package names] $package] != -1} {
	set packversions [package versions $package]
	set ok 0
	foreach packversion $packversions {
	    if {$reqversion != "" && [package vsatisfies $packversion $reqversion]} {
		set ok 1
	    }
	    puts -nonewline "$packversion "
	    flush stdout
	}
	puts ""

	if {$reqversion != "" && !$ok} {
	    return 0
	}
    } else {
	puts "not found."
	return 0
    }

    return 1
}

if {![check_package BWidget 1.2]} {
    puts "configure: error: package BWidget not found (version 1.2 required)"
    exit 1
}

if {![check_package xml]} {
    puts "configure: error: package xml not found"
    exit 1
}

if {![check_package dom]} {
    puts "configure: error: package dom not found"
    exit 1
}

if {![check_package dom_pretty_print]} {
    puts "configure: error: package dom_pretty_print not found"
    exit 1
}

if {![check_package Itcl] && ![check_package tcl++]} {
    puts "configure: error: either \[incr Tcl\] or tcl++ required"
    exit 1
}

if {![check_package combat 0.7]} {
    puts "configure: error: package combat not found (version 0.7 required)"
    exit 1
}

exit 0
