#!/bin/sh
# -*- TCL -*-
# the next line restarts using tclsh \
exec tclsh8.3 "$0" ${1+"$@"}

#
# Where am I?
#

if {[catch {::freewrap::iswrapped $::argv0}] || \
	[::freewrap::iswrapped $::argv0] == 0} {
    #
    # Not wrapped.
    #
    set homedir [file dirname [info script]]
    set selfdir $homedir
} else {
    #
    # Probably wrapped. Try to find my home directory using $argv0
    #
    set homedir [file join [pwd] [file dirname $::argv0]]
    set selfdir [file dirname [info script]]
}

#
# Load Packages
#

foreach package {
    combat-0.7
    tcldom-2.0
    tclxml-2.0
    tcl++
} {
    if {[file exists [file join $::selfdir lib $package]]} {
	lappend auto_path [file join $::selfdir lib $package]
    } elseif {[file exists [file join $::homedir lib $package]]} {
	lappend auto_path [file join $::homedir lib $package]
    } else {
	# file exists does not seem to work in freewrap
	lappend auto_path [file join $::selfdir lib $package]
    }
}

package require combat
package require dom

#
# create Ass namespace
#

namespace eval Ass {
    variable nodecounter 0
    variable packdata
    variable homedata
    variable compdata
    variable tempfiles [list]
    variable tempidx 0
    variable tempdir ""
    variable status ""
    variable setup

    if {[file exists /tmp] && [file isdirectory /tmp]} {
	set tempdir /tmp
    } else {
	set tempdir [pwd]
    }

    set setup(host) "<auto>"
    set setup(port) 1235
    set setup(unzip) "unzip"

    set packdata(allpackages) [list]
    set homedata(allhomes) [list]
    set compdata(allcomponents) [list]
}

#
# Source support files
#

if {[file exists [file join $::selfdir setup.tcl]]} {
    source [file join $::selfdir setup.tcl]
}

if {[file exists [file join $::homedir setup.tcl]]} {
    source [file join $::homedir setup.tcl]
}

source [file join $::selfdir assio.tcl]
source [file join $::selfdir fpsfhttpd.tcl]

#
# ======================================================================
# Main
# ======================================================================
#

#
# Initialize Combat
#

set argv [eval corba::init $argv]
source [file join [file dirname [info script]] lib CCM.tcl]
combat::ir add $_ir_CCM

#
# Handle Options
#

proc usage {} {
    global argv0
    puts stderr "usage: $argv0 \[options\] <archive>"
    puts stderr "  available options:"
    puts stderr "    --ccmd <IOR>"
    puts stderr "    --host <host>\[:port\]"
    puts stderr "    --server \[host:\]<portno>"
    puts stderr "    --dont-call-configuration-complete"
    puts stderr "    --collocated"
    puts stderr "    -v"
    exit 1
}

set archive ""
set ccmdior ""
set ccmdhost ""
set verbose 0
set collocated 0
set callcc 1

for {set i 0} {$i < [llength $argv]} {incr i} {
    switch -glob -- [lindex $argv $i] {
	--dont-call-configuration-complete {
	    set callcc 0
	}
	--collocate* -
	--colocate* {
	    incr collocated
	}
	--ccmd {
	    set ccmdior [lindex $argv [incr i]]
	}
	--host {
	    set ccmdhost [lindex $argv [incr i]]
	}
	--server {
	    set value [lindex $argv [incr i]]
	    if {[string first : $value]} {
		set ::Ass::setup(host) [lindex [split $value :] 0]
		set ::Ass::setup(port) [lindex [split $value :] 1]
	    } else {
		set ::Ass::setup(port) $value
	    }
	}
	-v {
	    incr verbose
	}
	--* {
	    usage
	}
	default {
	    set archive [lindex $argv $i]
	}
    }
}

if {$archive == ""} {
    usage
}

#
# Display info from XML parsers
#

proc StatusDisplay {name1 name2 op} {
    global verbose
    global oldstatus

    if {$verbose && $::Ass::status != $oldstatus} {
	set oldstatus $::Ass::status
	puts $::Ass::status
    }
}

set oldstatus ""
trace variable ::Ass::status w StatusDisplay

#
# Download by http if necessary
#

if {[catch {set filename [Ass::DownloadFile $archive]} err]} {
    puts stderr "error: archive $archive not found: $err"
    exit 1
}

#
# Read Assembly
#

if {[catch {Ass::FileReadAssembly $filename} err]} {
    Ass::CleanupTempFiles
    puts stderr "error: $err"

    if {$verbose} {
	puts stderr $errorInfo
    }

    exit 1
}

#
# Contact MicoCCM Daemon
#

if {$ccmdior == ""} {
    if {$ccmdhost == ""} {
	set ccmdhost "localhost:1234"
    } elseif {[string first : $ccmdhost] == -1} {
	append ccmdhost ":1234"
    }
    set ccmdior "corbaloc::$ccmdhost/MicoCCMD"
}

corba::try {
    set ccmd [corba::string_to_object $ccmdior]
} catch {... err} {
    puts stderr "error: cannot read IOR $ccmdior: $err"
    Ass::CleanupTempFiles
    exit 1
}

corba::try {
    set ia1 [$ccmd _is_a IDL:omg.org/Components/ComponentInstallation:1.0]
    set ia2 [$ccmd _is_a IDL:omg.org/Components/ServerActivator:1.0]
    set ia3 [$ccmd _is_a IDL:mico.org/MICOCCM/MicoCCMD:1.0]
} catch {... err} {
    puts stderr "error: cannot contact CCM Daemon: $err"
    Ass::CleanupTempFiles

    if {$verbose} {
	puts stderr $errorInfo
    }

    exit 1
}

if {!$ia1 || !$ia2} {
    puts stderr "error: ccmd object does not look like MicoCCM Daemon"
    Ass::CleanupTempFiles

    if {$verbose} {
	puts stderr $errorInfo
    }

    exit 1
}

if {$verbose} {
    puts "Deploying ..."
}

corba::try {
    Ass::DeployAssembly $ccmd $collocated $callcc
} catch {... err} {
    puts stderr "error: deployment failed: $err"

    if {$verbose} {
	puts stderr $errorInfo
    }
}

Ass::CleanupTempFiles
