#
# ----------------------------------------------------------------------
# My Tiny HTTPD Server, which serves a single file
# ----------------------------------------------------------------------
#

namespace eval fpsfhttpd {
    variable serversocket 0
    variable complete 0
    variable file ""
}

proc fpsfhttpd::serve {port _file _cookie} {
    variable serversocket
    variable connected
    variable complete
    variable serving
    variable cookie
    variable mutex
    variable file

    set file [open $_file r]
    fconfigure $file -translation binary

    set cookie $_cookie

    set connected 0
    set complete 0
    set serving 0
    set mutex 0

    if {[catch {set serversocket [socket -server fpsfhttpd::accept $port]}]} {
	catch {close $file}
	error "could not listen on port $port"
    }
}

proc fpsfhttpd::wait {} {
    variable serversocket
    variable complete
    variable mutex
    variable file

    if {!$mutex} {
	vwait ::fpsfhttpd::mutex
    }

    catch {close $serversocket}
    catch {close $file}
    return $complete
}

proc fpsfhttpd::stop {} {
    variable serving
    variable mutex

    if {$serving && !$mutex} {
	vwait ::fpsfhttpd::mutex
    } else {
	set mutex 1
    }
}

proc fpsfhttpd::accept {sock peerhost peerport} {
    variable connected
    variable indata

    if {$connected} {
	close $sock
    }

    set indata ""
    set connected 1
    fconfigure $sock -translation binary -blocking 0
    fileevent $sock readable "fpsfhttpd::read $sock"
}

proc fpsfhttpd::read {sock} {
    variable complete
    variable serving
    variable indata
    variable cookie
    variable mutex

    if {[catch {set iseof [eof $sock]}]} {
	fileevent $sock readable ""
	set complete -1
	set mutex 1
	return
    }

    if {$iseof} {
	fileevent $sock readable ""
	if {[catch {close $sock}]} {
	    set complete -1
	}
	set mutex 1
	return
    }

    if {[catch {set data [::read $sock]} err]} {
	fileevent $sock readable ""
	catch {close $sock}
	set complete -1
	set mutex 1
	return
    }

    append indata $data

    if {$serving} {
	return
    }

    if {[string first \n $indata] == -1} {
	return
    }

    set firstline [string range $indata 0 [string first \n $indata]]
    set firstline [string trimright $firstline]
    set words [split $firstline]

    if {[llength $words] < 2} {
	#
	# oops
	#
	puts -nonewline $sock "400 Bad Request\r\n"
	catch {close $sock}
	set complete -1
	set mutex 1
	return
    }

    if {[lindex $words 0] != "GET"} {
	puts -nonewline $sock "400 Bad Request\r\n"
	catch {close $sock}
	set complete -1
	set mutex 1
	return
    }

    set url [string range [lindex $words 1] 1 end]

    if {$url != $cookie} {
	puts -nonewline $sock "404 File not found\r\n"
	catch {close $sock}
	set complete -1
	set mutex 1
	return
    }

    if {[llength $words] == 2} {
	#
	# http 0.9 request
	#

	set serving 1
	fileevent $sock writable "fpsfhttpd::write $sock"
	return
    }

    #
    # wait for header to be complete
    #

    if {[string range $indata end-3 end] != "\r\n\r\n" && \
	    [string range $indata end-1 end] != "\n\n"} {
	return
    }

    #
    # http 1.0 or 1.1 request, serve as http 1.0
    #

    set serving 1
    if {[catch {
	puts -nonewline $sock "HTTP/1.0 200 Ok\r\n"
	puts -nonewline $sock "Content-Type: application/octet-stream\r\n"
	puts -nonewline $sock "\r\n"
    }]} {
	set complete -1
	set mutex 1
	return
    }

    fileevent $sock writable "fpsfhttpd::write $sock"
    return
}

proc fpsfhttpd::write {sock} {
    variable complete
    variable mutex
    variable file

    if {[eof $file]} {
	fileevent $sock writable ""
	if {[catch {close $sock}]} {
	    set complete -1
	} else {
	    set complete 1
	}
	set mutex 1
	return
    }

    if {[catch {
	set data [::read $file 10240]
	puts -nonewline $sock $data
    } err]} {
	fileevent $sock writable ""
	catch {close $sock}
	set complete -1
	set mutex 1
    }
}
