'\""
'\" Copyright (c) 2000-2001 Frank Pilhofer"
'\""
'\".so man.macros"
.TH Combat n 0.7 Combat "Tcl CORBA ORB"
.BS
.SH NAME
Combat \- Tcl CORBA Object Request Broker
.SH SYNOPSIS
.PD 0
\fBpackage require Tcl 8.1\fR
.PP
\fBpackage require Itcl 3.0\fR
.PP
\fBpackage require combat ?0.7?\fR
.PD
.PP
.PD 0
\fBcorba::init\fR \fR?\fIoptions ...\fR?\fR
.PP
\fBcorba::resolve_initial_references\fR \fIname\fR
.PP
\fBcorba::list_initial_services\fR
.PP
\fBcorba::register_initial_reference\fR \fIname\fR \fIobj\fR
.PP
\fBcorba::string_to_object\fR \fIstring\fR
.PP
\fBcorba::object_to_string\fR \fIhandle\fR
.PP
\fBcorba::release\fR \fR?\fItypecode\fR?\fR \fIvalue\fR
.PP
\fBcorba::duplicate\fR \fR?\fItypecode\fR?\fR \fIvalue\fR
.PP
\fBcorba::dii\fR ?\fIoptions\fR? \fIhandle\fR \fIspec\fR \fR?\fIargs\fR?\fR
.PP
\fBcorba::request\fR \fIcmd\fR \fR?\fIargs\fR?\fR
.PP
\fBcorba::const\fR \fIid\fR
.PP
\fBcorba::type\fR \fIcmd\fR \fR?\fIargs\fR?\fR
.PP
\fBcorba::throw\fR \fIexception\fR
.PP
\fBcorba::try\fR \fIm-block\fR \fR?\fBcatch {\fIexception \fR?\fIvar\fR?\fI\fB} \fIe-block\fR? \fI...\fR \fR?\fBfinally \fIf-block\fR?
.PP
\fBcombat::ir\fR \fIcmd\fR \fIargs\fR
.PD
.BE
.SH DESCRIPTION
.PP
The \fBCombat\fR package provides a CORBA Object Request Broker
(ORB). Using \fBCombat\fR, you can communicate with any CORBA
server using the standard IIOP protocol. This is not only useful for
test-driving your servers, but also for writing CORBA-enabled
applications, especially mixed with a Tk GUI.
.PP
Object references are represented by \fIhandles\fR, which are Tcl
commands that front for invocations. This looks quite natural:
.RS
.PP
set bank [corba::string_to_object IOR:...]
.PD 0
.PP
set account [$bank create]
.PP
$account deposit 1234
.PP
puts "Balance is [$account balance]"
.PD
.RE
.PP
In this example, \fB$bank\fR and \fB$account\fR are \fIhandles\fR
(strictly speaking, these are Tcl variables that contain a command
name). In an invocation, the first parameter is the operation name,
and following parameters are arguments to that operation. The format of
arguments is described in the section \fILanguage Mapping\fR below.
.PP
\fBCombat\fR works by reading information about the object's interface
from the Interface Repository. IDL files must be loaded into an
Interface Repository before they can be used. Read the \fIInterface
Repository\fR section below for more information.
.SH COMMANDS
.TP
\fBcorba::init\fR \fR?\fIoptions ...\fR?\fr
Initializes the ORB. The command accepts an arbitrary number of
options. Options starting with the \fI-ORB\fR prefix are processed
(see below), all list containing all other options is then
returned. \fBcorba::init\fR can therefore be called, at the beginning
of a script, as
.RS
.PP
set argv [eval corba::init $argv]
.PP
to remove all ORB-specific command-line parameters from the command
line. The call to \fBcorba::init\fR is optional. All other commands
implicitly call \fBcorba::init\fR with an empty parameter list if the
ORB was not initialized before.
.PP
The following options are accepted:
.IP "-ORBInitRef \fIname\fR=\fIvalue\fR"
Sets the initial reference \fIname\fR (see
\fBcorba::resolve_initial_references\fR) to \fIvalue\fR. The
\fIvalue\fR can have any format that is acceptable to
\fIcorba::string_to_object\fR.
.IP "-ORBDefaultInitRef \fIvalue\fR"
Sets the default initial reference. See
\fBcorba::resolve_initial_references\fR for more information.
.IP "-ORBDebug \fIlevel\fR"
Enables debugging output (which is sent to stderr). Valid values for
\fIlevel\fR are \fIgiop\fR (GIOP message exchange), \fIiiop\fR (IIOP
connection handling), \fItransport\fR (raw GIOP data), and \fIall\fR
(for all of them).
.IP "-ORBServerPort \fIport\fR"
Initializes a TCP socket to listen on port \fIport\fR for incoming
connections. This option can be repeated multiple times to listen on
several ports. If this option is not present, a port will be selected
automatically if the \fBRootPOA\fR is accessed for the first time.
.IP "-ORBGIOPMaxSize \fIvalue\fR"
Limits the maximum acceptable size of incoming GIOP messages. By
default, GIOP messages are accepted regardless of their size. This
allows a denial of service attack on a server by sending a huge
message, eventually causing memory exhaustion. If set, GIOP
messages whose size (in octets) exceeds \fIvalue\fR cause the
connection to be dropped. Each connection will also accept up to
\fIvalue\fR octets of GIOP fragments. The size of outgoing messages
is not limited by this option.
.IP "-ORBNativeCodeSet \fIvalue\fR"
Sets the native codeset to be used and advertised as SNCS-C. Usually,
the native codeset is determined from \fBencoding
system\fR. \fBvalue\fR can be an OSF registry value or Tcl encoding
name.
.PP
Any other options that start with \fI-ORB\fR cause an error.
.RE
.TP
\fBcorba::resolve_initial_references\fR \fIname\fR
Returns the initial reference \fIname\fR as set upon initialization,
as a handle (i.e. the value from \fI-ORBInitRef\fR is passed to
\fBcorba::string_to_object\fR). If no such initial reference is
configured, the default initial reference, if configured, and
\fIname\fR are concatenated and passed to
\fBcorba::string_to_object\fR to return a handle. If this fails, too,
the \fICORBA::ORB::InvalidName\fR exception is thrown.
.TP
\fBcorba::list_initial_services\fR
Returns a list of names configured as initial references.
.TP
\fBcorba::register_initial_reference\fR \fIname\fR \fIobj\fR
Registers the handle \fIobj\fR as an initial reference. \fIname\fR
will appear in the result of \fBlist_initial_services\fR and can be
used with \fBresolve_initial_references\fR to retrieve the handle.
.TP
\fBcorba::string_to_object\fR \fIstring\fR
Converts the \fIstring\fR to an object reference, and returns a handle
for that object reference. Several formats for \fIstring\fR are
accepted:
.RS
.IP \fBIOR:\fR
classic CORBA format for stringified object references, a very long
string consisting of the IOR: prefix and data in hex.
.IP \fBcorbaloc::\fIhost\fB:\fIport\fR/\fIObjectKey\fR
Format containing the IP address, port number and object key of the
target object.
.IP \fBcorbaname::\fIhost\fB[:\fIport\fB][/\fIKey\fB][#\fIName\fB]\fR
Format referencing an entry in the Naming Service by IP address. If
\fIport\fR is omitted, 2089 is used. \fIKey\fR is the object key of
the Naming Service. If omitted, \fBNameService\fR is used. \fIName\fR
identifies an entry in the Naming Service. This entry is read and
returned. If omitted, the reference of the Naming Service itself is
returned.
.IP \fBfile://[\fIhost\fB]\fIpath\fR
Reads the given file. \fIhost\fR should be empty or
\fBlocalhost\fR. \fIpath\fR is an absolute path name. The contents of
this file are read and then passed to \fBcorba::string_to_object\fR.
.IP \fBhttp://...\fR
The referenced URL is pulled (for this, the \fBhttp\fR package must be
present). Its contents are then passed to
\fBcorba::string_to_object\fR.
.RE
.TP
\fBcorba::object_to_string\fR \fIhandle\fR
Converts the object reference represented by \fIhandle\fR into a
string, using the \fBIOR:\fI format.
.TP
\fBcorba::release\fR \fR?\fItypecode\fR?\fR \fIvalue\fR
Releases a handle. Handles (as returned from
e.g. \fBcorba::string_to_object\fR) must be released when no longer
used, or memory leakage occurs. If \fItypecode\fR is omitted,
\fIvalue\fR must be a handle, which is then released. If
\fItypecode\fR is present, \fIvalue\fR must match that typecode. In
that case, all handles contained in \fIvalue\fR are released.
.TP
\fBcorba::duplicate\fR \fR?\fItypecode\fR?\fR \fIvalue\fR
Duplicates a handle. This method is primarily for use in server
applications, see the note in the \fIServer Side Scripting\fR
section. If \fItypecode\fR is omitted, \fIvalue\fR must be a handle,
which is then duplicated; the new handle, which points to the same
object reference, is returned. If \fItypecode\fR is present,
\fIvalue\fR must match that typecode. In that case, all handles
contained in \fIvalue\fR are duplicated, and a duplicate of the
complete \fIvalue\fR is returned.
.TP
\fBcorba::dii\fR ?\fIoptions\fR? \fIhandle\fR \fIspec\fR \fR?\fIargs\fR?\fR
Initiates a remote invocation using the \fIDynamic Invocation
Interface\fR. Invocations as described in the \fBInvocations\fR
section normally pull type information from the Interface Repository,
as described in the \fIInterface Repository\fR section below. In
contrast, an invocation using \fBdii\fR does not require type
information in the Interface Repository; here, type information is
passed along with each invocation in the \fIspec\fR parameter, which
is a list composed of three or four elements. The first element is the
typecode of the return value. The second element is the name of the
operation to be invoked. The third element describes the
parameters. The fourth element is a list of exception typecodes that
this operation may throw. The parameter description is a list that
contains one element per parameter. Each parameter is described by a
list of two elements. The first element is either \fBin\fR, \fBout\fR
or \fBinout\fR, and the second element is the typecode of the
parameter type.
.RS
.PP
\fIoptions\fR can be either \fB-async\fR or \fB-callback\fR to start
an asynchronous request; see below for more information.
.RE
.TP
\fBcorba::request\fR \fIcmd\fR \fR?\fIargs\fR?\fR
Manages asynchronous requests. See the \fBInvocations\fR section below
for information about initiating asynchronous requests using the
\fB-async\fR and \fB-callback\fR options to an invocation. \fIcmd\fR
and \fIargs\fR can be one of
.RS
.IP "\fBget\fR \fIaid\fR"
This operation enters the event loop until the asynchronous request
identified by \fIaid\fR is finished, and then returns its result. All
variables corresponding to \fBin\fR and \fBout\fR parameters on the
initial invocation are set in the current context.
.IP "\fBpoll\fR \fIaids ...\fR"
Checks whether any of the asynchronous requests identified by
\fIaids\fR has finished yet. If yes, its identifier is returned, and a
subsequent \fBget\fR for that identifier is guaranteed not to
block. If none of the given asynchronous requests has finished, an
empty string is returned.
.IP "\fBwait\fR \fR[\fIaids ...\fR]\fR"
Enters the event loop until any of the asynchronous requests
identified by \fIaids\fR finishes. If no parameter is given, all
outstanding asynchronous invocations qualify. The identifier for one
finished asynchronous invocation is returned.
.RE
.TP
\fBcorba::const\fR \fIid\fR
Looks up the \fIid\fR in the Interface Repository. \fIid\fR must be
the Repository Id or absolute name of a constant definition. The value
of that constant is then returned as an \fBany\fR value.
.TP
\fBcorba::type\fR \fIcmd\fR \fR?\fIargs\fR?\fR
Handles type definitions. \fIcmd\fR and \fIargs\fR can be one of
.RS
.IP "\fBof\fR \fIid\fR"
Looks up the \fIid\fR in the Interface Repository. \fIid\fR must be
the Repository Id or absolute name of a type definition. The typecode
of that type is returned.
.IP "\fBmatch\fR \fItypecode value\fR"
Checks whether \fIvalue\fR matches the \fItypecode\fR, and returns the
result as either 1 (matches) or 0 (does not match).
.IP "\fBequivalent\fR \fItc1 tc2\fR"
Checks whether the two typecodes \fItc1\fR and \fItc2\fR are
equivalent, and returns the result as either 1 (equivalent) or 0 (not
equivalent).
.RE
.TP
\fBcorba::throw\fR \fIexception\fR
Throws the CORBA exception \fIexception\fR. This is a convenience
operator; since exceptions use Tcl's normal error handling, you could
also throw exceptions using \fBerror\fR. See the \fILanguage
Mapping\fR section for information on the format of exeptions.
.TP
\fBcorba::try\fR \fIm-block\fR \fR?\fBcatch {\fIexception \fR?\fIvar\fR?\fI\fB} \fIe-block\fR? \fI...\fR \fR?\fBfinally \fIf-block\fR?
Provides a more convenient means of catching CORBA exceptions than
Tcl's native \fBcatch\fR command (all CORBA exceptions can be caught
with \fBcatch\fR, if preferred). First, the \fIm-block\fR is
executed. If an exception occurs, a matching \fBcatch\fR clause, whose
\fIexception\fR value matches the exception's Repository Id, is looked
for. If found, its \fIe-block\fR is executed, initializing the variable
denoted by \fIvar\fR, if present, to the exception value. Whether or
not an exception has occured, and whether or not this exception has
been handled by a \fBcatch\fR clause, the optional \fIf-block\fR is
then executed.
.RS
.PP
If an error occured in the \fIf-block\fR, that error is returned. If
an exception has occured, and it has been handled by a \fBcatch\fR
clause, the result from executing its \fIe-block\fR are
returned. Otherwise, the result from executing \fIm-block\fR is
returned.
.PP
To catch all exceptions, \fIexception\fR can be set to \fB...\fR
(three dots, similar to Java or C++).
.PP
Because exceptions are a specialization of Tcl errors,
\fBcorba::try\fR also responds to Tcl errors. If a Tcl error occurs in
the execution of \fIm-block\fR, the \fBfinally\fR clause will also be
executed. Tcl errors are caught in a \fBcatch\fR clause if \fB...\fR
is used as \fIexception\fR.
.PP
If there are no \fBcatch\fR clauses, an implicit clause that catches
\fB...\fR is used.
.RE
.TP
\fBcombat::ir\fR \fIcmd\fR \fIargs\fR
This command can be used to load interface information into a local
Interface Repository. At the moment, the only supported value for
\fIcmd\fR is \fIadd\fR. \fIargs\fR must be a string which is the
result of processing an IDL file with the \fIidl2tcl\fR compiler. By
adding information into a local Interface Repository, you can remove
your dependency on setting up an external Interface Repository in a
separate process. See the \fIInterface Repository\fR section for more
information.
.SH INVOCATIONS
Operations are invoked, and attributes are accessed, using the handle
as command. The first parameter to that command is the operation name,
and following parameters are arguments to that operation.
.RS
.PP
set handle [corba::string_to_object ...]
.PD 0
.PP
\fI$handle\fR ?\fIoptions\fR? \fIoperation\fR ?\fIargs ...\fR?
.PD
.RE
.PP
\fIoperation\fR must be a valid operation for the type of object
identified by \fIhandle\fR, and \fIargs\fR must match the parameter
list of that operation. For \fBinout\fR and \fBout\fR parameters, the
name of a variable must be passed. In the case of \fBinout\fR, that
variable must contain the value to be sent. In the case of both
\fBinout\fR and \fBout\fR, the variable is set to the value returned
by the operation.
.PP
The event loop is entered while waiting for the response from the
server. The operation's result is then returned as a result from the
invocation.
.PP
For attributes, use the attribute name for \fIoperation\fR. Without an
argument, the value of that attribute is returned. With one argument,
the attribute is set, and no value is returned.
.PP
The following operations are available for all handles:
.IP "\fB_is_a\fR \fIid\fR"
Returns 1 if the object identified by the handle is or is derived from
the Repository Id \fIid\fR. Returns 0 if the object identified by the
handle is not derived from \fIid\fR. Throws an exception if the
correct value can not be determined.
.IP "\fB_get_interface\fR"
Returns an object reference (as handle) to the \fBInterfaceDef\fR
object that identifies this type in the Interface Repository. This
requires that the server is connected to a well-configured Interface
Repository.
.IP "\fB_is_equivalent\fR \fIoh\fR"
Returns 1 if the object reference identified by the handle is
equivalent to the object reference identified by the handle
\fIoh\fR. Returns 0 if they can be reliably determined to be
different. Throws an exception otherwise.
.IP "\fB_duplicate\fR"
Returns a new handle for the same object reference. You need to use
this e.g. in servers which handle object references. See the \fIServer
Side Scripting\fR section below.
.PP
The following options are accepted:
.IP "\fB-async\fR"
Initiates an asynchronous invocation. Instead of entering the event
loop, the invocation returns immediately, and processing of the
invocation is handled in the background, within the event
loop. Instead of returning the operation's result, an \fIasynchronous
request identifier\fR is returned. This identifier can subsequently be
passed to the \fBcorba::request\fR to poll or wait for the operation
to complete and ultimately be used with the \fBget\fR subcommand of
\fBcorba::request\fR to receive the operation's result. Note that the
application must update the event loop in order to process
asynchronous requests.
.IP "\fB-callback\fR \fIcommand\fR"
Again, an asynchronous invocation is initiated, and an asynchronous
request identifier is returned. If the invocation has completed,
\fIcommand\fR is called at global level with the asynchronous request
identifier as single parameter. This callback procedure is expected to
call \fBcorba::request\fR \fBget\fR to receive the operation's result.
.SH LANGUAGE MAPPING
This section describes the mapping of IDL data types to Tcl types.
.IP "Primitive Types"
\fBshort\fR, \fBlong\fR, \fBunsigned short\fR, \fBunsigned long\fR,
\fBlong long\fR and \fBunsigned long long\fR values are mapped to
Tcl's integer type. Errors may occur if a value exceeds the numerical
range of Tcl's integer type.
.RS
.PP
\fBfloat\fR, \fBdouble\fR, \fBlong double\fR values are mapped to
Tcl's floating point type.
.PP
\fBstring\fR and \fBwstring\fR values are mapped to Tcl strings.
.PP
\fBboolean\fR values are accepted as 0, 1, true, false, yes and
no. In a result, they are always rendered as 0 (false) and 1 (true).
.PP
\fBoctet\fR, \fBchar\fR and \fBwchar\fR values are mapped to strings
of length 1.
.PP
\fBfixed\fR values are mapped to a floating-point value in exponential
representation. Depending on their scale and value, it may or may not
be possbile to use the value in a Tcl expression.
.RE
.IP "Structs"
\fBstruct\fR values are mapped to a list. For each element in
the structure, there are two elements in the list -- the first is
the element name, the second is the element's value. This allows to
easily assign structures from and to associative arrays, using
\fBarray get\fR and \fBarray set\fR.
.IP "Sequences"
\fBsequence\fR values are mapped to a list. As an exception, sequences
of \fBchar\fR, \fBoctet\fR or \fBwchar\fR are mapped to strings.
.IP "Arrays"
\fBarray\fR values are mapped to a list. As an exception, arrays of
\fBchar\fR, \fBoctet\fR or \fBwchar\fR are mapped to strings.
.IP "Enumerations"
\fBenum\fR values are mapped to the enumeration identifiers (without
any namespace qualifiers).
.IP "Unions"
\fBunion\fR values are mapped to a list of length 2. The first
element is the discriminator, or \fB(default)\fR for the default
member. The second element is the appropriate union member. Note that
the default case can also be represented by a concrete value distinct
from all other discriminator values.
.IP "Object References"
Object references are mapped to handles. Nil object references are
mapped to the integer value 0 (zero).
.IP "Exceptions"
\fBexception\fR values are mapped to a list of length one or two. The
first element is the Repository Id for the exception. If present, the
second element is the exception's contents, equivalent to the
structure mapping. The second element may be omitted if the exception
has no members. Exception handling is done using Tcl's normal error
handling mechanism; they are thrown using \fBerror\fR and can be
caught using \fBcatch\fR. The convenience operations
\fBcorba::throw\fR and \fBcorba::try\fR can be used instead (see
above).
.IP "Value Types"
\fBvaluetype\fR values are mapped to a list, like \fBstructs\fR. For
each element in the inheritance hierarchy of a \fBvaluetype\fR, there
are two elements in the list -- the first is the element name, and the
second is the element's value. An additional member \fB_tc_\fR may be
present. If present, its value must be a typecode. In an invocation,
this member determines the type to be sent. This mechanism allows to
send a derived valuetype where a base valuetype is expected. If no
\fB_tc_\fR member is present, the valuetype must be of the same type
as requested by the parameter. In receiving a valuetype, the _tc_
member is always added. A \fBvaluetype\fR can also be the integer 0
(zero) for a null value. \fBcustom\fR valuetypes are not supported.
.IP "Value Boxes"
Boxed \fBvaluetype\fR types are mapped to either the boxed type or to
the integer 0 (zero) for a null value. In the case of boxed integers,
the value 0 will always be read as a null value rather than a non-null
value containing the boxed integer zero. Shoot yourself in the foot if
you run into this problem.
.IP "Typecodes"
\fBTypeCode\fR values are mapped to a string containing a description
of the typecode:
.RS
.PP
Typecodes for the primitive types \fBvoid\fR, \fBboolean\fR,
\fBshort\fR, \fBlong\fR, \fBunsigned short\fR ,\fBunsigned long\fR,
\fBlong long\fR, \fBunsigned long long\fR, \fBfloat\fR, \fBdouble\fR,
\fBlong double\fR, \fBchar\fR, \fBoctet\fR, \fBstring\fR, \fBany\fR,
\fBTypeCode\fR are mapped to their name.
.PP
Bounded string typecodes are mapped to a list of length two. The first
element of the list is the identifier \fBstring\fR, the second element
is the bound.
.PP
Bounded wstring typecodes are mapped to a list of length two. The
first element of the list is the identifier \fBwstring\fR, the second
element is the bound.
.PP
\fBstruct\fR typecodes are mapped to a list of length three. The first
element is the identifier \fBstruct\fR. The second element is the
Repository Id, if available (else, the field may be empty). The third
element is a list with an even number of elements. The zeroth and
other even-numbered elements are member names, followed by the
member's typecode.
.PP
\fBunion\fR typecodes are mapped to a list of length four. The first
element is the identifier \fBunion\fR. The second element is the
Repository Id, if available (else, the field may be empty). The third
element is the typecode of the discriminator. The fourth element is a
list with an even number of elements. The zeroth and other
even-numbered elements are labels or the identifier \fBdefault\fR for
the default label, followed by the typecode of the associated member.
.PP
\fBexception\fR typecodes are mapped to a list of length three. The
first element is the identifier \fBexception\fR, the second element
the Repository Id, and the third element is a list with an even number
of elements. The zeroth and other even-numbered elements are member
names, followed by the member's typecode.
.PP
\fBsequence\fR typecodes are mapped to a list of length two or
three. The first element is the identifier \fBsequence\fR, the second
element is the typecode of the member type. The third element, if
present, denotes the sequence's bound. Otherwise, the sequence is
unbounded.
.PP
\fBarray\fR typecodes are mapped to a list of length three. The first
element is the identifier \fBarray\fR, the second element is the
typecode of the member type, the third element is the array's length.
.PP
\fBenum\fR typecodes are mapped to a list of length two. The first
element is the identifier \fBenum\fR, the second element is a list of
the enumeration identifiers.
.PP
Object reference typecodes are mapped to a list of length two. The
first element is the identifier \fBObject\fR, the second element is
the Repository Id of the IDL \fBinterface\fR.
.PP
\fBfixed\fR typecodes are mapped to a list of length three. The first
element is the identifier \fBfixed\fR. The second element is the
number of significant digits, the third element is the scale.
.PP
\fBvaluetype\fR typecodes are mapped to a list of length five. The
first element is the identifier \fBvaluetype\fR. The second element is
the Repository Id. The third element is a list of non-inherited
members. For each member, there are three elements in the list, a
visibility (\fBprivate\fR or \fBpublic\fR), the member name and the
member's typecode. The fourth element is the typecode of the
valuetype's concrete base, or 0 (zero) if the valuetype does not have
a concrete base. The fifth element is either an empty string or one of
the modifiers \fBcustom\fR, \fBabstract\fR or \fBtruncatable\fR.
.PP
Boxed \fBvaluetype\fR typecodes are mapped to a list of length 3. The
first element is the identifier \fBvaluebox\fR. The second element is
the Repository Id, and the third element is the typecode of the boxed
type.
.PP
A recursive reference to an outer type (in a \fBstruct\fR, \fBunion\fR
or \fBvaluetype\fR) can be expressed by a list of length two. The first
element is the identifier \fBrecursive\fR, the second element is the
Repository Id of the outer type, which must appear in the same
typecode description.
.PP
TypeCode values can be constructed manually, or retrieved from the
Interface Repository using the \fBcorba::type\fR command.
.RE
.IP "Any"
\fBany\fR values are mapped to a list of length two. The first element
is the typecode, and the second element is the value.
.RE
.SH INTERFACE REPOSITORY
The \fIInterface Repository\fR is vital for the operation of
\fBCombat\fR, and it is important that you understand its
importance. \fBCombat\fR is fully dynamic and posesses no
"compile-time" knowledge of interfaces. This is different from other
language mappings (e.g. C++ or Java), where such knowlege is available
by the way of an IDL "compiler" in the form of stubs and skeletons.
.PP
\fBCombat\fR instead pulls the information from the Interface
Repository, at runtime. \fBYou\fR must take care that an Interface
Repository is available and correctly configured. You will have to
load the Interface Repository with the IDL files that correspond to
your interfaces.
.PP
The \fBCombat\fR package does not include an Interface Repository, but
there is one available for download on the \fBCombat\fR home page,
from the Mico ORB. Using that Package, you would first start and load
the Interface Repository using, on the command line
.PP
\fBird --ior /tmp/ird.ior\fR
.PD 0
.TP
\fBidl --feed-ir -ORBInitRef InterfaceRepository=file:///tmp/ird.ior \fImyidlfile.idl\fR
.PD
.PP
On the second command line, replace \fImyidlfile.idl\fR with the name
of your IDL file. Repeat that command if you have multiple IDL files.
To make \fBCombat\fR use that Interface Repository, you would then use
.PP
\fBcorba::init -ORBInitRef InterfaceRepository=file:///tmp/ird.ior\fI
.PP
That above was a quick start, but unfortunately, things are more
complicated than that, which means that you have more options to
choose from. Read on for more detail.
.PP
Combat tries to do the following in order to get type information for
an interface:
.IP "1"
\fBCombat\fR sends a \fB_get_interface\fR request to the server. If
the server is connected to a correctly configured Interface Repository,
this will return a pointer to a server-side Interface Repository, and
\fBCombat\fR will then read interface information from
there. Unfortunately, few servers are set up this way.
.IP ""
You can try this for yourself by invoking the \fB_get_interface\fR
operation on a handle manually. If you get a handle in return, all is
fine. If you get a nil reference or an exception, this possibility has
failed.
.IP "2"
Next, \fBCombat\fR looks for a Repository Id inside the object
reference. Unfortunately, CORBA allows the Repository Id field in an
object reference to be empty, and the field is always empty for object
references of the \fBcorbaloc\fR kind. If a Repository Id is found,
\fBCombat\fR contacts its own Interface Repository, which must have
been configured as an initial reference by the name of
\fBInterfaceRepository\fR upon \fBcorba::init\fR, and tries to look up
the Repository Id in it.
.IP ""
You can try this for yourself by acquiring the
\fBInterfaceRepository\fR initial reference using
\fBcorba::resolve_initial_references\fR, and calling its
\fBlookup_id\fR operation with the Repository Id as parameter. If you
get a handle in return, all is fine. If you get an exception, this
possibility has failed.
.IP "3"
If both strategies above fail, then you have a problem. By the way,
one popular reason why the second strategy fails is that the object
reference is of a derived type, while only the base type is registered
in the Interface Repository (this frequently happens e.g. with the
Naming Service, where you load the Interface Repository with the
standard CosNaming IDL, but most ORBs actually implement a type that
is derived from CosNaming::NamingContextExt). Another frequent reason
is using \fBcorbaloc:\fR object references, which do not contain a
Repository Id.
.IP ""
In this case, you can try to force \fBCombat\fR into accepting a
certain type for that object reference by using the \fB_is_a\fR
operation:
.IP ""
\fI$handle\fR \fB_is_a\fR \fIid\fR
.IP ""
\fBCombat\fR will then ask the server whether it is compatible with
the Repository Id \fIid\fR. If yes, then \fBCombat\fR will use
\fIid\fR as the object reference's type, and look it up as above. In
the case of the Naming Service, you would e.g. use
.IP ""
.PD 0
\fBset ns [corba::resolve_initial_references NameService]\fR
.IP ""
\fB$ns _is_a IDL:omg.org/CosNaming/NamingContextExt:1.0\fR
.PD
.PP
The C++ version of \fBCombat\fR also includes a \fBidl2tcl\fR utility
that distils IDL files into Tcl strings, which can then be used to
populate an internal Interface Repository. \fBidl2tcl\fR is not
included here because of its dependency on the C++ version of Combat
and a separate ORB. Its files, however, can also be used with this Tcl
version of \fBCombat\fR.
.SH SERVER SIDE SCRIPTING
\fBCombat\fR includes a pretty complete implementation of the Portable
Object Adapter (POA). You can access the POA by resolving the
\fBRootPOA\fR initial reference. All operations on the POA follow the
normal language mapping rules as above.
.PP
A single exception that does not follow the language mapping rules is
the third parameter of the \fBcreate_POA\fR operation
(\fIpolicies\fR), which for convenience takes a list of policy
\fIidentifiers\fR instead of a list of policy \fIobjects\fR.
.PP
Servants are mapped to [incr Tcl] (or tcl++) objects and must
implement a public method \fB_Interface\fR that takes no parameters
and returns the Repository Id of the interface that the servant
implements. Operations are mapped to public methods, and attributes to
public variables. Servant classes may optionally inherit from
\fBPortableServer::ServantBase\fR. This inheritance is not required
per se, but only by this do servants inherit the \fB_this\fR method.
.PP
If your servant has operations that handle object references, then you
must be careful about memory management. As you may have noticed,
object references are not (cannot be) reference counted as they are in
e.g. C++. To avoid memory leakage, the runtime releases all handles
that are passed into your operations as \fBin\fR or \fBinout\fR
parameters. If you want to keep a copy, you must duplicate the handle
in question using the \fB_duplicate\fR method. Also, when you return
a handle as result, \fBinout\fR or \fBout\fR parameter, you must
return a duplicate, if you want to keep the original handle for
yourself. This rule applies recursively to all object references
embedded in other types. The \fBcorba::duplicate\fR command exists to
duplicate all handles embedded in a type.
.PP
Note that POA references are not object references, and need not be
released by \fBcorba::release\fR.
.PP
The \fBCookie\fR native type is mapped to string.
.SH CODESET SUPPORT
\fBCombat\fR can use about any popular codeset as transmission codeset
(\fITCS\fR), thanks to Tcl's excellent support in this area. See the
output of \fBencoding names\fR for supported codeset
encodings. However, a matching entry for each codeset must be present
in an internal table that maps the name of the Tcl encoding to the OSF
registry values used by CORBA. Currently, this table only includes
western codesets (because of the author's heritage and
laziness). Contact the author if you want support for other codesets.
.SH EVENT LOOP
\fBCombat\fR relies on fileevents for handling communication. In a
synchronous invocation, the event loop is entered until the response
from the server is received (internally, \fBupdate\fR is called). When
asynchronous requests are used, the application must call \fBupdate\fR
at times to allow processing to continue.
.PP
In a server, the event loop must be entered to allow reception of
incoming requests. Typically, a server application ends in a
`\fBvwait\fR \fIforever\fR' command, where the variable \fIforever\fR
is never set.
.SH FREQUENTLY ASKED QUESTIONS
.TP
All I get is \fBIDL:omg.org/CORBA/INTF_REPOS:1.0\fR!
This exception is thrown when \fBCombat\fR does not find type
information for an object reference. Make sure that an Interface
Repository is available, and set up to contain the correct
information. Read the \fIInterface Repository\fR section above for
more information.
.TP
I cannot talk to my Naming Service!
You need to configure the initial reference using \fB-ORBInitRef\fR
(see the documentation for \fBcorba::init\fR and
\fBcorba::string_to_object\fR).
.IP ""
Most ORBs have proprietary means of locating their own services. For
example the JDK ORB is able to find its own Naming Service
(\fItnameserv\fR) without any options or configuration efforts by
using default, compiled-in object references. Such proprietary
information is not available to \fBCombat\fR, so you must configure
all initial references using \fB-ORBInitRef\fR.
.IP ""
Also, in order to use the Naming Service, you will have to load the
IDL for the Naming Service into the Interface Repository.
.TP
Using \fBcorbaloc:\fR does not work!
The problem with \fBcorbaloc:\fR is that such object URLs do not
include the Repository Id. Hence, \fBCombat\fR will, as its last
resort to find out the object reference's type, send a
\fB_get_interface\fR request to the server, which will likely
fail. See the \fIInterface Repository\fR section above for more
information.
.IP ""
The solution is to give \fBCombat\fR a hint about the type by sending
a manual \fB_is_a\fR invocation with the expected Repository Id as a
parameter. \fBCombat\fR takes notice of succeeding \fB_is_a\fR
requests, and can then use that verified Repository Id for usage in
the local Interface Repository, e.g.
.IP ""
\fB$handle _is_a IDL:omg.org/CosNaming/NamingContextExt:1.0\fR
.IP ""
in the case of a Naming Service that has been contacted with a
\fBcorbaloc:\fR URL.
.TP
I cannot send or receive \fBwstring\fRs!
Codeset handling in \fBCombat\fR should be pretty correct. However,
the encodings for wide strings have been pretty broken in GIOP
versions prior to 1.2, and \fBCombat\fR refuses to send wstrings over
GIOP 1.0 or GIOP 1.1 connections. Check if the remote ORB supports
GIOP 1.2 (e.g. by running an iordump utility on its object
references). You might want to switch on GIOP debug messages
\fB-ORBDebug GIOP\fR, for \fBCombat\fR's comments on the situation.
.IP ""
One popular example is the ORB in JDK 1.2 or 1.3, which only supports
GIOP 1.1. The ORB in JDK 1.4 should work fine.
.SH TODO
Combat is a reasonably complete ORB. But here are ideas for new
features or sensible supporting projects.
.PP
Implement the remaining data types (long long, long double, fixed).
.PP
IIOP over SSL. Shouldn't be hard with TLS.
.PP
Portable Interceptors.
.PP
Rewrite \fBidl2tcl\fR in Tcl to avoid the dependency on
Combat/C++. Would be much easier if there was a grammar package (a la
Yacc) for Tcl.
.PP
Contact the author if you are interested in working on any of these
projects, or in funding them.
.SH SEE ALSO
The \fBCombat\fR homepage, http://www.fpx.de/Combat/
.SH KEYWORDS
CORBA, IIOP, GIOP, Object Request Broker
