#
# ======================================================================
# This file is part of Combat/Tcl, a Tcl CORBA Object Request Broker
#
# Please visit the Combat Web site at http://www.fpx.de/Combat/ for
# more information.
#
# Copyright (c) Frank Pilhofer, combat@fpx.de
#
# ======================================================================
#
# CVS Version Tag: $Id: combat.tcl,v 1.7 2002/02/26 10:50:59 fp Exp $
#
# ----------------------------------------------------------------------
# Main
# ----------------------------------------------------------------------
#

#
# Tcl 8.1 adds the internationalization features that are used for codeset
# negotiation. Also, [string] had some bugs with binary strings in 8.0.
#

package require Tcl 8.1

#
# Can work with Itcl 3.0 or Tcl++ 2.3. We prefer the former, but don't
# complain if the latter is already available.
#

if {[catch {package present tcl++ 2.3}]} {
    if {[catch {package require Itcl 3.0}]} {
	if {[catch {package require tcl++ 2.3}]} {
	    error "Oops in Combat initialization: neither \[incr Tcl\] nor tcl++ available"
	}
    } else {
	namespace import -force itcl::*
    }
}

#
# Load Combat
#

set _combat_mandatory_files {
    cdr giop iop iiop codeset str object orb corba
}

set _combat_optional_files {
    poa
}

foreach _combat_file $_combat_mandatory_files {
    set _combat_fullname [file join [file dirname [info script]] \
	    [set _combat_file].tcl]
    source $_combat_fullname
}

foreach _combat_file $_combat_optional_files {
    set _combat_fullname [file join [file dirname [info script]] \
	    [set _combat_file].tcl]
    if {[file exists $_combat_fullname]} {
	source $_combat_fullname
    }
}

#
# combat Namespace
#

namespace eval combat {
    proc ir {cmd data} {
	if {$cmd != "add"} {
	    error "illegal option for combat::ir, was expecting add"
	}
	::Combat::SimpleTypeRepository::add $data
    }
}

package provide corba 0.7
package provide combat 0.7

