#! /bin/sh
# \
exec tclsh8.3 "$0" ${1+"$@"}

lappend auto_path ../..
package require combat

#
# The Account server
#

class Account_impl {
    inherit PortableServer::ServantBase

    private variable mybank

    constructor {bank} {
	set mybank $bank
    }

    public method _Interface {} {
	return "IDL:Account:1.0"
    }

    private variable balance 0

    public method deposit { amount } {
	set balance [expr $balance + $amount]
    }

    public method withdraw { amount } {
	if {$amount > $balance} {
	    corba::throw [list IDL:Account/Bankrupt:1.0 \
		    [list balance $balance amount $amount]]
	}
	set balance [expr $balance - $amount]
    }

    public method balance {} {
	return $balance
    }

    public method destroy {} {
	set obj [corba::resolve_initial_references POACurrent]
	set poa [$obj get_POA]
	set oid [$obj get_object_id]
	$poa deactivate_object $oid
	$mybank goodbye $oid
	delete object $this
    }
}

class Bank_impl {
    inherit PortableServer::ServantBase

    private variable accounts
    private variable mypoa

    constructor {poa} {
	set mypoa $poa
    }

    public method _Interface {} {
	return "IDL:Bank:1.0"
    }

    public method create {name password} {
	if {[info exists accounts($name)]} {
	    if {$accounts($name) != $password} {
		corba::throw IDL:Bank/NotAuthorized:1.0
	    }
	    return [$mypoa id_to_reference $name]
	}
	set accounts($name) $password
	set acc [namespace current]::[::Account_impl #auto $this]
	$mypoa activate_object_with_id $name $acc
	return [$mypoa id_to_reference $name]
    }

    public method goodbye {name} {
	unset accounts($name)
    }
}

#
# Initialize ORB and feed the local Interface Repository
#

eval corba::init $argv
source account.tcl
combat::ir add $_ir_account

#
# Create a new POA with the USER_ID policy
#

set poa   [corba::resolve_initial_references RootPOA]
set mgr   [$poa the_POAManager]
set mypoa [$poa create_POA MyPOA $mgr {RETAIN USER_ID}]

#
# Create a Bank and activate it
#

set srv [Bank_impl #auto $mypoa]
set oid [$poa activate_object $srv]

#
# write Bank's IOR to file
#

set reffile [open "server.ior" w]
set ref [$poa id_to_reference $oid]
set str [corba::object_to_string $ref]
puts -nonewline $reffile $str
close $reffile

#
# Activate the POAs
#

$mgr activate

#
# .. and serve the bank ...
#

puts "Running."
vwait forever

puts "oops"
