




# use cmdline in tcllib for improved introspection
# fix-up version spaces.


package provide dom_pretty_print 0.2

package require Tcl 8.2
package require dom

namespace eval dom {
    namespace export pretty_print
}      


########
#
#	Copyright 2001, Phaseit, Inc.
#
########
########
#
#  This is an extremely crude implementation.  It's just a badly
#      hacked version of the [dom:DOMImplmentation serialize ...]
#      from TclDOM-1.3.  The first change on the way to better
#      maintainability would be to improve argument-passing, which
#      I've greatly corrupted from Steve's style.
#
########
#
#  Example invocations:
#      dom::pretty_print $root
#      dom::pretty_print $document \t
#      dom::pretty_print $node {   } { }
#
########
proc dom::pretty_print {document
                       {indentation "  "}
                       {attribute_style ""}
                       {? ?}
                       {depth 0}} {
            GetHandle documentFragment $document node
            return [eval [list PrettyPrint:$node(node:nodeType) $depth \
                                   $indentation $attribute_style] $document] 
}   



#################################################
#
# Serialization and pretty-printing
#
#################################################

# dom::PrettyPrint:documentFragment --
#
#	Produce text for documentFragment.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:documentFragment {depth indentation attribute_style token args} {
    array set node [set $token]

    if {[string compare "node1" $node(documentFragment:masterDoc)]} {
	return [eval [list PrettyPrint:node $depth $indentation $attribute_style $token] $args]
    } else {
	if {[string compare {} [GetField node(document:documentElement)]]} {
	    return [eval PrettyPrint:document $depth [list $indentation] [list $attribute_style] [list $token] $args]
	} else {
	    return -code error "document has no document element"
	}
    }

}

# dom::PrettyPrint:document --
#
#	Produce text for document.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:document {depth indentation attribute_style token args} {
    array set node [set $token]

    if {![info exists node(document:documentElement)]} {
	return -code error "document has no document element"
    } elseif {![string length node(document:doctype)]} {
	return -code error "no document type declaration given"
    } else {

	array set doctype [set $node(document:doctype)]

	# BUG: Want to serialize all children except for the 
	# document element, and then do the document element.

	# Bug fix: can't use PrettyPrint:attributeList for XML declaration,
	# since attributes must occur in a given order (XML 2.8 [23])

	return "<?xml[PrettyPrint:XMLDecl version $node(document:xmldecl)][PrettyPrint:XMLDecl encoding $node(document:xmldecl)][PrettyPrint:XMLDecl standalone $node(document:xmldecl)]?>\n<!DOCTYPE $doctype(doctype:name)[expr {[string length $doctype(doctype:externalid)] ? " PUBLIC[PrettyPrint:ExternalID $doctype(doctype:externalid)]" : {}}][expr {[string length $doctype(doctype:internaldtd)] ? " \[$doctype(doctype:internaldtd)\]" : {}}]>\n[eval PrettyPrint:element $depth $[list indentation] [list $attribute_style] [list $node(document:documentElement)] $args]"
    }

}

# dom::PrettyPrint:ExternalID --
#
#	Returned appropriately quoted external identifiers
#
# Arguments:
#	id	external indentifiers
#
# Results:
#	text

proc dom::PrettyPrint:ExternalID id {
    set result {}

    foreach ident $id {
	append result { } \"$ident\"
    }

    return $result
}

# dom::PrettyPrint:XMLDecl --
#
#	Produce text for an arbitrary node.
#	This simply serializes the child nodes of the node.
#
# Arguments:
#	attr	required attribute
#	attList	attribute list
#
# Results:
#	XML format text.

proc dom::PrettyPrint:XMLDecl {attr attrList} {
    array set data $attrList
    if {![info exists data($attr)]} {
	return {}
    } elseif {[string length $data($attr)]} {
	return " $attr='$data($attr)'"
    } else {
	return {}
    }
}

# dom::PrettyPrint:node --
#
#	Produce text for an arbitrary node.
#	This simply serializes the child nodes of the node.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:node {depth indentation attribute_style token args} {
    array set node [set $token]

    set result {}
    foreach childToken [set $node(node:childNodes)] {
	catch {unset child}
	array set child [set $childToken]
	append result [eval [list PrettyPrint:$child(node:nodeType) \
                      [expr $depth + 1] $indentation $attribute_style $childToken] $args]
    }

    return $result
}

# dom::PrettyPrint:element --
#
#	Produce text for an element.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:element {depth indentation attribute_style token args} {
    array set node [set $token]
    array set opts {-newline {}}
    array set opts $args

    set indent [string repeat $indentation $depth]
    set result {}
    set newline {}
    if {[lsearch $opts(-newline) $node(node:nodeName)] >= 0} {
	# append result \n
	set newline \n
    }
# I think this takes care of some top-level stuff.
    if $depth {
       append result \n
    }
    append result "$indent<$node(node:nodeName)"
    append result [PrettyPrint:attributeList $depth $indentation $attribute_style [array get $node(element:attributeList)]]

    if {![llength [set $node(node:childNodes)]]} {

	append result />

    } else {

        append result >
	# Do the children
	append result [eval PrettyPrint:node $depth [list $indentation] [list $attribute_style] [list $token] $args]

        # What's the right way to handle this?
        set is_text 0
        foreach child [dom::node children $token] {
            if [string match textNode [dom::node cget $child -nodeType]] {  
                set is_text 1
                break
            }
        }
        if {!$is_text} {
            append result \n$indent
        }
	append result "</$node(node:nodeName)>"

    }

    return $result
}

# dom::PrettyPrint:textNode --
#
#	Produce text for a text node.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:textNode {depth indentation attribute_style token args} {
    array set node [set $token]

    return [Encode $node(node:nodeValue)]
}

# dom::PrettyPrint:processingInstruction --
#
#	Produce text for a PI node.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:processingInstruction {depth indentation attribute_style token args} {
    array set node [set $token]

    return "<$node(node:nodeName) $node(node:nodeValue)>"
}

# dom::PrettyPrint:comment --
#
#	Produce text for a comment node.
#
# Arguments:
#	token	node token
#	args	configuration options
#
# Results:
#	XML format text.

proc dom::PrettyPrint:comment {depth indentation attribute_style token args} {
    array set node [set $token]

    return <!--$node(node:nodeValue)-->
}


# dom::PrettyPrint:attributeList --
#
#	Produce text for an attribute list.
#
# Arguments:
#	l	name/value paired list
#
# Results:
#	XML format text.

proc dom::PrettyPrint:attributeList {depth indentation attribute_style l} {

    set result {}
    foreach {name value} $l {

	append result { } $name ${attribute_style}=${attribute_style}

	# Handle special characters
	regsub -all < $value {\&lt;} value

	if {![string match *\"* $value]} {
	    append result \"$value\"
	} elseif {![string match *'* $value]} {
	    append result '$value'
	} else {
	    regsub -all \" $value {\&quot;} value
	    append result \"$value\"
	}

    }

    return $result
}
